/*
** $Id: agncmpt.h $
** Compatibility definitions
** See Copyright notices in this file, and - where not indicated - in agena.h
** initiated by Alexander Walz, before 2009
*/

#ifndef agncmpt_h
#define agncmpt_h

#include "agnhlps.h"
#include "prepdefs.h"

/* In OpenSUSE 10.3, trunc, log2, exp2, fma, fdim do not exist although their definitions are included in tgmath.h */

#if (defined(__SOLARIS)) || defined(__linux__) && !defined(OPENSUSE) || defined(__APPLE__)  /* trunc, log2, exp2, fma, fdim */
#include <tgmath.h>
#endif

/* DJGPP does not have fma, fdim, trunc, lstat */
#if !(defined(_WIN32) || defined(__SOLARIS) || defined(__linux__) && !defined(OPENSUSE) || defined(__APPLE__))  /* 2.8.4 / 2.17.0 */
#undef fma
#define fma tools_fma

#undef fdim
#define fdim tools_fdim

#undef trunc
#define trunc(x)  ( ((x) > 0) ? sun_floor((x)) : sun_ceil((x)) )

#undef lstat
#define lstat stat

#endif

#if !(defined(_WIN32) || defined(__SOLARIS) || defined(__linux__) && !defined(OPENSUSE) || defined(LUA_DOS) || defined(__APPLE__))  /* 2.8.4 fix */
#undef log2
#define log2(x)  ( ((x)) > 0 ? log((x)) * M_LOG2E : AGN_NAN )  /* sun_log creates round-off errors with 8 and 64 in OpenSUSE 10.3 ! Changed 2.12.0 RC 4 */
#endif

/* 2.8.4, we use the Cephes implementation for the antilog2 operator on all platforms: */
#undef exp2  /* 2.8.4 improvement */
#define exp2 cephes_exp2

/* faster than exp((x) * LN10//log(10)); DJGPP _has_ exp10 */
/* 2.8.4, we use the Cephes implementation for the antilog10 operator on all platforms: */
#undef exp10  /* 2.8.4 improvement */
#define exp10 cephes_exp10

#undef scalb  /* for any BSD-style code */
#define scalb ldexp

#if defined(__SOLARIS)  /* 2.8.4, 2.11.0 RC2 */
#ifndef isunordered
#define isunordered(x, y) __builtin_isunordered(x, y)
#endif
#elif !(defined(_WIN32) || defined(__linux__) && !defined(OPENSUSE) || defined(__APPLE__) || defined(__OS2__) )
#ifndef isunordered  /* prevent warnigns in OpenSUSE */
#define isunordered(x, y)  (!(((x) != AGN_NAN) && ((y) != AGN_NAN)))
#endif
#endif

#if defined(__SOLARIS)  /* 2.8.4, 2.11.0 RC2  */
#ifndef isfinite
#define isfinite(x) __builtin_isfinite(x)
#endif
#elif !(defined(_WIN32) || defined(__linux__) && !defined(OPENSUSE) || defined(__APPLE__) || defined(__OS2__) || defined(__HAIKU__) )
#ifndef isfinite  /* prevent warnigns in OpenSUSE */
#define isfinite tools_isfinite
#endif
#endif

#if defined(__SOLARIS)  /* 2.8.4 fix and extension, 2.11.0 RC2  */
#ifndef signbit
#define signbit(x) __builtin_signbit(x)
#endif
#elif !(defined(_WIN32) || defined(__linux__) && !defined(OPENSUSE) || defined(__APPLE__) || defined(__OS2__) )  /* DJGPP does not know signbit */
#undef signbit
#define signbit tools_signbit  /* FIXME: testing for -0.0 does not work in OpenSUSE 10.3/GCC 4.4.5 */
#endif

/* Solaris 10 does not have isinf */
#if defined(__SOLARIS)  /* 2.9.0 */
#ifndef isinf
#define isinf(x)  ((x) == -HUGE_VAL || (x) == HUGE_VAL)
#endif
#endif

/* needed to compile under Solaris; orig in MinGW: 0x8000 */
#if !defined(_WIN32) && !defined(LUA_DOS) && !defined(__OS2__)
#define O_BINARY      0x0000
#endif

/* at least in GCC there are problems with fmax, see http://gcc.gnu.org/ml/gcc-bugs/2004-01/msg00620.html:
  `The symptom is that values of the mathematical functions fmaxf, fminf, fmax, fmin, fmaxl, fminl are
   all wrong. This makes the results of some numerical computations junks.` - 0.26.4. */
#define fMax(a, b)  (((a) > (b)) ? (a) : (b))  /* substitute */
#define fMin(a, b)  (((a) < (b)) ? (a) : (b))  /* substitute */

/* define constant `undefined` */
#ifdef NAN
#define AGN_NAN NAN
#elif defined(INFINITY)
#define AGN_NAN (INFINITY/INFINITY)
#else
#define AGN_NAN (tools_nan())
#endif

#if defined(LUA_RASPI_STRETCH) || defined(__linux__)
#define stretch_isnan(z) ( luai_numisnan(creal(z)) || luai_numisnan(cimag(z)) )
#endif
/* define complex unit I, which natively is not known to Solaris */
#ifdef __unix__
#undef _Complex_I
#define _Complex_I   (0.0F + 1.0iF)
#undef I
#define I   _Complex_I
#endif

/* The floating-point expression evaluation method, not known in MinGW/GCC 3.4.6
       -1  indeterminate
        0  evaluate all operations and constants just to the range and
           precision of the type
        1  evaluate operations and constants of type float and double
           to the range and precision of the double type, evaluate
           long double operations and constants to the range and
           precision of the long double type
        2  evaluate all operations and constants to the range and
           precision of the long double type;
   According to various newsgroup, its value is always 0. */
#ifndef FLT_EVAL_METHOD
#define FLT_EVAL_METHOD 0
#endif


#ifndef math_opt_barrier
#define math_opt_barrier(x) \
({ __typeof (x) __x = x; __asm ("" : "+m" (__x)); __x; })
#define math_force_eval(x) __asm __volatile ("" : : "m" (x))
#endif

#if defined(LUA_DOS)
#  define fopen64 fopen
#  define ftello64 ftell
#  define _ftelli64 ftello64
#  define fseeko64 fseek
#  define _fseeki64 fseeko64
#  define lseek64 lseek
#else
#if defined(__SOLARIS)  /* 2.11.0 RC3 */
#  define _fseeki64 fseeko64
#  define _ftelli64 ftello64
#else
#  define fopen64 fopen
#  define ftello64 ftell
#  define _ftelli64 ftello64
#  define fseeko64 fseek
#  define _fseeki64 fseeko64
#  define lseek64 lseek
#endif
#endif

#ifndef LUA_DOS
#  include <stdint.h>     /* for int32_t */
#  include <sys/types.h>  /* for off_t & off64_t */
#  ifndef off64_t
#    define off64_t long long int
#  endif
#endif


/* TYPEDEF 2.14.4 */
#include <stdint.h>

/* 2.10.4 fix as MinGW and Solaris 10 use a 4-byte integers with int64_t and off64_t */
#if defined(_WIN32) || (defined(__SOLARIS)) || defined(LUA_DOS) || defined(__linux__) || defined(OPENSUSE)
#undef int64_t
#define int64_t long long int
#undef off64_t  /* 2.11.0 RC2 fix */
#define off64_t int64_t
#else
#  ifndef int64_t
#     define int64_t off64_t
#  endif
#endif

#ifndef ULLONG_MAX
#  ifdef ULONG_LONG_MAX
#    define ULLONG_MAX ULONG_LONG_MAX
#  endif
#endif

#ifndef LLONG_MIN
#  ifdef LONG_LONG_MIN
#    define LLONG_MIN LONG_LONG_MIN
#  endif
#endif

#ifndef LLONG_MAX
#  ifdef LONG_LONG_MAX
#    define LLONG_MAX LONG_LONG_MAX
#  endif
#endif


#define TRUNC	trunc

#if !defined(umount2) && defined(umount)
  #define umount2((x), (y))	umount((x))
#endif

#ifndef FP_ZERO
#define FP_ZERO          0x4000
#endif
#ifndef FP_NORMAL
#define FP_NORMAL        0x0400
#endif
#ifndef FP_NAN
#define FP_NAN           0x0100
#endif
#ifndef FP_SUBNORMAL
#define FP_SUBNORMAL     (FP_NORMAL | FP_ZERO)
#endif
#ifndef FP_INFINITE
#define FP_INFINITE      (FP_NAN | FP_NORMAL)
#endif
#ifndef FP_ILOGBNAN
#define FP_ILOGBNAN      ((int)0x80000000)
#endif

/* Predicitve branching, 2.15.5, idea and macros taken from Arm Optimized Routines */
#ifdef __GNUC__
# define likely(x)       __builtin_expect (!!(x), 1)
# define unlikely(x)     __builtin_expect ((x), 0)
#else
# define likely(x)       (x)
# define unlikely(x)     (x)
#endif

#endif

