/*
** $Id: agnhlps.c $
** C Helper routines
** See Copyright notices in this file, and - where not indicated - in agena.h
** initiated Alexander Walz, July 20, 2007
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdarg.h>
#include <limits.h>     /* for PATH_MAX length, cast_num */
#include <sys/stat.h>
#include <math.h>       /* for pow, floor, fabs, etc. */
#include <locale.h>     /* for localeconv */
#include <errno.h>      /* for my_ioerror */
#include <time.h>
#include <float.h>      /* for sincos, rem_pio2 */
#include <sys/timeb.h>  /* ftime, for milliseconds */

#include "prepdefs.h"   /* preprocessor aliases, repositioned 2.14.5 */

#ifdef __linux__
#include <stdint.h>
#endif

#ifdef CPUID
#include <cpuid.h>
#endif

#ifndef PROPCMPLX
#include <complex.h>
#endif

#ifdef _WIN32
#include <sys/locking.h>
#include <windows.h>   /* free memory */
#include <ntdef.h>     /* for GetProcAddress, etc */
#endif

#if defined(__unix__) || defined(__APPLE__)  /* for getch */
#include <termios.h>
#include <unistd.h>
#endif

#if defined(_WIN32) || defined(__DJGPP)
#include <conio.h>   /* getch; a UNIX version is included in agnhlps.c */
#endif

#ifdef __APPLE__
  #include <crt_externs.h>
  #define environ (*_NSGetEnviron())
#elif defined(__unix__)
  extern char **environ;
#endif

#if defined(__OS2__)
#include <unistd.h>
#include <conio.h>  /* for kbhit */
#include <os2emx.h>
#endif

#define agnhlps_c
#define LUA_LIB

#include "agncfg.h"    /* read in endianness */
#include "agncmpt.h"   /* for O_BINARY and FP_* */
#include "agnhlps.h"   /* for INLINE alias */
#include "cephes.h"    /* for powi, cephes_gamma */
#include "sofa.h"      /* for iauJd2cal */
#include "lucase.def"  /* for tools_lowercase */
#include "lmathlib.h"  /* for seeds m_z, m_w */

/* instead of time.h: */
#include "agnt64.h"

#define BASE_MOVELEN  16384L     /* number of bytes in index section to be moved in a write or delete operation */
#define BASE_OFFSET     256L     /* first position of an index to a key */
#define MAXNRECORDS     104L     /* position of Maximum Records allowed */
#define BASECOMMENT     129L     /* position of comment */

#define uchar(c) ((unsigned char)(c))

#ifndef MAX_PATH
#define MAX_PATH PATH_MAX
#endif


/* See: https://www.greenend.org.uk/rjk/tech/inline.html */
#if defined __GNUC__ && !defined __GNUC_STDC_INLINE__ && !defined __GNUC_GNU_INLINE__
# define __GNUC_GNU_INLINE__ 1
#endif

/* In Linux and with DJGPP, all inline functions must be pre-declared here: */
INLINE int tools_isfloat (double x);
INLINE int tools_isregint (double x);
INLINE int tools_isregnonnegint (double x);
INLINE int tools_isint (double x);
INLINE int tools_isnegint (double x);
INLINE int tools_isposint (double x);
INLINE int tools_isnonnegint (double x);
INLINE int tools_isuint (double x, int bits);
INLINE double tools_random (int mode);
INLINE int tools_randomrange (int l, int u, int mode);
INLINE ptrdiff_t tools_posrelat (ptrdiff_t pos, size_t len);
INLINE uint32_t tools_msb (register uint32_t x);
INLINE uint32_t tools_lsb (register uint32_t x);
INLINE uint32_t tools_clz32 (register uint32_t x, uint32_t *y);
INLINE uint32_t tools_ctz32 (register uint32_t x, uint32_t *y);
INLINE uint32_t tools_onebits (register uint32_t x);
INLINE uint32_t tools_flip (register uint32_t n);
INLINE uint32_t tools_fibmod (uint32_t k, uint32_t m, int exact);
INLINE double tools_mhypot (double a, double b);
INLINE void *tools_realloc (void *p, size_t size);
INLINE int tools_cpuid (uint32_t call, uint32_t *eax, uint32_t *ebx, uint32_t *ecx, uint32_t *edx);
INLINE int tools_isevenorodd (double x);

INLINE double sun_log (double x);
INLINE double sun_log2 (double x);
INLINE double sun_log10 (double x);
INLINE double sun_exp (double x);
INLINE double sun_pow (double x, double y);
INLINE double sun_hypot (double x, double y);
INLINE int32_t sun_exponent (double x);
INLINE int sun_irregular (double x);

INLINE size_t tools_optstrlen (size_t l, size_t *chunks);
INLINE char *tools_stralloc (size_t len);

INLINE void *tools_memchr (const void *src, int c, size_t l);
INLINE int tools_memcmp (const void *haystack, const void *needle, size_t l);
INLINE const char *tools_lmemfind (const char *s1, size_t l1, const char *s2, size_t l2);
INLINE size_t tools_strlen (const char *t);

/* do not inline/INLINE cpack* functions here, as this causes trouble in OpenSUSE 10.3 & Mac */


/* Constants used by Sun's functions */

static volatile double
pi_lo   = 1.2246467991473531772E-16,       /* 0x3CA1A626, 0x33145C07 */
tiny    = 1.0e-300,
twom1000= 9.33263618503218878990e-302;     /* 2**-1000=0x01700000,0*/


LUALIB_API void tools_quicksort (off64_t v[], long int left, long int right) {  /* 2.11.0 RC2 change */
  register long int i, last;
  register off64_t temp;
  if (left >= right) {
    return;
  }
  SWAP(v[left], v[(left + right)/2], temp);
  last = left;
  for (i=left + 1; i <= right; i++) {
    if (v[i] < v[left]) {
      last++;
      SWAP(v[last], v[i], temp);
    }
  }
  SWAP(v[left], v[last], temp);
  tools_quicksort(v, left, last - 1);
  tools_quicksort(v, last + 1, right);
}


/* Array must be sorted !  cnt is the number of elements in array a. z is the integer to be searched. */
LUALIB_API int tools_binsearch (off64_t *a, off64_t cnt, off64_t z) {  /* 2.11.0 RC 2 */
  off64_t i, j, x;
  i = 0; j = cnt - 1;
  while (i <= j) {
    x = (i + j)/2;
    if (z == a[x])
      return 1;
    if (a[x] < z)
      i = x + 1;
    else
      j = x - 1;
  }
  return 0;
}


/* left: left border, right: right border of array */
LUALIB_API void tools_dquicksort (double v[], long int left, long int right) {
  register long int i, last;
  register double temp;
  if (left >= right) return;
  SWAP(v[left], v[(left + right)/2], temp);
  last = left;
  for (i=left + 1; i <= right; i++) {
    if (v[i] < v[left]) {
      last++;
      SWAP(v[last], v[i], temp);
    }
  }
  SWAP(v[left], v[last], temp);
  tools_dquicksort(v, left, last - 1);
  tools_dquicksort(v, last + 1, right);
}


/* Based on the heapsort algorithm as published in Niklaus Wirth's `Algorithms and Data Structures` (for Oberon),
   extended to optionally sort only a portion of an array; 2.3.0 */

static INLINE void dsift (double *a, long int L, long int R, long int offset) {  /* there is no benefit when integrating `dsift` into dheapsort, inlined 2.16.13 */
  register long int i, j;
  register double x;
  i = L; j = 2*i + 1 - offset; x = a[i];
  if (j < R && a[j] < a[j + 1]) j++;
  while (j <= R && x < a[j]) {
    a[i] = a[j]; i = j; j = 2*j + 1 - offset;
    if (j < R && a[j] < a[j + 1]) j++;
  }
  a[i] = x;
};

LUALIB_API void tools_dheapsort (double *a, long int l, long int r) {
  register long int L, R;
  register double temp;
  L = l + (r - l + 1)/2; R = r;
  while (L > l) {
    L--; dsift(a, L, R, l);
  }
  while (R > l) {
    SWAP(a[l], a[R], temp);  /* 2.15.3 change */
    R--; dsift(a, L, R, l);
  }
}


/* left: left border, right: right border of array, depth initially is 0, threshold is 2*log(#entries)/log(2) */
LUALIB_API void tools_dintrosort (double v[], long int left, long int right, size_t depth, size_t threshold) {  /* 2.3.0 */
  register long int i, last;
  register double temp;
  if (left >= right) return;
  depth++;
  if (depth == threshold) {
    tools_dheapsort(v, left, right);
    return;
  }
  SWAP(v[left], v[(left + right)/2], temp);
  last = left;
  for (i=left + 1; i <= right; i++) {
    if (v[i] < v[left]) {
      last++;
      SWAP(v[last], v[i], temp);
    }
  }
  SWAP(v[left], v[last], temp);
  tools_dintrosort(v, left, last - 1, depth, threshold);
  tools_dintrosort(v, last + 1, right, depth, threshold);
}


/*

Function: pixel_qsort

Old and supposedly optimised quicksort algorithm.

Function : pixel_qsort()

    In : pixel array, size of the array
    Out : int
    Job : sort out the array of pixels
    Note : optimized implementation.

References PIX_STACK_SIZE, and PIX_SWAP.

Taken from https://github.com/wme7/aero-shock/blob/master/mycpp/test/benchmed.c,
written by Nicolas Devillard <ndevilla@free.fr> August 1998. This code in public domain.

The function _can_ be 15 % faster on older systems than the recursive quicksort implementation for
lists of numbers in random order. Only if the list is already sorted in descending order, pixel sort
is much slower.

See also: http://www.eso.org/projects/dfs/papers/jitter99/node27.html

Modified by Alexander Walz to return a success code: 0 - failure, 1 - success, and to dynamically
increase the internal stack instead of throwing an error if the stack size has become too small.

*/

#define PIX_SWAP(a,b) { double temp=(a);(a)=(b);(b)=temp; }
#define PIX_STACK_SIZE 50


LUALIB_API int pixel_qsort (double *pix_arr, int npix) {  /* changed from void to int by Alexander Walz */
  int i, ir, j, k, l, j_stack;
  int *i_stack;
  double  a;
  ir = npix;
  l = 1;
  j_stack = 0;
  i_stack = malloc(PIX_STACK_SIZE * sizeof(double));
  if (i_stack == NULL) return 0;  /* 2.3.0 */
  for (;;) {
    if (ir - l < 7) {
      for (j=l + 1; j <= ir; j++) {
        a = pix_arr[j - 1];
        for (i=j - 1; i >= 1; i--) {
          if (pix_arr[i - 1] <= a) break;
          pix_arr[i] = pix_arr[i - 1];
        }
        pix_arr[i] = a;
      }
      if (j_stack == 0) break;
      ir = i_stack[j_stack-- -1];
      l  = i_stack[j_stack-- -1];
    } else {
      k = (l+ir) >> 1;
      PIX_SWAP(pix_arr[k - 1], pix_arr[l])
      if (pix_arr[l] > pix_arr[ir - 1]) {
        PIX_SWAP(pix_arr[l], pix_arr[ir - 1])
      }
      if (pix_arr[l - 1] > pix_arr[ir - 1]) {
        PIX_SWAP(pix_arr[l - 1], pix_arr[ir - 1])
      }
      if (pix_arr[l] > pix_arr[l - 1]) {
        PIX_SWAP(pix_arr[l], pix_arr[l - 1])
      }
      i = l + 1;
      j = ir;
      a = pix_arr[l - 1];
      for (;;) {
        do i++; while (pix_arr[i - 1] < a);
        do j--; while (pix_arr[j - 1] > a);
        if (j < i) break;
        PIX_SWAP(pix_arr[i - 1], pix_arr[j - 1]);
      }
      pix_arr[l - 1] = pix_arr[j - 1];
      pix_arr[j - 1] = a;
      j_stack += 2;
      if (j_stack > PIX_STACK_SIZE) {
         i_stack = realloc(i_stack, (j_stack + PIX_STACK_SIZE) * sizeof(double));  /* changed by Alexander Walz */
         if (i_stack == NULL) return 0;  /* changed by Alexander Walz */
      }
      if (ir-i+1 >= j - l) {
        i_stack[j_stack - 1] = ir;
        i_stack[j_stack - 2] = i;
        ir = j- 1;
      } else {
        i_stack[j_stack - 1] = j - 1;
        i_stack[j_stack - 2] = l;
        l = i;
      }
    }
  }
  xfree(i_stack);
  return 1;
}


/*---------------------------------------------------------------------------

   Algorithm from N. Wirth's book, implementation by N. Devillard.
   This code in public domain.

   Function :   kth_smallest()
   In       :   array of elements, # of elements in the array, rank k
   Out      :   one element
   Job      :   find the kth smallest element in the array
   [Notice  :   use the median() macro defined below to get the median.]
   Notice   :   works on both sorted as well as unsorted data

                Reference:

                  Author: Wirth, Niklaus
                   Title: Algorithms + data structures = programs
               Publisher: Englewood Cliffs: Prentice-Hall, 1976
    Physical description: 366 p.
                  Series: Prentice-Hall Series in Automatic Computation

 ---------------------------------------------------------------------------*/

/* #define ELEM_SWAP(a,b) { register double t = (a); (a) = (b); (b) = t; } */

LUALIB_API double tools_kth_smallest (double a[], long int n, long int k) {
  register long int i, j, l, m;
  register double x, temp;
  if (n == 0) return AGN_NAN;  /* 2.4.4 */
  l = 0; m = n - 1;
  while (l < m) {
    x = a[k];
    i = l;
    j = m;
    do {
      while (a[i] < x) i++;
      while (x < a[j]) j--;
      if (i <= j) {
        SWAP(a[i], a[j], temp);
        i++; j--;
      }
    } while (i <= j);
    if (j < k) l = i;
    if (k < i) m = j;
  }
  return a[k];
}


/* Dr. F.H.Toor's database C functions */

LUALIB_API int my_open (const char *file) {
  int hnd;
  hnd = open(file, O_BINARY | O_RDWR);
  if (hnd < 3)
    return -1;
  else
    return(hnd);
}


LUALIB_API FILE *my_fopen (const char *file) {  /* 2.11.5, returns NULL on failure */
  FILE *hnd;
  hnd = fopen64(file, "rb+");  /* Open a file for update (both for input and output). The file must exist. */
  return hnd;
}


/* read-only open */
LUALIB_API int my_roopen (const char *file) {
  int hnd;
  hnd = open(file, O_BINARY | O_RDONLY); /*|_S_IWRITE); */
  return (hnd < 3) ? -1 : hnd;
}


LUALIB_API FILE *my_froopen (const char *file) {  /* 2.11.5, returns NULL on failure */
  FILE *hnd;
  hnd = fopen64(file, "rb");  /* Open file for read operations. The file must exist. */
  return hnd;
}


LUALIB_API int my_create (const char *file) {
  int hnd;
  hnd = open(file, O_BINARY | O_RDWR | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);  /* GCC 4.3.2 needs a 3rd arg */
  return (hnd < 3) ? -1 : hnd;
}


/* wb+: write/update: Create an empty file and open it for update (both for input and output). If a file with the
   same name already exists its contents are DISCARDED and the file is treated as a new empty file.
   Source: http://www.cplusplus.com/reference/cstdio/fopen */
LUALIB_API FILE *my_fcreate (const char *file) {  /* 2.11.5, returns NULL on failure */
  FILE *hnd;
  hnd = fopen64(file, "wb+");  /* GCC 4.3.2 needs a 3rd arg */
  return hnd;
}


LUALIB_API int my_close (int hnd) {
  return close(hnd);
}


LUALIB_API off64_t my_seek (int hnd, off64_t pos) {  /* 2.11.0 fix */
  return (lseek64(hnd, pos, SEEK_SET) == -1) ? -1 : 1;
}


LUALIB_API off64_t my_lof (int hnd) {  /* 2.11.0 fix */
  off64_t size;
  size = lseek64(hnd, 0, SEEK_END);
  if (size == -1)
    fprintf(stderr, "Agena IO subsystem: LOF error\n");
  return(size);
}


LUALIB_API void my_read (int hnd, void *data, size_t size) {
  ssize_t z;
  if ((z = read(hnd, data, size)) != (ssize_t)size) {
    fprintf(stderr, "Agena IO subsystem: read error\n");
  }
}


LUALIB_API size_t sec_read (int hnd, void *data, size_t size) {
  return (read(hnd, data, size) == (ssize_t)size);
}


LUALIB_API int32_t sec_readl (int hnd, ssize_t *success) {
  int32_t data;
  *success = (read(hnd, &data, sizeof(int32_t)) > 0);
  if (*success == 0) return -1;
#if BYTE_ORDER != BIG_ENDIAN
  tools_swapint32_t(&data);
#endif
  return(data);
}


LUALIB_API uint32_t sec_readul (int hnd, ssize_t *success) {
  uint32_t data;
  *success = (read(hnd, &data, sizeof(uint32_t)) > 0);
  if (*success == 0 || data < 0) return -1;
#if BYTE_ORDER != BIG_ENDIAN
  tools_swapuint32_t(&data);
#endif
  return(data);
}


/* longs are always stored in Big Endian notation */

LUALIB_API int32_t my_readl (int hnd) {
  int32_t data;
  my_read(hnd, &data, sizeof(int32_t));
#if BYTE_ORDER != BIG_ENDIAN
  tools_swapint32_t(&data);
#endif
  return(data);
}


LUALIB_API char my_readc (int hnd) {
  char data;
  my_read(hnd, &data, sizeof(char));
  return(data);
}


LUALIB_API void my_write (int hnd, void *data, size_t size) {
  if (write(hnd, data, size) != (ssize_t)size)
    fprintf(stderr, "Agena IO subsystem: write error\n");
}

#ifndef ENOFILE
#define ENOFILE 2
#endif

LUALIB_API const char *my_ioerror (int en) {  /* 2.6.1 */
  switch (en) {
    case EAGAIN:  return "write operation was not complete";
    case EBADF:   return "file descriptor is invalid or not open";
    case EFBIG:   return "size of the file becomes too large";
    case EINTR:   return "operation interrupted by a signal";
    case EIO:     return "hardware error encountered";
    case ENOSPC:  return "device is full";
    case EPIPE:   return "pipe or FIFO is not open or broken";
    case EINVAL:  return "position and size not aligned to block size";
    /* added 2.10.4 */
    case ENOFILE: return "file or directory does not exist";
    case ENXIO:   return "device or address does not exist";
    case ENOMEM:  return "not enough space available";
    case EACCES:  return "permission denied";
    case EEXIST:  return "file or directory already exists";
    case ENODEV:  return "device does not exist";
    case ENOTDIR: return "target is not a directory";
    case EISDIR:  return "target is a directory";
    case ENFILE:  return "there are too many open files in system";
    case EMFILE:  return "there are too many open files";
    case ENOTTY:  return "I/O control operation is inappropriate";
    case EROFS:   return "file system is read-only";
    case EMLINK:  return "there are too many links";
    case ENAMETOOLONG:  return "filename is too long";
    case ENOLCK:  return "locks are not available";
    case ENOTEMPTY:  return "directory is not empty";
#ifndef LUA_DOS
    case EILSEQ:  return "byte sequence is illegal";
    case ERANGE:  return "result is too large";
#endif
    case EDOM:    return "domain error";
    case ESPIPE:  return "seek is invalid (on a pipe ?)";
    case EXDEV:   return "link is improper";
    case EBUSY:   return "strerror reports \'Resource device\'";
    case EFAULT:  return "bad address";
    case EPERM:   return "operation is not permitted";
    case ESRCH:   return "process not found";
    case E2BIG:   return "argument list is too long";
    case ENOEXEC: return "exec format is invalid";
    case ECHILD:  return "child process does not exist";
    default: {  /* 2.10.4 improvement */
      int err;
      char firstletter[2];
      char *rest, *errmsg;
      errmsg = strerror(en);
      firstletter[0] = tools_lowercase[uchar(errmsg[0])];
      firstletter[1] = '\0';
      rest = substr(errmsg, 1, strlen(errmsg) - 1, &err);
      if (err == -1) return NULL;
      return concat(firstletter, rest, NULL);
    }
  }
}


/* longs are always stored in Big Endian notation */

LUALIB_API void my_writel (int hnd, int32_t data) {
#if BYTE_ORDER != BIG_ENDIAN
  tools_swapint32_t(&data);
#endif
  my_write(hnd, &data, sizeof(int32_t));
}


LUALIB_API void my_writec (int hnd, char data) {
  my_write(hnd, &data, sizeof(char));
}


/* in UNIX and Windows, if size == 0 then lock entire file */
LUALIB_API int my_lock (int hnd, off64_t start, off64_t size) {
  #ifdef _WIN32
  /* see: http://msdn.microsoft.com/en-us/library/8054ew2f.aspx, topic `_locking` */
  int r;
  off64_t oldpos, cursor;
  if (size == 0) size = (off64_t)pow(2, 63);  /* = lock 2^63 bytes (cannot be more in Windows) */
  /* reset file cursor to `start' */
  oldpos = my_fpos(hnd);
  if (oldpos == -1) return -1;
  cursor = lseek64(hnd, start, SEEK_SET);
  if (cursor == -1) return -1;
  /* lock the file */
  r = _locking(hnd, _LK_NBLCK, size);
  /* set file cursor to original file position */
  cursor = lseek64(hnd, oldpos, SEEK_SET);
  if (cursor == -1 || r != 0) return -1;
  #elif defined(__unix__) || defined(__APPLE__)
  struct flock file;
  file.l_type = F_WRLCK | F_RDLCK;   /* exclusive read and write lock */
  file.l_start = (off_t)start;   /* locking from file position `start` */
  file.l_whence = SEEK_SET;      /* offset is relative to the beginning of the file */
  file.l_len = (size == 0) ? 0L : (off_t)size;  /* lock size bytes (size == 0 query: better sure than sorry) */
  if (fcntl(hnd, F_SETLK, &file) == -1)
    return -1;
  #endif
  return 0;
}


LUALIB_API int my_unlock (int hnd, off64_t start, off64_t size) {
  #ifdef _WIN32
  /* see: http://msdn.microsoft.com/en-us/library/8054ew2f.aspx, topic `_locking` */
  int r;
  off64_t oldpos, cursor;
  if (size == 0) size = (off64_t)pow(2, 63);  /* = lock 2^63 bytes (cannot be more in Windows) */
  /* reset file cursor to `start' */
  oldpos = my_fpos(hnd);
  if (oldpos == -1) return -1;
  cursor = lseek64(hnd, start, SEEK_SET);
  if (cursor == -1) return -1;
  /* lock the file */
  r = _locking(hnd, LK_UNLCK, size);
  /* set file cursor to original file position */
  cursor = lseek64(hnd, oldpos, SEEK_SET);
  if (cursor == -1 || r != 0) return -1;
  #elif defined(__unix__) || defined(__APPLE__)
  struct flock file;
  file.l_type = F_WRLCK | F_RDLCK;
  file.l_whence = SEEK_SET;
  file.l_start = (off_t)start;
  file.l_len = (size == 0) ? 0L : (off_t)size;  /* unlock size bytes (size == 0 query: better sure than sorry) */
  if (fcntl(hnd, F_UNLCK, &file) == -1)
    return -1;
  #endif
  return 0;
}


/* my_move(hnd, low*4L+24L, low*4L+28L, cnt*4L+24L); */
/* fpos: from position, tpos: to position, size */
LUALIB_API void my_move (int hnd, off64_t fpos, off64_t tpos, off64_t size) {  /* 2.11.0 fix */
  off64_t segment;
  char buff[BASE_MOVELEN];
  segment = size - fpos;
  if (segment < 1)
    return;
  if (fpos < 0 || tpos < 0)
    fprintf(stderr, "Agena IO subsystem: move error\n");
  if (fpos > tpos) {
    while (segment >= BASE_MOVELEN) {
      my_seek(hnd, fpos);
      my_read(hnd, buff, BASE_MOVELEN);
      my_seek(hnd, tpos);
      my_write(hnd, buff, BASE_MOVELEN);
      fpos += BASE_MOVELEN;
      tpos += BASE_MOVELEN;
      segment -= BASE_MOVELEN;
    }
    if (segment > 0) {
      my_seek(hnd, fpos);
      my_read(hnd, buff, segment);
      my_seek(hnd, tpos);
      my_write(hnd, buff, segment);
    }
  }
  if (fpos < tpos) {
    while (segment >= BASE_MOVELEN) {
      my_seek(hnd, fpos + segment - BASE_MOVELEN);
      my_read(hnd, buff, BASE_MOVELEN);
      my_seek(hnd, tpos + segment - BASE_MOVELEN);
      my_write(hnd, buff, BASE_MOVELEN);
      segment -= BASE_MOVELEN;
    }
    if (segment > 0) {
      my_seek(hnd, fpos);
      my_read(hnd, buff, segment);
      my_seek(hnd, tpos);
      my_write(hnd, buff, segment);
    }
  }
}


/* return current file position, written by Alexander Walz on June 28, 2007; extended
   0.32.5 */

LUALIB_API off64_t my_fpos (int hnd) {
  return lseek64(hnd, 0, SEEK_CUR);
}


/* my_expand: core functionality for expanding a database by a) reshifting all records, thus
   b) adding more index entries.
   Assumes that the file is open and locked when envoked.
   Restores the file cursor to the position before it was called.

   written by Alexander Walz on June 28, 2007

   Arguments:
      hnd   - the file handler
      mrc   - maximum number of records currently allowed
      cnt   - current number of actual records
      count - number of records to be added */

LUALIB_API void my_expand (int hnd, int mrc, int cnt, int count, int *error) {
  int32_t dsbegin, dsend, offset, j, index, bufsize, cfpos, commentpos, commentlen, newindex;
  int i;
  char *buffer;
  cfpos = my_fpos(hnd);           /* save the current file position for later restoration */
  my_seek(hnd, BASECOMMENT);
  commentpos = my_readl(hnd);
  commentlen = 0;
  if (commentpos != 0) {
    my_seek(hnd, commentpos);
    commentlen = my_readl(hnd) + 4L;
  }
  dsbegin = mrc*4L + BASE_OFFSET;   /* start of current dataset section */
  dsend = my_lof(hnd) - 1;          /* end of current dataset section */
  /* write 0L at end of file count times */
  my_seek(hnd, dsend + 1);
  for (i=0; i < count; i++) {
    my_writel(hnd, 0L);          /* write zeros */
  }
  offset = count * 4L;
  buffer = malloc(offset*sizeof(char));
  if (buffer == NULL) {
    *error = 1;
    return;
  }
  /* beginning with the last data set move data sets to new eof */
  bufsize = offset;
  j = dsend + 1;
  while (j > dsbegin) {
    if ((j - dsbegin) >= offset) {
      bufsize = offset;
      my_seek(hnd, j - bufsize);
      my_read(hnd, buffer, bufsize);
      my_seek(hnd, j);
      my_write(hnd, buffer, bufsize);
    } else {  /* last few bytes need special treatment */
      my_seek(hnd, dsbegin);
      my_read(hnd, buffer, j - dsbegin);
      my_seek(hnd, j + bufsize - (j - dsbegin));
      my_write(hnd, buffer, j - dsbegin);
    }
    j -= bufsize;
  }
  /* zero start of old data section, 2.11.0 RC2 fix */
  my_seek(hnd, dsbegin);
  for (i=0; i < count; i++)
    my_writel(hnd, 0L);
  /* update indices */
  for (j=0; j < cnt; j++) {
    my_seek(hnd, j*4L + BASE_OFFSET);
    index = my_readl(hnd);
    my_seek(hnd, j*4L + BASE_OFFSET);
    newindex = (index > commentpos) ? index + offset + commentlen : index + offset;
    my_writel(hnd, newindex);
  }
  /* add new indices to end of index section */
  for (j=0; j < count; j++) {
    my_writel(hnd, 0L);
  }
  /* update maximum number of entries field */
  my_seek(hnd, MAXNRECORDS);
  my_writel(hnd, mrc + count);
  /* update position of comment if present */
  if (commentpos != 0) {
    my_seek(hnd, BASECOMMENT);
    my_writel(hnd, commentpos + offset);
  }
  my_seek(hnd, cfpos);  /* reset cursor to position before the function was called */
  xfree(buffer);
  *error = 0;
}


#if (defined(__unix__) && !defined(LUA_DOS)) || defined(__APPLE__)

static struct termios new_io, old_io;

/* Taken from `C von A bis Z` by Jrgen Wolf, 2nd Ed., pp. 452 */
/* Funktion schaltet das Terminal in den cbreak-Modus:         */
/* Kontrollflag ECHO und ICANON auf 0 setzen                   */
/* Steuerzeichen: Leseoperation liefert 1 Byte VMIN=1 VTIME=1  */
LUALIB_API int cbreak (int fd) {
  /*Sichern unseres Terminals*/
  if ((tcgetattr(fd, &old_io)) == -1)
    return -1;
  new_io = old_io;
  /* Wir verndern jetzt die Flags fr den cbreak-Modus */
  new_io.c_lflag &= ~(ECHO | ICANON);
  new_io.c_cc[VMIN] = 1;
  new_io.c_cc[VTIME]= 0;
  /* Jetzt setzen wir den cbreak-Modus */
  if ((tcsetattr(fd, TCSAFLUSH, &new_io)) == -1)
    return -1;
  return 1;
}


LUALIB_API int getch (void) {
  int c;
  if (cbreak(STDIN_FILENO) == -1) {
    fprintf(stderr, "Agena IO subsystem: error in function cbreak (called by getch).\n");
    exit(EXIT_FAILURE);
  }
  c = getchar();
  /* Alten Terminal-Modus wiederherstellen */
  tcsetattr(STDIN_FILENO, TCSANOW, &old_io);
  return c;
}

#endif

/* Copyright (c) 2008    Johnathon

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.

Source: http://code.google.com/p/linux-kbhit/linux_kbhit.tar.gz

*/

#if (defined(__unix__) && !defined(LUA_DOS))  /* 2.3.4, use DJGPP's built-in kbhit function instead to avoid stdin corruption */

LUALIB_API int kbhit (void) {  /* does not work on Mac */
  struct termios oldt, newt;
  int ch;
  if (tcgetattr(STDIN_FILENO, &oldt) == -1) return -1;
  newt = oldt;
  newt.c_cc[VMIN]  = 0;
  newt.c_cc[VTIME] = 1;
  newt.c_lflag &= ~(ICANON | ECHO);
  if (tcsetattr(STDIN_FILENO, TCSANOW, &newt) == -1) return -1;
  ch = getchar();  /* ch == -1 means key not pressed */
  if (tcsetattr(STDIN_FILENO, TCSANOW, &oldt) == -1) return -1;
  return (ch != -1);
}

#endif


#if defined(__APPLE__)
#include <sys/select.h>

/* found on: http://stackoverflow.com/questions/448944/c-non-blocking-keyboard-input, originally written by Alnitak;
   XXX this implementation most of the times still echoes the key being pressed at stdin */

LUALIB_API int kbhit (void) {
  int r, i;
  fd_set fds;
  struct termios newt, oldt;
  struct timeval tv = {0L, 0L};
  if (tcgetattr(STDIN_FILENO, &oldt) == -1) return -1;  /* save current terminal mode */
  newt = oldt;
  /* cfmakeraw(&newt); */  /* see the GNU C manual for the more info on this function */
  newt.c_lflag &= ~(ICANON | ECHO);
  newt.c_cc[VMIN]  = 1;  /* default: 1 */
  newt.c_cc[VTIME] = 0;  /* default: 0 */
  if (tcsetattr(STDIN_FILENO, TCSANOW, &newt) == -1) return -1;  /* change terminal mode */
  FD_ZERO(&fds);
  FD_SET(STDIN_FILENO, &fds);
  r = select(FD_SETSIZE, &fds, NULL, NULL, &tv);  /* listen for signals */
  i = tcsetattr(STDIN_FILENO, TCSANOW, &oldt);  /* reset terminal mode */
  FD_CLR(STDIN_FILENO, &fds);
  return (i == -1) ? -1 : r;  /* -1: an error has occurred, 0: no input, 1: key pressed */
}

#endif


#ifdef __OS2__  /* 2.3.4 */

LUALIB_API int kbhit (void) {
  KBDKEYINFO data;
  KbdPeek(&data, 0);
  return (data.fbStatus & 0xe0) != 0;
}

#endif

/* stores the current working directory in the already allocated buffer variable buffer,
   and returns 0 on success or -1 otherwise. In case of failure, the function frees buffer,
   so you do not have to free it in the function calling tools_cwd. */
LUALIB_API int tools_cwd (char *buffer) {  /* Agena 1.0.4, Agena 1.5.0, Agena 1.6.0 Valgrind */
  if (buffer == NULL) return -1;
  if (getcwd(buffer, PATH_MAX) == buffer) return 0;
  xfree(buffer);
  return -1;
}


#define luai_numfrac(a)     ((a) - trunc((a)))
#define NANO_SECOND_MULTIPLIER  1000000000L  /* 1 millisecond = 1,000,000 Nanoseconds */
#define MICRO_SECOND_MULTIPLIER 1000000L     /* 1 microsecond */
LUALIB_API int tools_wait (double x) {  /* 2.4.0 */
  int r;
  if (x < 0) return 0;
#ifdef _WIN32
  Sleep(x*1000);  /* microsecs */
  r = 1;
#elif defined (LUA_DOS) || (defined(__SOLARIS))
  r = usleep(x*MICRO_SECOND_MULTIPLIER) == 0;
#elif defined(__unix__) || defined(__OS2__) || defined(__APPLE__)
  struct timespec tv;
  tv.tv_sec  = (time_t)x;
  tv.tv_nsec = (long int)(luai_numfrac(x) * (lua_Number)NANO_SECOND_MULTIPLIER);
  r = nanosleep(&tv, NULL) == 0;  /* none-busy wait */
#elif defined(__HAIKU__)
  if (tools_isfloat(x)) x = (int)(trunc(x) + 1);
  sleep(x);
  r = 1;
#else
  r = 0;
#endif
  return r;
}


static int ndays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

LUALIB_API int tools_checkdatetime (int year, int month, int day, int hour, int minute, int second, int mseconds) {  /* changed 2.10.0 */
  if (month < 1 || month > 12 || day < 1 ||
    hour < 0 || hour > 24 || minute < 0 || minute > 59 || second < 0 || second > 59 || mseconds < 0 || mseconds > 999)
    return 0;
  else if (hour == 24 && (minute != 0 || second != 0))  /* not midnight ? */
    return 0;
  /* now check for February 29 */
  return day <= ndays[month - 1] + (month == 2) * tools_isleapyear(year);
}


LUALIB_API Time64_T tools_maketime (int year, int month, int day, int hour, int minute, int second, double *alternative) {  /* Agena 1.8.0, added 2.9.8 */
  struct TM time_str;
  Time64_T r;
  *alternative = 0;
  if (tools_checkdatetime(year, month, day, hour, minute, second, 0) == 0) return -2;  /* 1.9.1 */
  time_str.tm_year = year - 1900;
  time_str.tm_mon = month - 1;
  time_str.tm_mday = day;
  time_str.tm_hour = hour;
  time_str.tm_min = minute;
  time_str.tm_sec = second;
  time_str.tm_isdst = -1;
  r = mktime64(&time_str);
  if (r == -1 && (year < 1970 || (year == 1970 && month == 1 && day == 1 && hour == 0))) {  /* 2.16.2 fix for MinGW mktime bug for dates y 1970/1/1 */
    *alternative = iauJuliandate(year, month, day, hour, minute, second) - 2415020.5 + 1 + ((year == 1900 && month > 2) || year > 1900);
    return -1;  /* 2415020.5 = 1900/1/1 = LSD Day 1; alternative is HUGE_VAL if undeterminable. */
  }  /* 2.16.1 patch for buggy GCC mktime[64] */
  return r;
}


LUALIB_API double tools_time (void) {  /* 2.12.4, patched 2.12.6, changed 2.14.0 */
  time_t t;
  int32_t msecs;  /* 2.16.5 patch for Mac OS X */
  double r;
#if defined(_WIN32) || defined(__unix__) || defined(__APPLE__) || defined(__OS2__)
  struct timeb tp;
#endif
  t = (time_t)(-1);
  r = 1;
  msecs = 0;
#if defined(_WIN32) || defined(__OS2__)
  ftime(&tp);  /* UTC, 2.12.6 patch */
  t = tp.time;
  msecs = tp.millitm;
#elif defined(__unix__) || defined(__APPLE__)
  if (ftime(&tp) == 0) { /* set milliseconds only when query has been successful */
    t = tp.time;
    msecs = tp.millitm;
  }
#endif
  if (t == (time_t)(-1))
    r = -1;
  else {
    r = (double)t;
    if (msecs != -1) r += 0.001*msecs;
  }
  return r;
}


LUALIB_API INLINE int tools_cpuid (uint32_t call, uint32_t *eax, uint32_t *ebx, uint32_t *ecx, uint32_t *edx) {  /* 2.14.3 */
#ifdef CPUID
  uint32_t ax, bx, cx, dx;
  __cpuid(0, ax, bx, cx, dx);
  if (call <= ax) {
    __cpuid(call, ax, bx, cx, dx);
    *eax = ax; *ebx = bx; *ecx = cx; *edx = dx;
    return 0;
  } else
#endif
  return -1;
}


/* Determination of endianness; taken from Jrgen Wolf, `C von A bis Z`, 2. Auflage, Galileo Computing */

LUALIB_API char tools_endian (void) {
  unsigned int Word = 0x22CCDDEE;
  unsigned char *Byte;
  Byte = (unsigned char *) &Word;
  if (Byte[0] == ((Word >> 0) & 0xFF))
    return 0;  /* little endian */
  else if (Byte[0] == ((Word >> 24) & 0xFF))
    return 1;  /* big endian */
  else
    return -1;
}


/* Returns the contents of the given environment variable. To avoid the value from being overwritten in subsequent calls,
   it is duplicated, so FREE it !!! 2.16.11 */
LUALIB_API char *tools_getenv (const char *name) {
/* tested in Windows, Mac OS X, Solaris 10, Linux, OS/2, DOS */
#if defined(_WIN32)
  /* 2.8.6, taken from lua_sys package written by Nodir Temirkhodjaev, <nodir.temir@gmail.com> */
  unsigned int len;
  len = GetEnvironmentVariableA(name, NULL, 0);  /* get number of characters to be stored in buffer, _including_ terminating \0 */
  if (len) {
    char *buf = tools_stralloc(len - 1);  /* 2.16.12 change */
    if (buf && GetEnvironmentVariableA(name, buf, len) == len - 1) return buf;  /* now GetEnvironmentVariableA returns #chars excluding \0 */
  }
  return NULL;
#elif defined(__OS2__)     /* 2.16.11, see: http://www.edm2.com/index.php/DosScanEnv */
  PSZ r = NULL;            /* stores environment variable */
  APIRET rc = NO_ERROR;    /* return code */
  rc = DosScanEnv((PCSZ)name, &r);  /* Get contents of PATH environment */
  return (rc != NO_ERROR) ? NULL : strdup((char *)r);
#elif defined(LUA_DOS)
  char *result = getenv(name);
  return (result) ? strdup(result) : NULL;  /* 2.17.0a fix to prevent segfaults with more recent versions of DJGPP if strdup is called with NULL */
#elif defined(__unix__) || defined(__APPLE__)
  /* char *result = getenv(name); */  /* getenv will not work at least in Linux, Solaris */
  const char *nam, *val, *end;  /* taken from os_environ in loslib.c */
  const char **env;
  for (env = (const char **)environ; (nam = *env); env++) {
    end = strchr(val = strchr(nam, '=') + 1, '\0');
    if (strncmp(name, nam, val - nam - 1) == 0) {
      char *result = tools_stralloc(end - val);
      if (result == NULL) return NULL;
      memcpy(result, val, end - val);
      return result;
    }
  }
  return NULL;
#else
  return NULL;
#endif
}


LUALIB_API double tools_intpow (double x, int n) {
  double r;
  if (n == INT_MIN) return AGN_NAN;  /* undefined or +/- infinity ? -> protect against an infinite loop;
    n is an int, not a double, so isinf/isnan do not work here. */
  r = 1.0;
  if (n <= 0) {  /* 2.5.1 fix */
    if (x == 0) return AGN_NAN;  /* 0.32.4 patch */
    x = 1.0 / x;
    n = -n;
  }
  while (1) {
    if (n & 1) r *= x;  /* power is odd ? */
    n >>= 1;
    if (n == 0) return r;  /* 2.4.1 change */
    x *= x;
  };
  /* never reached */
}


/* OBSOLETE, is number finite ?  Use C's isfinite instead */
LUALIB_API int tools_isfinite (const double a) {
  const double b = a - a;
  int r = (b == b);
  return r;
}


/* OBSOLETE, is number NaN ?  Use C's isnan instead */
LUALIB_API int tools_isnan (const double x) {
  int r = (x != x);
  return r;
}


/* define NaN */
LUALIB_API double tools_nan (void) {
  return (0.0/0.0);
}


/* code taken from http://www.dmh2000.com/cpp/dswap.shtml */

#if BYTE_ORDER != BIG_ENDIAN

LUALIB_API double tools_tobigendian (double d) {  /* 2.14.9, swap bytes */
  double_cast u;
  uint64_t a;
  size_t i, s;
  unsigned char *src = (unsigned char *)&d;
  unsigned char *dst = (unsigned char *)&a;
  s = sizeof(double);
  for (i=0; i < s; i++) dst[i] = src[s - i - 1];  /* 2.9.0 */
  u.i = a;
  return u.f;
}

LUALIB_API uint64_t tools_twoint32touint64 (int32_t d, int32_t e) {  /* Little Endian, 2.9.8, used by xbase.writetime */
  uint64_t a;
  size_t i, s;
  unsigned char *dst = (unsigned char *)&a;
  unsigned char *src = (unsigned char *)&d;
  s = sizeof(int32_t);
  for (i=0; i < s; i++) dst[i] = src[i];
  src = (unsigned char *)&e;
  for (; i < 2*s; i++) dst[i] = src[i - s];
  return a;
}

/* 2.9.8, k=1: first int32_t, k=2: second one; used by xbase.DBFReadTimeStampAttribute, Little Endian */
LUALIB_API int32_t tools_uint64toint32 (uint64_t d, char k) {
  int32_t a;
  size_t i, s, q;
  unsigned char *dst = (unsigned char *)&a;
  unsigned char *src = (unsigned char *)&d;
  s = sizeof(int32_t);
  q = s*(k - 1);
  for (i=q; i < k*s; i++) dst[i - q] = src[i];
  return a;
}

/* Little Endian, signed 4-byte integer, 2.9.8, used by xbase.DBFWriteAttribute (binary I type) */
LUALIB_API int32_t tools_swapint32 (int32_t d) {
  int32_t a;
  size_t i, s;
  unsigned char *dst = (unsigned char *)&a;
  unsigned char *src = (unsigned char *)&d;
  s = sizeof(int32_t);
  for (i=0; i < s; i++) dst[i] = src[i];
  return a;
}

LUALIB_API double tools_sint2double (int32_t a) {  /* Little Endian, 2.9.8, used by xbase.DBFReadAttribute */
  int32_t d;
  size_t i, s;
  unsigned char *src = (unsigned char *)&a;
  unsigned char *dst = (unsigned char *)&d;
  s = sizeof(int32_t);
  for (i=0; i < s; i++) dst[i] = src[i];
  return (double)d;
}
#else
/* BIG ENDIAN */
LUALIB_API double tools_tolittleendian (double d) {  /* 2.14.9, swap bytes */
  double_cast u;
  uint64_t a;
  size_t i, s;
  unsigned char *src = (unsigned char *)&d;
  unsigned char *dst = (unsigned char *)&a;
  s = sizeof(double);
  for (i=0; i < s; i++) dst[i] = src[s - i - 1];  /* 2.9.0 */
  u.i = a;
  return u.f;
}

LUALIB_API double tools_tobigendian (double d) {  /* 2.14.9, swap bytes */
  return d;
}

LUALIB_API uint64_t tools_swapdouble (double d) {  /* Big Endian, used by binio.writenumber */
  uint64_t a;
  size_t i, s;
  unsigned char *dst = (unsigned char *)&a;
  unsigned char *src = (unsigned char *)&d;
  s = sizeof(double);
  for (i=0; i < s; i++) dst[i] = src[i];  /* 2.9.0, do nothing */
  return a;
}

/* unswap using char pointers */
LUALIB_API double tools_unswapdouble (uint64_t a) {  /* Big Endian, used by binio.readnumber, binio.readobjectaux */
  double d;
  size_t i, s;
  unsigned char *src = (unsigned char *)&a;
  unsigned char *dst = (unsigned char *)&d;
  s = sizeof(double);
  for (i=0; i < s; i++) dst[i] = src[i];  /* 2.9.0 */
  return d;
}

LUALIB_API int32_t tools_swapint32 (int32_t a) {  /* Big Endian, 2.9.8 */
  int32_t d;
  size_t i, s;
  unsigned char *dst = (unsigned char *)&d;
  unsigned char *src = (unsigned char *)&a;
  s = sizeof(int32_t);
  for (i=0; i < s; i++) dst[i] = src[s - i - 1];
  return d;
}

LUALIB_API uint64_t tools_doubletouint64andswap (double a) {  /* Big Endian, used by xbase.DBFWriteAttribute (binary B type) */
  uint64_t d;
  size_t i, s;
  unsigned char *dst = (unsigned char *)&d;
  unsigned char *src = (unsigned char *)&a;
  s = sizeof(double);
  for (i=0; i < s; i++) dst[i] = src[s - i - 1];  /* 2.9.0 */
  return d;
}


LUALIB_API uint64_t tools_twoint32touint64 (int32_t d, int32_t e) {  /* Big Endian, 2.9.8, used by xbase.writetime */
  uint64_t a;
  size_t i, s;
  unsigned char *dst = (unsigned char *)&a;
  unsigned char *src = (unsigned char *)&d;
  s = sizeof(int32_t);
  for (i=0; i < s; i++) dst[i] = src[s - i - 1];
  src = (unsigned char *)&e;
  for (; i < 2*s; i++) dst[i] = src[2*s - i - 1];
  return a;
}

/* Big Endian, 2.9.8, k=1: first int32_t, k=2, second one, used by xbase.DBFReadTimeStampAttribute */
LUALIB_API int32_t tools_uint64toint32 (uint64_t d, char k) {
  int32_t a;
  size_t i, s, q;
  unsigned char *dst = (unsigned char *)&a;
  unsigned char *src = (unsigned char *)&d;
  s = sizeof(int32_t);
  q = s*(k - 1);
  /* k=1 -> q = 0; k=2 -> q = s */
  /* q  i  s*k - i - 1 + q
     4  4  7
     4  5  6
     4  6  5
     4  7  4 */
  for (i=q; i < k*s; i++) dst[i - q] = src[s*k - i - 1 + q];
  return a;
}


LUALIB_API double tools_uint64todoubleandswap (uint64_t a) {  /* Big Endian, used by xbase.DBFReadAttribute */
  double d;
  size_t i, s;
  unsigned char *src = (unsigned char *)&a;
  unsigned char *dst = (unsigned char *)&d;
  s = sizeof(double);
  for (i=0; i < s; i++) dst[i] = src[s - i - 1];  /* 2.10.0 */
  return d;
}


LUALIB_API double tools_sint2double (int32_t a) {  /* Big Endian, 2.9.8, used by xbase.DBFReadAttribute */
  int32_t d;
  size_t i, s;
  unsigned char *src = (unsigned char *)&a;
  unsigned char *dst = (unsigned char *)&d;
  s = sizeof(int32_t);
  for (i=0; i < s; i++) dst[i] = src[s - i - 1];  /* 2.9.0 */
  return (double)d;
}
#endif


LUALIB_API void tools_swapint32_t (int32_t *n) {
  size_t i;
#ifndef __HAIKU__
  size_t s = sizeof(int32_t);
#endif
  union {
    int32_t d;
#ifdef __HAIKU__
    unsigned char b[4];
#else
    unsigned char b[s];
#endif
  } p, q;
  p.d = *n;
#ifndef __HAIKU__
  for (i=0; i < s; i++) q.b[i] = p.b[s - i - 1];  /* 2.9.0 */
#else
  for (i=0; i < 4; i++) q.b[i] = p.b[4 - i - 1];  /* 2.9.4 */
#endif
  *n = q.d;
}


LUALIB_API void tools_swapuint32_t (uint32_t *n) {
  size_t i;
#ifndef __HAIKU__
  size_t s = sizeof(uint32_t);
#endif
  union {
    uint32_t d;
#ifdef __HAIKU__
    unsigned char b[4];
#else
    unsigned char b[s];
#endif
  } p, q;
  p.d = *n;
#ifndef __HAIKU__
  for (i=0; i < s; i++) q.b[i] = p.b[s - i - 1];  /* 2.9.0 */
#else
  for (i=0; i < 4; i++) q.b[i] = p.b[4 - i - 1];  /* 2.9.4 */
#endif
  *n = q.d;
}


/* Determines in which half-plane a!b lies, see `csgn` online help in Maple V Release 4 */
LUALIB_API double tools_csgn (double a, double b) {
  if (a > 0 || (a == 0 && b > 0))
    return 1.0;
  else if (a < 0 || (a == 0 && b < 0))
    return -1.0;
  else
    return 0.0;
}


LUALIB_API char *charreplace (char *s, char from, char to, int flag) {
  char *olds, *e;
  olds = s;
  while (NULL != (e = strchr(s, from))) {  /* traverses the entire string */
    s = e;
    *s = to;
  }
  s = s + strlen(s) - 1;  /* set pointer to end of string */
  /* delete trailing slash if present, only if str has more than one char, 0.31.4 */
  if (flag && strlen(olds) > 1 && (*s == '/' || *s == '\\')) *s = '\0';
  return olds;
}


/*
 * concat() - allocate memory and safely concatenate strings in portable C
 * (and C++ if you like).
 *
 * This code deals gracefully with potential integer overflows (perhaps when
 * input strings are maliciously long), as well as with input strings changing
 * from under it (perhaps because of misbehavior of another thread).  It does
 * not depend on non-portable functions such as snprintf() and asprintf().
 *
 * Written by Solar Designer <solar at openwall.com> and placed in the
 * public domain.
 *
 * Originally written for and currently maintained as a part of popa3d,
 * a POP3 server:
 *
 *   http://www.openwall.com/popa3d/
 *   file misc.c
 */

/* Concatenates strings and returns a null-terminated string, or NULL if something failed.

   Always pass NULL as the last argument to denote the end of the strings to be concatenated, e.g.
   concat(argv[0], argv[1], argv[2], argv[3], NULL).

   You must FREE the memory allocated by concat in case of success. ONLY assign the result to a variable,
   use it in the VM or API, and then finally FREE it !

   Please note that pushing strings onto the stack and then calling `lua_concat` may be 30 % faster in
   standard situations. */

LUALIB_API char *concat (const char *s1, ...) {
  va_list args;
  const char *s;
  char *p, *result;
  unsigned long l, m, n;
  m = n = strlen(s1);
  va_start(args, s1);
  while ((s = va_arg(args, char *))) {
    l = strlen(s);
    if ((m += l) < l) break;
  }
  va_end(args);
  if (s || m >= INT_MAX) return NULL;
  result = tools_stralloc(m);  /* 2.16.5, (char *)malloc((m + 1)*sizeof(char)); */
  if (!result) return NULL;
  memcpy(p = result, s1, n);
  p += n;
  va_start(args, s1);
  while ((s = va_arg(args, char *))) {
    l = strlen(s);
    if ((n += l) < l || n > m) break;
    memcpy(p, s, l);
    p += l;
  }
  va_end(args);
  if (s || m != n || p != result + n) {
    xfree(result);
    return NULL;
  }
  *p = 0;
  return result;
}


/* Taken from: http://c-faq.com/lib/regex.html, maintained by Steve Summit

   Quick little wildcard matcher by Arjan Kenter (Copyright 1995, Arjan Kenter). Processes ? and *
   wildcards and return 0 (false) or 1 (true).

   With this definition, the call match("a*b.c", "aplomb.c") would return 1. */

LUALIB_API int glob (const char *pat, const char *str) {
  switch(*pat) {
    case '\0':  return !*str;
    case '*':   return glob(pat+1, str) ||
                  (*str && glob(pat, str+1));
    case '?':   return *str && glob(pat+1, str+1);
    default:    return *pat == *str && glob(pat+1, str+1);
  }
}


/* based on oa_substr released with the sources of OpenAxiom 1.4.1, see file open-axiom-1.4.1/src/lib/cfuns-c.c;
   modified by Alexander Walz, 1.10.0.

   Returns a null-erminated substring. argument begin must be in the range 0 .. strlen()-1. error will be set to 1
   in case of an index-out-of-range error, to -1 in case of a failed memory allocation, and 0 in case of success.

   FREE the return after use !

   A light edition of this function without the error parameter and index out-of-range checks is not faster, although
   this may sound strangely. */

LUALIB_API char *substr (const char *str, const size_t begin, const size_t end, int *error) {
  char *substring;
  size_t len, s;    /* modified by Alex Walz */
  s = strlen(str);  /* modified by Alex Walz */
  if (str == NULL || s == 0 || s < begin || end >= s || begin > end || begin < 0 || end < 0) {
    *error = 1;     /* index out-of-range; modified by Alex Walz */
    return NULL;
  }
  len = (end - begin) + 2;  /* including terminating \0 */
  substring = (char*)malloc(len * sizeof(char));
  if (substring == NULL) {  /* memory allocation failed ?  Modified by Alex Walz */
    *error = -1;
    return NULL;
  }
  memset(substring, '\0', len);
  memcpy(substring, str + begin, len - 1);
  *error = 0;
  return substring;
}


/* ISO 8859/1 Latin-1 alphabetic and upper and lower case bit vector tables.

   Taken from the entropy utility ENT written by John Walker, January 28th, 2008,
   Fourmilab, http://www.fourmilab.ch.

   This software is in the public domain. Permission to use, copy, modify, and distribute this software
   and its documentation for any purpose and without fee is hereby granted, without any conditions or
   restrictions. This software is provided as is without express or implied warranty. */

/* !!! NEEDED by isISO* macros in agnhlps.h !!! */

unsigned char isoalpha[32] = {
    0,   0,   0,   0,   0,   0,   0,   0,
  127, 255, 255, 224, 127, 255, 255, 224,
    0,   0,   0,   0,   0,   0,   0,   0,
  255, 255, 254, 255, 255, 255, 254, 255
};

unsigned char isoupper[32] = {
    0,   0,   0,   0,   0,   0,   0,   0,
  127, 255, 255, 224,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,
  255, 255, 254, 254,   0,   0,   0,   0
};

unsigned char isolower[32] = {
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0, 127, 255, 255, 224,
    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   1, 255, 255, 254, 255
};


/* taken from: http://stackoverflow.com/questions/11258019/conversion-from-iso-8859-15-latin9-to-utf-8
   ISO 8859-15 is ISO8859/1 plus the EUR symbol;
   written by Nominal Animal, http://stackoverflow.com/users/1475978/nominal-animal.

   Creates a dynamically allocated copy of string, changing the encoding from ISO-8859-15 to UTF-8. */

LUALIB_API char *latin9_to_utf8 (const char *string) {  /* FREE IT ! */
  char *result;
  size_t  n = 0;
  if (string) {
    const unsigned char *s = (const unsigned char *)string;
    while (*s) {
      if (*s < 128) {
        s++;
        n += 1;
      } else
      if (*s == 164) {
        s++;
        n += 3;
      } else {
        s++;
        n += 2;
      }
    }
  }
  /* Allocate n+1 (to n+7) bytes for the converted string. */
  result = malloc(sizeof(char)*((n | 7) + 1));
  if (!result) return NULL;
  /* Clear the tail of the string, setting the trailing NUL. */
  memset(result + (n | 7) - 7, 0, 8);
  if (n) {
    const unsigned char *s = (const unsigned char *)string;
    unsigned char *d = (unsigned char *)result;
    while (*s) {
      if (*s < 128) {
        *(d++) = *(s++);
      } else
      if (*s < 192) switch (*s) {
        case 164: *(d++) = 226; *(d++) = 130; *(d++) = 172; s++; break;
        case 166: *(d++) = 197; *(d++) = 160; s++; break;
        case 168: *(d++) = 197; *(d++) = 161; s++; break;
        case 180: *(d++) = 197; *(d++) = 189; s++; break;
        case 184: *(d++) = 197; *(d++) = 190; s++; break;
        case 188: *(d++) = 197; *(d++) = 146; s++; break;
        case 189: *(d++) = 197; *(d++) = 147; s++; break;
        case 190: *(d++) = 197; *(d++) = 184; s++; break;
        default:  *(d++) = 194; *(d++) = *(s++); break;
      } else {
        *(d++) = 195;
        *(d++) = *(s++) - 64;
      }
    }
  }
  /* Done. Remember to free() the resulting string when no longer needed. */
  return result;
}


/* taken from: http://stackoverflow.com/questions/11258019/conversion-from-iso-8859-15-latin9-to-utf-8
   ISO 8859-15 is ISO8859/1 plus the EUR symbol;
   written by Nominal Animal, http://stackoverflow.com/users/1475978/nominal-animal.

   Creates a dynamically allocated copy of string, changing the encoding from UTF-8 to ISO-8859-1/15.
   Unsupported code points are not ignored. */

LUALIB_API char *utf8_to_latin9 (const char *string) {
  size_t size = 0;
  size_t used = 0;
  unsigned char *result = NULL;
  if (string) {
    const unsigned char *s = (const unsigned char *)string;
    while (*s) {
      if (used >= size) {
        void *const old = result;
        size = (used | 255) + 257;
        result = realloc(result, size*sizeof(char));
        if (!result) {
          if (old) free(old);
          return NULL;
        }
      }
      if (*s < 128) {
        result[used++] = *(s++);
        continue;
      } else
      if (s[0] == 226 && s[1] == 130 && s[2] == 172) {
        result[used++] = 164;
        s += 3;
        continue;
      } else
      if (s[0] == 194 && s[1] >= 128 && s[1] <= 191) {
        result[used++] = s[1];
        s += 2;
        continue;
      } else
      if (s[0] == 195 && s[1] >= 128 && s[1] <= 191) {
        result[used++] = s[1] + 64;
        s += 2;
        continue;
      } else
      if (s[0] == 197 && s[1] == 160) {
        result[used++] = 166;
        s += 2;
        continue;
      } else
      if (s[0] == 197 && s[1] == 161) {
        result[used++] = 168;
        s += 2;
        continue;
      } else
      if (s[0] == 197 && s[1] == 189) {
        result[used++] = 180;
        s += 2;
        continue;
      } else
      if (s[0] == 197 && s[1] == 190) {
        result[used++] = 184;
        s += 2;
        continue;
      } else
      if (s[0] == 197 && s[1] == 146) {
        result[used++] = 188;
        s += 2;
        continue;
      } else
      if (s[0] == 197 && s[1] == 147) {
        result[used++] = 189;
        s += 2;
        continue;
      } else
      if (s[0] == 197 && s[1] == 184) {
        result[used++] = 190;
        s += 2;
        continue;
      }
      if (s[0] >= 192 && s[0] < 224 &&
        s[1] >= 128 && s[1] < 192) {
        s += 2;
        continue;
      } else
      if (s[0] >= 224 && s[0] < 240 &&
        s[1] >= 128 && s[1] < 192 &&
        s[2] >= 128 && s[2] < 192) {
        s += 3;
        continue;
      } else
      if (s[0] >= 240 && s[0] < 248 &&
        s[1] >= 128 && s[1] < 192 &&
        s[2] >= 128 && s[2] < 192 &&
        s[3] >= 128 && s[3] < 192) {
        s += 4;
        continue;
      } else
      if (s[0] >= 248 && s[0] < 252 &&
        s[1] >= 128 && s[1] < 192 &&
        s[2] >= 128 && s[2] < 192 &&
        s[3] >= 128 && s[3] < 192 &&
        s[4] >= 128 && s[4] < 192) {
        s += 5;
        continue;
      } else
      if (s[0] >= 252 && s[0] < 254 &&
        s[1] >= 128 && s[1] < 192 &&
        s[2] >= 128 && s[2] < 192 &&
        s[3] >= 128 && s[3] < 192 &&
        s[4] >= 128 && s[4] < 192 &&
        s[5] >= 128 && s[5] < 192) {
        s += 6;
        continue;
      }
      /* if (*s == 132 ||  //  from console
          *s == 148 ||  //  from console
          *s == 129 ||  //  from console
          *s == 225 ||  //  from console
          *s == 142 ||  //  from console
          *s == 153 ||  //  from console
          *s == 154 ||  //  from console
          *s == 155 ||  //  from console
          *s == 138 ||  //  from console
          *s == 130 ||  //  from console
          *s == 144 ||  //  from console
          *s == 144)    //  from console */
      /* 2.12.0 RC 2, don't drop char but put it into result */
      result[used++] = *(s++);
    }
  }
  {
    void *const old = result;
    size = (used | 7) + 1;
    result = realloc(result, size*sizeof(char));
    if (!result) {
      if (old) free(old);
      return NULL;
    }
    memset(result + used, 0, (size - used)*sizeof(char));
  }
  /* Done. Remember to free() the resulting string when no longer needed. */
  return (char *)result;
}


LUALIB_API size_t utf8_to_latin9_len (const char *string, size_t *utf8len) {
  const unsigned char *s;
  size_t used = 0;
  *utf8len = 0;
  if (string == NULL) return 0;
  s = (const unsigned char *)string;
  while (*s) {
    if (*s < 128) {
      used++; s++;
      continue;
    } else
    if (s[0] == 226 && s[1] == 130 && s[2] == 172) {
      used++;
      s += 3;
      continue;
    } else
    if (s[0] == 194 && s[1] >= 128 && s[1] <= 191) {
      used++;
      s += 2;
      continue;
    } else
    if (s[0] == 195 && s[1] >= 128 && s[1] <= 191) {
      used++;
      s += 2;
      continue;
    } else
    if (s[0] == 197 && s[1] == 160) {
      used++;
      s += 2;
      continue;
    } else
    if (s[0] == 197 && s[1] == 161) {
      used++;
      s += 2;
      continue;
    } else
    if (s[0] == 197 && s[1] == 189) {
      used++;
      s += 2;
      continue;
    } else
    if (s[0] == 197 && s[1] == 190) {
      used++;
      s += 2;
      continue;
    } else
    if (s[0] == 197 && s[1] == 146) {
      used++;
      s += 2;
      continue;
    } else
    if (s[0] == 197 && s[1] == 147) {
      used++;
      s += 2;
      continue;
    } else
    if (s[0] == 197 && s[1] == 184) {
      used++;
      s += 2;
      continue;
    }
    if (s[0] >= 192 && s[0] < 224 &&
      s[1] >= 128 && s[1] < 192) {
      s += 2;
      continue;
    } else
    if (s[0] >= 224 && s[0] < 240 &&
      s[1] >= 128 && s[1] < 192 &&
      s[2] >= 128 && s[2] < 192) {
      s += 3;
      continue;
    } else
    if (s[0] >= 240 && s[0] < 248 &&
      s[1] >= 128 && s[1] < 192 &&
      s[2] >= 128 && s[2] < 192 &&
      s[3] >= 128 && s[3] < 192) {
      s += 4;
      continue;
    } else
    if (s[0] >= 248 && s[0] < 252 &&
      s[1] >= 128 && s[1] < 192 &&
      s[2] >= 128 && s[2] < 192 &&
      s[3] >= 128 && s[3] < 192 &&
      s[4] >= 128 && s[4] < 192) {
      s += 5;
      continue;
    } else
    if (s[0] >= 252 && s[0] < 254 &&
      s[1] >= 128 && s[1] < 192 &&
      s[2] >= 128 && s[2] < 192 &&
      s[3] >= 128 && s[3] < 192 &&
      s[4] >= 128 && s[4] < 192 &&
      s[5] >= 128 && s[5] < 192) {
      s += 6;
      continue;
    }
    used++; s++;
  }
  *utf8len = s - (const unsigned char *)string;  /* length of UTF-8 string */
  return used;  /* size of string converted to Latin */
}


/* detects that the given string is in UTF-8 encoding;
   written by Christoph, http://stackoverflow.com/users/48015/christoph,
   see: http://stackoverflow.com/questions/1031645/how-to-detect-utf-8-in-plain-c */

LUALIB_API int is_utf8 (const char *string, size_t *pos) {
  if (!string) return 0;
  const unsigned char *bytes = (const unsigned char *)string;
  while (*bytes) {
    if ( (  /* ASCII use bytes[0] <= 0x7F to allow ASCII control characters */
      bytes[0] == 0x09 ||
      bytes[0] == 0x0A ||
      bytes[0] == 0x0D ||
      (0x20 <= bytes[0] && bytes[0] <= 0x7E)
      )
    ) {
      bytes += 1;
      continue;
    }
    if ( (  /* non-overlong 2-byte */
      (0xC2 <= bytes[0] && bytes[0] <= 0xDF) &&
      (0x80 <= bytes[1] && bytes[1] <= 0xBF)
    )
    ) {
      bytes += 2;
      continue;
    }
    if ( (  /* excluding overlongs */
        bytes[0] == 0xE0 &&
        (0xA0 <= bytes[1] && bytes[1] <= 0xBF) &&
        (0x80 <= bytes[2] && bytes[2] <= 0xBF)
      ) ||
      (  /* straight 3-byte */
        ((0xE1 <= bytes[0] && bytes[0] <= 0xEC) ||
          bytes[0] == 0xEE ||
          bytes[0] == 0xEF) &&
        (0x80 <= bytes[1] && bytes[1] <= 0xBF) &&
        (0x80 <= bytes[2] && bytes[2] <= 0xBF)
      ) ||
      (  /* excluding surrogates */
        bytes[0] == 0xED &&
        (0x80 <= bytes[1] && bytes[1] <= 0x9F) &&
        (0x80 <= bytes[2] && bytes[2] <= 0xBF)
      )
    ) {
      bytes += 3;
      continue;
    }
    if ( (  /* planes 1-3 */
        bytes[0] == 0xF0 &&
        (0x90 <= bytes[1] && bytes[1] <= 0xBF) &&
        (0x80 <= bytes[2] && bytes[2] <= 0xBF) &&
        (0x80 <= bytes[3] && bytes[3] <= 0xBF)
      ) ||
      (  /* planes 4-15 */
        (0xF1 <= bytes[0] && bytes[0] <= 0xF3) &&
        (0x80 <= bytes[1] && bytes[1] <= 0xBF) &&
        (0x80 <= bytes[2] && bytes[2] <= 0xBF) &&
        (0x80 <= bytes[3] && bytes[3] <= 0xBF)
      ) ||
      (  /* plane 16 */
        bytes[0] == 0xF4 &&
        (0x80 <= bytes[1] && bytes[1] <= 0x8F) &&
        (0x80 <= bytes[2] && bytes[2] <= 0xBF) &&
        (0x80 <= bytes[3] && bytes[3] <= 0xBF)
      )
    ) {
      bytes += 4;
      continue;
    }
    *pos = bytes - (const unsigned char *)string + 1;  /* 2.12.6 */
    return 0;
  }
  pos = 0;
  return 1;
}


/* Assumes a string is UTF-8 encoded and determines its size, counting UTF-8 marks (0xEF) but not succeeding UTF-8 multibytes; written by mpez0,
   http://stackoverflow.com/users/27898/mpez0 */

LUALIB_API size_t size_utf8 (const char *str) {  /* changed 2.12.6 */
  size_t j = 0;
  while (*str) {
    if ((*str++ & 0xc0) != 0x80) j++;  /* not in [128, 191] ? */
  }
  return j;
}


LUALIB_API int tools_isnumericstring (const char *s) {  /* 26.08.2012, Agena 1.7.7/1.9.1 */
  struct lconv *cv = localeconv();
  char decpoint = (cv ? cv->decimal_point[0] : '.');
  int flag = 1;
  size_t l = strlen(s);
  if (*s == '\0') return 0;
  for ( ; *s != '\0'; s++) {
    if (uchar(*s) < '0' || uchar(*s) > '9') {
      if (uchar(*s) == decpoint && flag && !(l == 1))
        flag = 0;
      else return 0;
    }
  }
  return 1;
}


/* taken from: http://stackoverflow.com/questions/779875/what-is-the-function-to-replace-string-in-c,
   written by rampion. FREE it !  Removed const declarations with 2.16.1 to avoid
   GCC 8.3.0 warnings. */
LUALIB_API char *strreplace (const char *original, const char *pattern, const char *replacement) {
  const char *oriptr, *patloc;
  char *retptr;
  size_t replen = strlen(replacement);
  size_t patlen = strlen(pattern);
  size_t orilen = strlen(original);
  size_t patcnt = 0;
  /* find how many times the pattern occurs in the original string */
  for (oriptr=original; (patloc = strstr(oriptr, pattern)); oriptr = patloc + patlen)
    patcnt++;
  /* allocate memory for the new string */
  size_t retlen = orilen + patcnt * (replen - patlen);
  char *returned = tools_stralloc(retlen);  /* 2.16.5, (char *)malloc(sizeof(char) * (retlen + 1)); */
  if (returned != NULL) {
    /* copy the original string, replacing all the instances of the pattern */
    retptr = returned;
    for (oriptr=original; (patloc = strstr(oriptr, pattern)); oriptr = patloc + patlen) {
      size_t const skplen = patloc - oriptr;
      /* copy the section until the occurrence of the pattern */
      strncpy(retptr, oriptr, skplen);
      retptr += skplen;
      /* copy the replacement */
      strncpy(retptr, replacement, replen);
      retptr += replen;
    }
    /* copy the rest of the string. */
    strcpy(retptr, oriptr);
  }
  return returned;
}


/* Reverses a string in-place. 2.10.0, based upon the StackOverflow article `c -
   Is the strrev() function not available in Linux`, answer #13 */
LUALIB_API void tools_strrev (char *str) {
  char ch;
  int i, j;
  i = strlen(str) - 1;
  j = 0;
  while (i > j) {
    ch = str[i];
    str[i] = str[j];
    str[j] = ch;
    i--; j++;
  }
}


/* Like strcmp, but compares case-insensitively, 2.16.1 */
LUALIB_API int tools_stricmp (const char* p, const char* q) {
  while (tolower(uchar(*p)) == tolower(uchar(*q))) {
    if (*p == '\0') return 0;
    p++; q++;
  }
  return (int)tolower(uchar(*p)) - (int)tolower(uchar(*q));
}


LUALIB_API int tools_streqx (const char *src, ...) {
  va_list args;
  const char *x, *s;
  s = src;
  va_start(args, src);
  while ((x = va_arg(args, const char *))) {
    if (tools_streq(s, x)) { va_end(args); return 1; }
  }
  va_end(args);
  return 0;
}


/* Determines the optimal memory allocation size for a string of length l, with l NOT including the terminating trailing
   NULL character, i.e. '\0'.
   The return is the optimal number of bytes to be allocated including a terminating NULL character, and the number of
   4/8-byte chunks to be allocated in parameter `chunks`. See also: tools_stralloc. */
LUALIB_API INLINE size_t tools_optstrlen (size_t l, size_t *chunks) {
  /* add extra bytes to fill up the last 4/8-byte chunk if necessary, regardless whether the strings on the system are aligned of not */
  size_t rem = (++l) % AGN_BLOCKSIZE;  /* length including \0 */
  l += (rem != 0)*(AGN_BLOCKSIZE - rem);
  *chunks = l / AGN_BLOCKSIZE;         /* chunks of bytes to be allocated */
  return l;                            /* optimal string length in bytes, works flawlessly even with empty strings */
}


/* Convert double (or int, etc.) into a string, taken from agnconf.h, safe for doubles

  ! FREE the string after usage ! */

#define MAXNUMBER2STR 32  /* 16 digits, sign, point, and \0 */

LUALIB_API char *tools_dtoa (double x) {
  char *buf = malloc(MAXNUMBER2STR * sizeof(char));
  if (!buf) return NULL;  /* 2.16.5 */
  sprintf(buf, "%.14g", x);
  return buf;
}


/* changes a float value with a decimal comma to one with a decimal dot, in-place; 2.3.0 RC 2 */
LUALIB_API void tools_commatodot (char *str, int *result) {
  size_t numcolons = 0;
  char *oldpos = str;
  *result = 0;
  while (*str) {
    if (*str == ',')
      numcolons++;
    else if (*str < '0' || *str > '9')
      return;
    str++;
  }
  if (numcolons == 1) {
    str = oldpos;
    charreplace(str, ',', '.', 0);
    *result = 1;
  }
  return;
}


LUALIB_API INLINE void *tools_realloc (void *p, size_t size) {
  if (p == NULL) return calloc(1, size);
  p = (void *)realloc(p, size);
  return p;
}

LUALIB_API char *tools_getline (FILE *fp, size_t *linesize, int *r) {  /* reads one line of any length, FREE result ! */
  size_t bufsize;
  char buf[BUFSIZ], *line, *q;
  *linesize = 0;
  line = NULL;
  do {
    if (fgets(buf, BUFSIZ, fp) == NULL) break;
    q = strchr(buf, '\n');
    if (q != NULL) *q = '\0';
    bufsize = strlen(buf);
    line = (char *)tools_realloc(line, sizeof(char) * (bufsize + *linesize + 1) );
    if (line == NULL) {  /* memory allocation error ? */
      *r = 1;
      return NULL;
    }
    *linesize += bufsize;
    strcat(line, buf);
  } while (q == NULL);
  *r = 0;
  return line;
}


/* retrieve information on your platform;
0.12.2, October 12, 2008; extended to correctly identify Vista, Windows 2003, and Windows 2007 Server
on April 08, 2009 - 0.13.4; updated 1.7.9, 07.09.2012. */
#ifdef _WIN32
/* for Windows 7 or later, https://stackoverflow.com/questions/32115255/c-how-to-detect-windows-10, posted by Michael Haephrati;
   returns 6 for Windows 8.1, but 10 for Windows 10. */
LUALIB_API int getSysOpType (int *minorVersion, int *BuildNumber, int *PlatformId, int *ProductType, double *winver, uint8_t *SPmaj, uint8_t *SPmin) {
  int ret = 0;
  NTSTATUS(WINAPI *RtlGetVersion)(LPOSVERSIONINFOEXW);
  OSVERSIONINFOEXW osInfo;
  *(FARPROC*)&RtlGetVersion = GetProcAddress(GetModuleHandleA("ntdll"), "RtlGetVersion");
  *minorVersion = *BuildNumber = *PlatformId = *ProductType = *SPmaj = *SPmin = -1;
  if (RtlGetVersion != NULL) {
    osInfo.dwOSVersionInfoSize = sizeof(osInfo);
    RtlGetVersion(&osInfo);
    ret = osInfo.dwMajorVersion;
    *minorVersion = osInfo.dwMinorVersion;
    *BuildNumber  = osInfo.dwBuildNumber;
    *PlatformId   = osInfo.dwPlatformId;
    *ProductType  = osInfo.wProductType;
    *SPmaj        = osInfo.wServicePackMajor;
    *SPmin        = osInfo.wServicePackMinor;
    *winver       = ret + 0.1*(*minorVersion);
  } else
    *winver = 0.0;
  return ret;
}

/* correct MinorVersion, etc. */
#define correctWinVerData() { \
  winver->MinorVersion = minversion; \
  winver->BuildNumber  = build; \
  winver->PlatformId   = platform; \
  winver->ProductType  = product; \
}

LUALIB_API int getWindowsVersion (struct WinVer *winver) {
  int result, MajorVersion, MinorVersion, ProductType, PlatformId;
  double winversion;
  uint8_t SPmaj, SPmin;
  OSVERSIONINFOEX version;
  version.dwOSVersionInfoSize = sizeof(version);
  GetVersionEx((OSVERSIONINFO *)&version);
  /*  GREP "static const char *versions[]" * when extending this procedure

  dwMajorVersion
  Major version number of the operating system. This member can be one of the following values. Value Meaning
  4 Windows NT 4.0, Windows Me, Windows 98, or Windows 95.
  5 Windows Server 2003, Windows XP, or Windows 2000
  6 Windows Vista, Windows Server 2008, Windows 7, 8, 8.1

  dwMinorVersion
  Minor version number of the operating system. This member can be one of the following values. Value Meaning
  0 Windows 2000, Windows NT 4.0, Windows 95, or Windows Vista
  1 Windows XP, Windows 2008 server
  2 Windows Server 2003, Windows 7, Windows Server 2012
  3 Windows Server 2012 RC 2, Windows 8, 8.1
  10 Windows 98
  90 Windows Me

  dwPlatformId
  Operating system platform. This member can be one of the following values. Value Meaning
  VER_PLATFORM_WIN32_NT
  2 Windows Server 2003, Windows XP, Windows 2000, or Windows NT.
  VER_PLATFORM_WIN32_WINDOWS
  1 Windows Me, Windows 98, or Windows 95.
  VER_PLATFORM_WIN32s
  0 Win32s on Windows 3.1. */
  result = 18;  /* unknown default */
  winver->BuildNumber = (int)version.dwBuildNumber;
  MajorVersion = winver->MajorVersion = (int)version.dwMajorVersion;
  MinorVersion = winver->MinorVersion = (int)version.dwMinorVersion;
  PlatformId = winver->PlatformId = (int)version.dwPlatformId;
  ProductType = winver->ProductType = (char)version.wProductType;
  if (MajorVersion == 4) {
    if (MinorVersion == 0 && PlatformId == 1)
      result = MS_WIN95;      /* 2, Win 95 */
    else if (MinorVersion == 0 && PlatformId == 2)
      result = MS_WINNT4;     /* 3, Win NT 4.0 */
    else if (MinorVersion == 10)
      result = MS_WIN98;      /* 4, Win 98 */
    else if (MinorVersion == 90)
      result = MS_WINME;      /* 5, Win Me, return value changed 1.9.3 */
  } else if (MajorVersion == 5) {
    if (MinorVersion == 0)
      result = MS_WIN2K;      /* 6, Win 2000, return value changed 1.9.3  */
    else if (MinorVersion == 1)
      result = MS_WINXP;      /* 7, Win XP */
    else if (MinorVersion == 2)
      result = MS_WIN2003;    /* 8, Win 2003 */
  } else if (MajorVersion == 6) {
    if (MinorVersion == 0 && ProductType == VER_NT_WORKSTATION)
      result = MS_WINVISTA;   /* 9, Win Vista */
    else if (MinorVersion >= 0 && ProductType != VER_NT_WORKSTATION)
      result = MS_WINS2008;   /* 10, Win Server 2008 */
    else if (MinorVersion == 1 && ProductType == VER_NT_WORKSTATION)
      result = MS_WIN7;       /* 11, Windows 7 */
    else if (MinorVersion == 2 && ProductType == VER_NT_WORKSTATION)
      result = MS_WIN80;      /* 12, Windows 8 */
    else if (MinorVersion == 2 && ProductType != VER_NT_WORKSTATION)
      result = MS_WINS2012;   /* 13, Windows Server 2012 */
  }
  else if (PlatformId == 0)
    result = MS_WIN32S;       /* 1, Win32s under Windows 3.1 */
  else  /* CHANGE result initalisation at the top if you change this line */
    result = MS_WIN10POST;    /* 18, 2.8.5 patch, assume higher Windows version, checked other source files former 0 return. */
  if (result >= MS_WIN80) {   /* DON'T MOVE TO TOP due to unknown behaviour on Vista and earlier!
    2.14.1 fix: we are at least on Windows 8.0 but additionally we have to check for 8.1 and later,
    since on 8.1 and 10, MajorVersion == 6 && MinorVersion == 2 depicting 8.0 */
    int majversion, minversion, build, platform, product;
    majversion = getSysOpType(&minversion, &build, &platform, &product, &winversion, &SPmaj, &SPmin);
    if (majversion == 6 && minversion == 3 && ProductType == VER_NT_WORKSTATION) {
      correctWinVerData();
      result = MS_WIN81;      /* 14, Windows 8.1 */
    } else if (majversion == 6 && minversion == 3 && ProductType != VER_NT_WORKSTATION) {
      correctWinVerData();
      result = MS_WINS2012R2; /* 15, Windows Server 2012 R2 */
    } else if (majversion == 10 && product == 1) {
      correctWinVerData();
      result = MS_WIN10;      /* 16, Windows 10 */
    } else if (majversion == 10 && product != 1) {
      correctWinVerData();
      result = MS_WIN10S;     /* 17, Windows 10 Server */
    }
  }
  return result;
}
#endif


/* Get full path of running application, 2.16.1.
   FREE it !  Note that free() according to the C standard just quits when pointer is NULL.
   See Internet article: "executable - Absolute path to currently executing program - Unix & Linux Stack Exchange" */
LUALIB_API char *getModuleFileName (void) {
  char *apath;
#if defined(_WIN32)
  int winversion;
  struct WinVer winver;
  winversion = getWindowsVersion(&winver);
  if (winversion < MS_WINXP) return NULL;  /* not(XP or later) ? */
#endif
  apath = tools_stralloc(MAX_PATH);  /* 2.16.5 */
  if (apath == NULL) return NULL;
#if defined(_WIN32)
  if (GetModuleFileName(NULL, apath, MAX_PATH + 1)) return apath;
#elif defined(__linux__)
  int nchar;
  nchar = readlink("/proc/self/exe", apath, MAX_PATH + 1);
  if (nchar >= 0 && nchar < (MAX_PATH + 1)) {
    apath[nchar] = '\0';
    return apath;
  }
#endif
  return NULL;
}


/* 2.1.3, compares two numbers using Donald Knuth's approximation method, see approx in lbaselib */
LUALIB_API int tools_approx (double x, double y, double eps) {
  if (unlikely((x == y) || (isnan(x) && isnan(y)))) {  /* 2.5.15 optimisation */
    return 1;
  } else {
    double dist = fabs(x - y);
    return dist <= eps || dist <= (eps * fMax(fabs(x), fabs(y)));
  }
}


/* sets bit 0 or 1 at position pos (always positive) to integer x */
LUALIB_API INLINE void tools_setbit (int *x, int pos, int bit) {  /* 2.3.3 */
  if (pos < 1 || bit < 0) return;  /* do nothing */
  *x = (*x & ~(1 << (pos - 1))) | (bit << (pos - 1));
}


LUALIB_API INLINE int tools_getbit (int x, int pos) {  /* 2.3.3 */
  if (pos < 1) return -1;
  if (x >= 0)
    return (x & (1 << (pos - 1))) != 0;
  else
    return (-x & (1 << (pos - 1))) != 0;
}


#ifndef PROPCMPLX
LUALIB_API complex double tools_centier (agn_Complex a) {  /* 2.3.3 */
  double aa, bb, re, im;
  agn_Complex X;
  re = creal(a);
  im = cimag(a);
  aa = re - sun_floor(re);
  bb = im - sun_floor(im);
  if (aa + bb < 1)
    X = 0 + 0*I;
  else if ((aa + bb >= 1) && (aa >= bb))
    X = 1 + 0*I;
  else
    X = 1*I;
  return sun_floor(re) + I*sun_floor(im) + X;
}
#else
LUALIB_API void tools_centier (double a, double b, double *x, double *y) {
  double aa, bb, X, Y;
  aa = a - sun_floor(a);
  bb = b - sun_floor(b);
  if ((aa + bb) < 1) {
    X = 0; Y = 0; }
  else if ((aa + bb) >= 1 && (aa >= bb)) {
    X = 1; Y = 0; }
  else {
    X = 0; Y = 1; }
  *x = sun_floor(a) + X; *y = sun_floor(b) + Y;
}
#endif


/* Copyright (C) 2002 by  Red Hat, Incorporated. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * is freely granted, provided that this notice is preserved.
 */
LUALIB_API double tools_fdim (double x, double y) {
  if (isnan(x) || isnan(y)) return AGN_NAN;
  return x > y ? x - y : 0.0;
}


/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* FIXME (?): There are strange situations on some Windows platforms where both this SunPro function and built-in platform
   signbit individually return correct results in Agena, but when both call results with the same argument -0 are compared
   with the (in)equality Agena operator the result differs, probably due to overflows ??? */
LUALIB_API int tools_signbit (double x) {  /* 2.8.4 */
  int32_t hx;
  __GET_HI(hx, x);
  return hx & 0x80000000;
}


/* Fused multiply-add pendant, 2.8.4 */
typedef struct { double hi; double lo; } doublefloat;  /* do not use float for the two components hi and lo ! */

doublefloat aux_split (double a) {  /* using volatile double t, hi, lo is not needed in GCC */
  double t = ((1 << 12) + 1)*a;     /* 12 is the best shift for this version to come close to C's fma function */
  double hi = t - (t - a);
  double lo = a - hi;
  return (doublefloat){hi, lo};
}

LUALIB_API double tools_fma (double a, double b, double c) {
  doublefloat as, bs;
  as = aux_split(a), bs = aux_split(b);
  return ((as.hi*bs.hi + c) + as.hi*bs.lo + as.lo*bs.hi) + as.lo*bs.lo;
}


LUALIB_API INLINE double tools_mhypot (double a, double b) {  /* hypot4, sqrt(a^2 - b^2), 2.11.0 RC2 */
  lua_Number t;
  if (unlikely( isnan(a) || isnan(b) || fabs(a) < fabs(b))) return AGN_NAN;  /* 2.11.1 fix, 2.5.15 optimisation */
  if (unlikely( isinf(a) != 0 || isinf(b) != 0 )) return HUGE_VAL;  /* 2.5.15 optimisation */
  if (a < 0) a = -a;
  if (b < 0) b = -b;
  if (b > a) { SWAP(a, b, t) }
  if (a + b == a) return (a);  /* -> a == 0 || b == 0 */
  t = b/a;  /* minimum divided by maximum */
  return a * sqrt(1 - t*t);
}


/* portable version of fpclassify which is not available on all platforms */
LUALIB_API int tools_fpclassify (double x) {
  uint32_t hx, lx;
  __EXTRACT_WORDS (hx, lx, x);
  lx |= hx & 0xfffff;
  hx &= 0x7ff00000;
  if ((hx | lx) == 0)
    return FP_ZERO;
  else if (hx == 0x7ff00000)
    return (lx == 0) ? FP_INFINITE : FP_NAN;
  else if (hx == 0)
    return FP_SUBNORMAL;
  return FP_NORMAL;
}


/* taken from: @(#)s_nextafter.c 5.1 93/09/24
 *
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 *
 * IEEE functions
 *   nextafter(x,y)
 *   return the next machine floating-point number of x in the direction toward y.
 *   Special cases:
 *
 * Modification taken from www.koders.com /..GNU/l/libc/libc/math/s_nextafter.c
 * because the original Sun implementation is non-ANSI C and does not work with GCC.
 *
 * This implementation is ten percent faster than the one available in GNU C Lib.
 */

LUALIB_API double sun_nextafter (double x, double y) {
  int32_t hx, hy, ix, iy;
  uint32_t lx, ly;
  __EXTRACT_WORDS(hx, lx, x);
  __EXTRACT_WORDS(hy, ly, y);
  ix = hx & 0x7fffffff;      /* |x| */
  iy = hy & 0x7fffffff;      /* |y| */
  if (unlikely(((ix >= 0x7ff00000) && ((ix - 0x7ff00000) | lx) != 0) ||   /* x is nan, 2.5.15 optimisation */
     ((iy >= 0x7ff00000) && ((iy - 0x7ff00000) | ly) != 0)))      /* y is nan */
    return x + y;
  if (x == y) return y;      /* x = y, return y */
  if ((ix | lx) == 0) {      /* x == 0 */
    double u;
    __INSERT_WORDS(x, hy & 0x80000000, 1);   /* return +-minsubnormal */
    u = math_opt_barrier(x);
    u = u*u;
    math_force_eval(u);     /* raise underflow flag */
    return x;
  }
  if (hx >= 0) {            /* x > 0 */
    if (hx > hy || ((hx == hy) && (lx > ly))) {   /* x > y, x -= ulp */
      if (lx == 0) hx -= 1;
      lx -= 1;
    } else {                /* x < y, x += ulp */
      lx += 1;
      if (lx == 0) hx += 1;
    }
  } else {                  /* x < 0 */
    if (hy >= 0 || hx > hy || ((hx == hy) && (lx > ly))) {  /* x < y, x -= ulp */
      if (lx == 0) hx -= 1;
      lx -= 1;
    } else {                /* x > y, x += ulp */
      lx += 1;
      if (lx == 0) hx += 1;
    }
  }
  hy = hx & 0x7ff00000;
  if (unlikely(hy >= 0x7ff00000)) {  /* 2.5.15 optimisation */
    x = x + x;   /* overflow */
    if (FLT_EVAL_METHOD != 0 && FLT_EVAL_METHOD != 1)
      asm ("" : "+m"(x));
    return x;   /* overflow */
  }
  if (unlikely(hy < 0x00100000)) {  /* 2.5.15 optimisation */
    double u = x * x;       /* underflow */
    math_force_eval(u);     /* raise underflow flag */
  }
  __INSERT_WORDS(x, hx, lx);
  return x;
}


#define TWO54   1.80143985094819840000e+16;  /* 0x43500000, 0x00000000 */

/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 *
 * The following two functions are based on Sun's frexp implementation
 * (source file s_frexp.c), taken from the ulibc 0.33.2 library.
 */

/*
 * for non-zero x
 *	x = frexp(arg,&exp);
 * return a double fp quantity x such that 0.5 <= |x| < 1.0
 * and the corresponding binary exponent "exp". That is
 *	arg = x*2^exp.
 * If arg is inf, 0.0, or NaN, then frexp(arg,&exp) returns arg
 * with *exp=0.
 *
 * A little bit modified to return either the exponent (mode = 1) or
 * mantissa (mode = 0).
 */

LUALIB_API double sun_frexp_man (double x) {  /* returns mantissa, 2.10.0 */
  int32_t hx, ix, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ix = 0x7fffffff & hx;  /* absolute value |x| */
  if (ix >= 0x7ff00000 || ((ix | lx) == 0)) return x;  /* 0, inf, nan */
  if (unlikely(ix < 0x00100000)) {  /* subnormal, 2.5.15 optimisation */
    x *= TWO54;
    __GET_HI(hx, x);
  }
  hx = (hx & 0x800fffff) | 0x3fe00000;
  __SET_HI(x, hx);
  return x;
}


LUALIB_API int sun_frexp_exp (double x) {  /* returns exponent, 2.10.0 */
  int eptr;
  int32_t hx, ix, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ix = 0x7fffffff & hx;  /* absolute value */
  eptr = 0;
  if (ix >= 0x7ff00000 || ((ix | lx) == 0)) return x;  /* 0, inf, nan */
  if (unlikely(ix < 0x00100000)) {  /* subnormal, 2.5.15 optimisation */
    x *= TWO54;
    __GET_HI(hx, x);
    ix = hx & 0x7fffffff;
    eptr = -54;
  }
  return eptr + (ix >> 20) - 1022;
}


LUALIB_API double sun_xfrexp (double x, int *eptr, int *isneg) {  /* returns mantissa, exponent, signbit, 2.10.0 */
  int32_t hx, ix, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ix = 0x7fffffff & hx;
  *eptr = 0;
  *isneg = (hx >> 31) & 1;
  if (ix >= 0x7ff00000 || ((ix | lx) == 0)) {
    hx &= 0x7fffffff;  /* absolute value (for -0) */
    __SET_HI(x, hx);
    return x;  /* 0, inf, nan */
  }
  if (unlikely(ix < 0x00100000)) {  /* subnormal, 2.5.15 optimisation */
    x *= TWO54;
    __GET_HI(hx, x);
    ix = hx & 0x7fffffff;
    *eptr = -54;
  }
  *eptr += (ix >> 20) - 1022;
  hx = (hx & 0x800fffff) | 0x3fe00000;  /* nil entire exponent, retain sign and mantissa */
  hx &= 0x7fffffff;  /* absolute value */
  __SET_HI(x, hx);
  return x;
}


/* Unbiased exponent: returns math.exponent(x) - 1, except 0 -> -1023; suited for expressions such like 2^52 -> 52;
   based on Sun's `rint`, 2.16.0 */
LUALIB_API INLINE int32_t sun_exponent (double x) {
  int32_t hx;
  GET_HIGH_WORD(hx, x);
  hx &= 0x7fffffff;  /* absolute value */
  return ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;
}


LUALIB_API INLINE int sun_irregular (double x) {  /* x > 2^53, i.e. not all larger integers representable ?  2.16.0 */
  int32_t hx, lx;
  GET_HIGH_WORD(hx, x);
  hx &= 0x7fffffff;  /* absolute value */
  if (hx < 0x43400000) return 0;
  GET_LOW_WORD(lx, x);
  if (hx == 0x43400000 && lx == 0) return 0;
  return 1;
}


LUALIB_API int sun_subnormal (double x) {
  int32_t hx, ix, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ix = 0x7fffffff & hx;  /* absolute value */
  if (ix >= 0x7ff00000 || ((ix | lx) == 0)) return 0;  /* 0, inf, nan */
  return (ix < 0x00100000);  /* subnormal */
}

/* # define fMod(a, b) ((a) - (b) * floor((a)/(b))) */
LUALIB_API double tools_reducerange (double x, double min, double max) {
  if (min <= x && x < max) return x;  /* just to save computation time */
  x -= min;
  max -= min;
  return min + fmod(max + fmod(x, max), max);  /* 2.11 RC4 tuning */
}


/* Returns its argument x - a number - if x is non-negative, otherwise returns 0. By passing any
   non-negative optional number d (the `direction`), the return is the same.

   By passing any non-positive optional number d, returns x if it is negative, otherwise returns 0.

   See also independent version math.branch. Used by skycrane.tolerance. */
LUALIB_API double tools_branch (double x, int d) {  /* 2.10.0 */
  /* x the number, d the direction or interval: if d > 0 -> return in [0, infinity] */
  double r;
  r = signbit(x) == 0;  /* if d <= 0 -> return in [-infinity, 0] */
  r = (d > 0) ? r*x : (!r)*x;
  if (r == 0 && signbit(r) != 0) r = 0;  /* convert -0 to 0 */
  return r;
}


/* Computes a mathematical epsilon value that is dependent on the value resp. magnitude of its argument x. It can be
   used in difference quotients, etc., for it prevents huge precision errors with computations on very small or large
   numbers (i.e. C doubles).
   See: http://stackoverflow.com/questions/5543783/java-estimate-a-derivative-at-a-point, answered by trashgod */

LUALIB_API double tools_matheps (double x) {
  x = fabs(x);  /* 2.10.5 patch */
  if (x < 1) return AGN_EPSILON;
  return x * sqrt(sun_nextafter(x, HUGE_VAL) - x);
}


/* Excel Serial Date, 2.9.8, does not handle the DST issue (which Excel does not handle, too) */
LUALIB_API double tools_esd (int y, int m, int d, int hh, int mm, int ss) {
  long int md;
  double dh;
  if (y == 1900 && m == 2 && d == 29)  /* Lotus bug: there is no 29th February 1900 since 1900 is not a leap year */
    return 60;
  /* calculate Modified `Julian Lotus Serial Date`, do not remove brackets, integer arithmetic ! */
  md = (1461 * (y + 4800 + ((m - 14) / 12))) / 4 +
       (367 * (m - 2 - 12 * ((m - 14) / 12))) / 12 -
       (3 * (((y + 4900 + ((m - 14) / 12)) / 100))) / 4 +
       d - 2415019 - 32075;
  if (md < 60) md--;  /* any serial date under 60 is one day off */
  dh = (hh*3600.0 + mm*60.0 + ss)/86400.0;  /* time fraction */
  return md + dh;
}


/* round then trunc x at d decimal digits
   x: number to round
   d: dec digits to round/trunc to: 0 - 14

   taken from: https://bytes.com/topic/c/answers/521405-rounding-nearest-nth-digits, posted by Larry I Smith,
   original function name: `FixDec`.

   "This code dates from 1991. It was used to display financial data on reports. Use at your own risk..."

   Extended by Alexander Walz. Mr. Smith's function prevents round-off errors.

   t = 0: round towards infinity if z is positive, round towards -infinity if x is negative
   t = 4: apply round half up method
   t = 9: always truncate (towards zero).

   If t is 0 and x is negative, rounds toward 0. */
LUALIB_API double tools_roundf (double x, int d, int t) {
  double m, f, s;
  if (x == 0.0) return(x);
  s = 1;
  if (x < 0) {                  /* behave symmetrically */
    s = -1;
    x = -x;
  }
  if (d < -14) d = -14;         /* these if's really do not influence performance perceptibly */
  else if (d > 14) d = 14;      /* IEEE doubles max precision minus 1 */
  /* comments show what would happen if z = 1234.5678 & d = 2 */
  m = sun_pow(10.0, (double)d);     /* m: = 100.0 */
  x *= m;                       /* x: 1234.5678 -> 123456.78 */
  f = sun_modf(x, &x);          /* x: 123456.78 -> 123456.0 & f = 0.78; 2.11.0 improvement */
  f *= 10.0;                    /* f: 0.78 -> 7.8 */
  if (t != 0) sun_modf(f, &f);  /* f: 7.8 -> 7.0, &f gets the integral part, (int)f is the (d+1)-th digit; 2.11.0 improvement */
  if ((t == 0 && f > 0) ||      /* with t = 0, solely checking the (d+1)-th digit does not suffice */
     ((int)f > t)) x += 1.0;    /* with t = 4 (round half up method): x: 123456.0 -> 123457.0 */
  return(s * x / m);            /* 123457.0 / 100.0 = 1234.57 */
}


LUALIB_API double tools_binomial (double n, double k) {  /* 2.10.4 */
  if (tools_isfloat(n) || tools_isfloat(k)) {  /* 2.10.4 improvement, works as Maple does */
    if (tools_isint(k) && k < 0)
      return 0;
    else
      return cephes_gamma(n + 1)/cephes_gamma(k + 1)/cephes_gamma(n - k + 1);
  } else if (n < 0) {
    /* 2.10.4, work as Maple does, see: https://stackoverflow.com/questions/41195063/general-smoothstep-equation */
    int i;
    int r = 1;
    if (k < 0) {
      if (n >= 0) r = 0;
      else if (n - k < 0) r = 0;
      else k = n - k;
    } else if (n < 0) {
      r = sun_pow(-1, k);
      n = k - n - 1;
    }
    for (i=1; i <= k; i++, n--) {
      /* to avoid overflows, split c*n/i into (c/i*i + c%i)*n/i, see https://en.wikipedia.org/wiki/Binomial_coefficient */
      r = r/i*n + (int)r%(int)i*n/i;
    }
    if (r == -0) r = 0;
    return r;
  } else if (n < k)  /* 0.28.3 patch */
    return 0;
  else
    return sun_exp(lgamma(n + 1) - lgamma(k + 1) - lgamma(n - k + 1));
  return 1;
}


/* returns exp(x^2) if sign >= 0 and exp(-x^2) if sign < 0.
   Based on file `expx2.c` in the Cephes Math Library
   Release 2.9:  June, 2000, Copyright 2000 by Stephen L. Moshier */
#define MAXLOG       +170.0

LUALIB_API double tools_expx2 (double x, int sign) {
  double u, u1, m, f;
  if (x < 0) x = -x;
  if (sign < 0) x = -x;
  if (unlikely(x == -HUGE_VAL))  /* 2.10.1, avoid numeric exceptions with exp(-infinity), i.e. wrong value `undefined`, 2.5.15 optimisation */
    return 0;
  else if (unlikely(x == HUGE_VAL))  /* 2.10.1, dito, 2.5.15 optimisation */
    return HUGE_VAL;
  else {
    /* Represent x as an exact multiple of M (= 128) plus a residual.
       M is a power of 2 chosen so that exp(m * m) does not overflow
       or underflow and so that |x - m| is small. */
    m = .0078125 * sun_floor(128.0*x + 0.5);
    f = x - m;
    /* x^2 = m^2 + 2mf + f^2 */
    u = m*m;
    u1 = 2*m*f + f*f;
    if (sign < 0) {
      u = -u;
      u1 = -u1;
    }
    if (unlikely((u + u1) > MAXLOG))  /* 2.5.15 optimisation */
      return HUGE_VAL;
    else  /* u is exact, u1 is small.  */
      return(sun_exp(u) * sun_exp(u1));
  }
}

LUALIB_API double tools_cexpx2 (double a, double b, double sign, double *im) {
  double t1, t2, t4, t5;
  t1 = a*a;
  t2 = b*b;
  t4 = (sign < 0) ? sun_exp(-t1 + t2) : sun_exp(t1 - t2);
  t5 = a*b;
  *im = -((sign < 0) ? 1 : -1)*t4*sun_sin(2.0*t5);  /* imaginary part */
  return t4*sun_cos(2.0*t5);  /* real part */
}


/* @(#)s_sincos.c 5.1 13/07/15 */
/*
 * ====================================================
 * Copyright (C) 2013 Elliot Saba. All rights reserved.
 *
 * Developed at the University of Washington.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* sincos(x, s, c)
 * Several applications need sine and cosine of the same
 * angle x. This function computes both at the same time,
 * and stores the results in *sin and *cos.
 *
 * kernel function:
 *	__kernel_sin		... sine function on [-pi/4,pi/4]
 *	__kernel_cos		... cose function on [-pi/4,pi/4]
 *	__ieee754_rem_pio2	... argument reduction routine
 *
 * Method.
 *      Borrow liberally from s_sin.c and s_cos.c, merging
 *  efforts where applicable and returning their values in
 * appropriate variables, thereby slightly reducing the
 * amount of work relative to just calling sin/cos(x)
 * separately
 *
 * Special cases:
 *      Let trig be any of sin, cos, or tan.
 *      sincos(+-INF, s, c)  is NaN, with signals;
 *      sincos(NaN, s, c)    is that NaN;
 */

#define	__STRICT_ASSIGN(type, lval, rval) do {	\
	volatile type __lval;			\
						\
	if (sizeof(type) >= sizeof(double))	\
		(lval) = (rval);		\
	else {					\
		__lval = (rval);		\
		(lval) = __lval;		\
	}					\
} while (0)

/* @(#)k_rem_pio2.c 1.3 95/01/18
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/*
 * __kernel_rem_pio2(x,y,e0,nx,prec)
 * double x[],y[]; int e0,nx,prec;
 *
 * __kernel_rem_pio2 return the last three digits of N with
 *		y = x - N*pi/2
 * so that |y| < pi/2.
 *
 * The method is to compute the integer (mod 8) and fraction parts of
 * (2/pi)*x without doing the full multiplication. In general we
 * skip the part of the product that are known to be a huge integer (
 * more accurately, = 0 mod 8 ). Thus the number of operations are
 * independent of the exponent of the input.
 *
 * (2/pi) is represented by an array of 24-bit integers in ipio2[].
 *
 * Input parameters:
 * 	x[]	The input value (must be positive) is broken into nx
 *		pieces of 24-bit integers in double precision format.
 *		x[i] will be the i-th 24 bit of x. The scaled exponent
 *		of x[0] is given in input parameter e0 (i.e., x[0]*2^e0
 *		match x's up to 24 bits.
 *
 *		Example of breaking a double positive z into x[0]+x[1]+x[2]:
 *			e0 = ilogb(z)-23
 *			z  = scalbn(z,-e0)
 *		for i = 0,1,2
 *			x[i] = floor(z)
 *			z    = (z-x[i])*2**24
 *
 *
 *	y[]	ouput result in an array of double precision numbers.
 *		The dimension of y[] is:
 *			24-bit  precision	1
 *			53-bit  precision	2
 *			64-bit  precision	2
 *			113-bit precision	3
 *		The actual value is the sum of them. Thus for 113-bit
 *		precison, one may have to do something like:
 *
 *		long double t,w,r_head, r_tail;
 *		t = (long double)y[2] + (long double)y[1];
 *		w = (long double)y[0];
 *		r_head = t+w;
 *		r_tail = w - (r_head - t);
 *
 *	e0	The exponent of x[0]. Must be <= 16360 or you need to
 *              expand the ipio2 table.
 *
 *	nx	dimension of x[]
 *
 *  	prec	an integer indicating the precision:
 *			0	24  bits (single)
 *			1	53  bits (double)
 *			2	64  bits (extended)
 *			3	113 bits (quad)
 *
 * External function:
 *	double scalbn(), floor();
 *
 *
 * Here is the description of some local variables:
 *
 * 	jk	jk+1 is the initial number of terms of ipio2[] needed
 *		in the computation. The minimum and recommended value
 *		for jk is 3,4,4,6 for single, double, extended, and quad.
 *		jk+1 must be 2 larger than you might expect so that our
 *		recomputation test works. (Up to 24 bits in the integer
 *		part (the 24 bits of it that we compute) and 23 bits in
 *		the fraction part may be lost to cancelation before we
 *		recompute.)
 *
 * 	jz	local integer variable indicating the number of
 *		terms of ipio2[] used.
 *
 *	jx	nx - 1
 *
 *	jv	index for pointing to the suitable ipio2[] for the
 *		computation. In general, we want
 *			( 2^e0*x[0] * ipio2[jv-1]*2^(-24jv) )/8
 *		is an integer. Thus
 *			e0-3-24*jv >= 0 or (e0-3)/24 >= jv
 *		Hence jv = max(0,(e0-3)/24).
 *
 *	jp	jp+1 is the number of terms in PIo2[] needed, jp = jk.
 *
 * 	q[]	double array with integral value, representing the
 *		24-bits chunk of the product of x and 2/pi.
 *
 *	q0	the corresponding exponent of q[0]. Note that the
 *		exponent for q[i] would be q0-24*i.
 *
 *	PIo2[]	double precision array, obtained by cutting pi/2
 *		into 24 bits chunks.
 *
 *	f[]	ipio2[] in floating point
 *
 *	iq[]	integer array by breaking up q[] in 24-bits chunk.
 *
 *	fq[]	final product of x*(2/pi) in fq[0],..,fq[jk]
 *
 *	ih	integer. If >0 it indicates q[] is >= 0.5, hence
 *		it also indicates the *sign* of the result.
 *
 */


/*
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 */

static const int init_jk[] = {3, 4, 4, 6}; /* initial value for jk */

/*
 * Table of constants for 2/pi, 396 Hex digits (476 decimal) of 2/pi
 *
 *		integer array, contains the (24*i)-th to (24*i+23)-th
 *		bit of 2/pi after binary point. The corresponding
 *		floating value is
 *
 *			ipio2[i] * 2^(-24(i+1)).
 *
 * NB: This table must have at least (e0-3)/24 + jk terms.
 *     For quad precision (e0 <= 16360, jk = 6), this is 686.
 */
static const int32_t ipio2[] = {
0xA2F983, 0x6E4E44, 0x1529FC, 0x2757D1, 0xF534DD, 0xC0DB62,
0x95993C, 0x439041, 0xFE5163, 0xABDEBB, 0xC561B7, 0x246E3A,
0x424DD2, 0xE00649, 0x2EEA09, 0xD1921C, 0xFE1DEB, 0x1CB129,
0xA73EE8, 0x8235F5, 0x2EBB44, 0x84E99C, 0x7026B4, 0x5F7E41,
0x3991D6, 0x398353, 0x39F49C, 0x845F8B, 0xBDF928, 0x3B1FF8,
0x97FFDE, 0x05980F, 0xEF2F11, 0x8B5A0A, 0x6D1F6D, 0x367ECF,
0x27CB09, 0xB74F46, 0x3F669E, 0x5FEA2D, 0x7527BA, 0xC7EBE5,
0xF17B3D, 0x0739F7, 0x8A5292, 0xEA6BFB, 0x5FB11F, 0x8D5D08,
0x560330, 0x46FC7B, 0x6BABF0, 0xCFBC20, 0x9AF436, 0x1DA9E3,
0x91615E, 0xE61B08, 0x659985, 0x5F14A0, 0x68408D, 0xFFD880,
0x4D7327, 0x310606, 0x1556CA, 0x73A8C9, 0x60E27B, 0xC08C6B,

#if LDBL_MAX_EXP > 1024
#if LDBL_MAX_EXP > 16384
#error "ipio2 table needs to be expanded"
#endif
0x47C419, 0xC367CD, 0xDCE809, 0x2A8359, 0xC4768B, 0x961CA6,
0xDDAF44, 0xD15719, 0x053EA5, 0xFF0705, 0x3F7E33, 0xE832C2,
0xDE4F98, 0x327DBB, 0xC33D26, 0xEF6B1E, 0x5EF89F, 0x3A1F35,
0xCAF27F, 0x1D87F1, 0x21907C, 0x7C246A, 0xFA6ED5, 0x772D30,
0x433B15, 0xC614B5, 0x9D19C3, 0xC2C4AD, 0x414D2C, 0x5D000C,
0x467D86, 0x2D71E3, 0x9AC69B, 0x006233, 0x7CD2B4, 0x97A7B4,
0xD55537, 0xF63ED7, 0x1810A3, 0xFC764D, 0x2A9D64, 0xABD770,
0xF87C63, 0x57B07A, 0xE71517, 0x5649C0, 0xD9D63B, 0x3884A7,
0xCB2324, 0x778AD6, 0x23545A, 0xB91F00, 0x1B0AF1, 0xDFCE19,
0xFF319F, 0x6A1E66, 0x615799, 0x47FBAC, 0xD87F7E, 0xB76522,
0x89E832, 0x60BFE6, 0xCDC4EF, 0x09366C, 0xD43F5D, 0xD7DE16,
0xDE3B58, 0x929BDE, 0x2822D2, 0xE88628, 0x4D58E2, 0x32CAC6,
0x16E308, 0xCB7DE0, 0x50C017, 0xA71DF3, 0x5BE018, 0x34132E,
0x621283, 0x014883, 0x5B8EF5, 0x7FB0AD, 0xF2E91E, 0x434A48,
0xD36710, 0xD8DDAA, 0x425FAE, 0xCE616A, 0xA4280A, 0xB499D3,
0xF2A606, 0x7F775C, 0x83C2A3, 0x883C61, 0x78738A, 0x5A8CAF,
0xBDD76F, 0x63A62D, 0xCBBFF4, 0xEF818D, 0x67C126, 0x45CA55,
0x36D9CA, 0xD2A828, 0x8D61C2, 0x77C912, 0x142604, 0x9B4612,
0xC459C4, 0x44C5C8, 0x91B24D, 0xF31700, 0xAD43D4, 0xE54929,
0x10D5FD, 0xFCBE00, 0xCC941E, 0xEECE70, 0xF53E13, 0x80F1EC,
0xC3E7B3, 0x28F8C7, 0x940593, 0x3E71C1, 0xB3092E, 0xF3450B,
0x9C1288, 0x7B20AB, 0x9FB52E, 0xC29247, 0x2F327B, 0x6D550C,
0x90A772, 0x1FE76B, 0x96CB31, 0x4A1679, 0xE27941, 0x89DFF4,
0x9794E8, 0x84E6E2, 0x973199, 0x6BED88, 0x365F5F, 0x0EFDBB,
0xB49A48, 0x6CA467, 0x427271, 0x325D8D, 0xB8159F, 0x09E5BC,
0x25318D, 0x3974F7, 0x1C0530, 0x010C0D, 0x68084B, 0x58EE2C,
0x90AA47, 0x02E774, 0x24D6BD, 0xA67DF7, 0x72486E, 0xEF169F,
0xA6948E, 0xF691B4, 0x5153D1, 0xF20ACF, 0x339820, 0x7E4BF5,
0x6863B2, 0x5F3EDD, 0x035D40, 0x7F8985, 0x295255, 0xC06437,
0x10D86D, 0x324832, 0x754C5B, 0xD4714E, 0x6E5445, 0xC1090B,
0x69F52A, 0xD56614, 0x9D0727, 0x50045D, 0xDB3BB4, 0xC576EA,
0x17F987, 0x7D6B49, 0xBA271D, 0x296996, 0xACCCC6, 0x5414AD,
0x6AE290, 0x89D988, 0x50722C, 0xBEA404, 0x940777, 0x7030F3,
0x27FC00, 0xA871EA, 0x49C266, 0x3DE064, 0x83DD97, 0x973FA3,
0xFD9443, 0x8C860D, 0xDE4131, 0x9D3992, 0x8C70DD, 0xE7B717,
0x3BDF08, 0x2B3715, 0xA0805C, 0x93805A, 0x921110, 0xD8E80F,
0xAF806C, 0x4BFFDB, 0x0F9038, 0x761859, 0x15A562, 0xBBCB61,
0xB989C7, 0xBD4010, 0x04F2D2, 0x277549, 0xF6B6EB, 0xBB22DB,
0xAA140A, 0x2F2689, 0x768364, 0x333B09, 0x1A940E, 0xAA3A51,
0xC2A31D, 0xAEEDAF, 0x12265C, 0x4DC26D, 0x9C7A2D, 0x9756C0,
0x833F03, 0xF6F009, 0x8C402B, 0x99316D, 0x07B439, 0x15200C,
0x5BC3D8, 0xC492F5, 0x4BADC6, 0xA5CA4E, 0xCD37A7, 0x36A9E6,
0x9492AB, 0x6842DD, 0xDE6319, 0xEF8C76, 0x528B68, 0x37DBFC,
0xABA1AE, 0x3115DF, 0xA1AE00, 0xDAFB0C, 0x664D64, 0xB705ED,
0x306529, 0xBF5657, 0x3AFF47, 0xB9F96A, 0xF3BE75, 0xDF9328,
0x3080AB, 0xF68C66, 0x15CB04, 0x0622FA, 0x1DE4D9, 0xA4B33D,
0x8F1B57, 0x09CD36, 0xE9424E, 0xA4BE13, 0xB52333, 0x1AAAF0,
0xA8654F, 0xA5C1D2, 0x0F3F0B, 0xCD785B, 0x76F923, 0x048B7B,
0x721789, 0x53A6C6, 0xE26E6F, 0x00EBEF, 0x584A9B, 0xB7DAC4,
0xBA66AA, 0xCFCF76, 0x1D02D1, 0x2DF1B1, 0xC1998C, 0x77ADC3,
0xDA4886, 0xA05DF7, 0xF480C6, 0x2FF0AC, 0x9AECDD, 0xBC5C3F,
0x6DDED0, 0x1FC790, 0xB6DB2A, 0x3A25A3, 0x9AAF00, 0x9353AD,
0x0457B6, 0xB42D29, 0x7E804B, 0xA707DA, 0x0EAA76, 0xA1597B,
0x2A1216, 0x2DB7DC, 0xFDE5FA, 0xFEDB89, 0xFDBE89, 0x6C76E4,
0xFCA906, 0x70803E, 0x156E85, 0xFF87FD, 0x073E28, 0x336761,
0x86182A, 0xEABD4D, 0xAFE7B3, 0x6E6D8F, 0x396795, 0x5BBF31,
0x48D784, 0x16DF30, 0x432DC7, 0x356125, 0xCE70C9, 0xB8CB30,
0xFD6CBF, 0xA200A4, 0xE46C05, 0xA0DD5A, 0x476F21, 0xD21262,
0x845CB9, 0x496170, 0xE0566B, 0x015299, 0x375550, 0xB7D51E,
0xC4F133, 0x5F6E13, 0xE4305D, 0xA92E85, 0xC3B21D, 0x3632A1,
0xA4B708, 0xD4B1EA, 0x21F716, 0xE4698F, 0x77FF27, 0x80030C,
0x2D408D, 0xA0CD4F, 0x99A520, 0xD3A2B3, 0x0A5D2F, 0x42F9B4,
0xCBDA11, 0xD0BE7D, 0xC1DB9B, 0xBD17AB, 0x81A2CA, 0x5C6A08,
0x17552E, 0x550027, 0xF0147F, 0x8607E1, 0x640B14, 0x8D4196,
0xDEBE87, 0x2AFDDA, 0xB6256B, 0x34897B, 0xFEF305, 0x9EBFB9,
0x4F6A68, 0xA82A4A, 0x5AC44F, 0xBCF82D, 0x985AD7, 0x95C7F4,
0x8D4D0D, 0xA63A20, 0x5F57A4, 0xB13F14, 0x953880, 0x0120CC,
0x86DD71, 0xB6DEC9, 0xF560BF, 0x11654D, 0x6B0701, 0xACB08C,
0xD0C0B2, 0x485551, 0x0EFB1E, 0xC37295, 0x3B06A3, 0x3540C0,
0x7BDC06, 0xCC45E0, 0xFA294E, 0xC8CAD6, 0x41F3E8, 0xDE647C,
0xD8649B, 0x31BED9, 0xC397A4, 0xD45877, 0xC5E369, 0x13DAF0,
0x3C3ABA, 0x461846, 0x5F7555, 0xF5BDD2, 0xC6926E, 0x5D2EAC,
0xED440E, 0x423E1C, 0x87C461, 0xE9FD29, 0xF3D6E7, 0xCA7C22,
0x35916F, 0xC5E008, 0x8DD7FF, 0xE26A6E, 0xC6FDB0, 0xC10893,
0x745D7C, 0xB2AD6B, 0x9D6ECD, 0x7B723E, 0x6A11C6, 0xA9CFF7,
0xDF7329, 0xBAC9B5, 0x5100B7, 0x0DB2E2, 0x24BA74, 0x607DE5,
0x8AD874, 0x2C150D, 0x0C1881, 0x94667E, 0x162901, 0x767A9F,
0xBEFDFD, 0xEF4556, 0x367ED9, 0x13D9EC, 0xB9BA8B, 0xFC97C4,
0x27A831, 0xC36EF1, 0x36C594, 0x56A8D8, 0xB5A8B4, 0x0ECCCF,
0x2D8912, 0x34576F, 0x89562C, 0xE3CE99, 0xB920D6, 0xAA5E6B,
0x9C2A3E, 0xCC5F11, 0x4A0BFD, 0xFBF4E1, 0x6D3B8E, 0x2C86E2,
0x84D4E9, 0xA9B4FC, 0xD1EEEF, 0xC9352E, 0x61392F, 0x442138,
0xC8D91B, 0x0AFC81, 0x6A4AFB, 0xD81C2F, 0x84B453, 0x8C994E,
0xCC2254, 0xDC552A, 0xD6C6C0, 0x96190B, 0xB8701A, 0x649569,
0x605A26, 0xEE523F, 0x0F117F, 0x11B5F4, 0xF5CBFC, 0x2DBC34,
0xEEBC34, 0xCC5DE8, 0x605EDD, 0x9B8E67, 0xEF3392, 0xB817C9,
0x9B5861, 0xBC57E1, 0xC68351, 0x103ED8, 0x4871DD, 0xDD1C2D,
0xA118AF, 0x462C21, 0xD7F359, 0x987AD9, 0xC0549E, 0xFA864F,
0xFC0656, 0xAE79E5, 0x362289, 0x22AD38, 0xDC9367, 0xAAE855,
0x382682, 0x9BE7CA, 0xA40D51, 0xB13399, 0x0ED7A9, 0x480569,
0xF0B265, 0xA7887F, 0x974C88, 0x36D1F9, 0xB39221, 0x4A827B,
0x21CF98, 0xDC9F40, 0x5547DC, 0x3A74E1, 0x42EB67, 0xDF9DFE,
0x5FD45E, 0xA4677B, 0x7AACBA, 0xA2F655, 0x23882B, 0x55BA41,
0x086E59, 0x862A21, 0x834739, 0xE6E389, 0xD49EE5, 0x40FB49,
0xE956FF, 0xCA0F1C, 0x8A59C5, 0x2BFA94, 0xC5C1D3, 0xCFC50F,
0xAE5ADB, 0x86C547, 0x624385, 0x3B8621, 0x94792C, 0x876110,
0x7B4C2A, 0x1A2C80, 0x12BF43, 0x902688, 0x893C78, 0xE4C4A8,
0x7BDBE5, 0xC23AC4, 0xEAF426, 0x8A67F7, 0xBF920D, 0x2BA365,
0xB1933D, 0x0B7CBD, 0xDC51A4, 0x63DD27, 0xDDE169, 0x19949A,
0x9529A8, 0x28CE68, 0xB4ED09, 0x209F44, 0xCA984E, 0x638270,
0x237C7E, 0x32B90F, 0x8EF5A7, 0xE75614, 0x08F121, 0x2A9DB5,
0x4D7E6F, 0x5119A5, 0xABF9B5, 0xD6DF82, 0x61DD96, 0x023616,
0x9F3AC4, 0xA1A283, 0x6DED72, 0x7A8D39, 0xA9B882, 0x5C326B,
0x5B2746, 0xED3400, 0x7700D2, 0x55F4FC, 0x4D5901, 0x8071E0,
#endif

};

static const double PIo2[] = {
  1.57079625129699707031e+00, /* 0x3FF921FB, 0x40000000 */
  7.54978941586159635335e-08, /* 0x3E74442D, 0x00000000 */
  5.39030252995776476554e-15, /* 0x3CF84698, 0x80000000 */
  3.28200341580791294123e-22, /* 0x3B78CC51, 0x60000000 */
  1.27065575308067607349e-29, /* 0x39F01B83, 0x80000000 */
  1.22933308981111328932e-36, /* 0x387A2520, 0x40000000 */
  2.73370053816464559624e-44, /* 0x36E38222, 0x80000000 */
  2.16741683877804819444e-51, /* 0x3569F31D, 0x00000000 */
};

static const double
zero    =  0.00000000000000000000e+00, /* 0x00000000, 0x00000000 */
one     =  1.00000000000000000000e+00, /* 3FF00000, 00000000 */
two24   =  1.67772160000000000000e+07, /* 0x41700000, 0x00000000 */
twon24  =  5.96046447753906250000e-08; /* 0x3E700000, 0x00000000 */

/* __kernel_rem_pio2(x,y,e0,nx,prec)
 * double x[],y[]; int e0,nx,prec;
 *
 * __kernel_rem_pio2 return the last three digits of N with
 *		y = x - N*pi/2
 * so that |y| < pi/2.
 *
 * The method is to compute the integer (mod 8) and fraction parts of
 * (2/pi)*x without doing the full multiplication. In general we
 * skip the part of the product that are known to be a huge integer (
 * more accurately, = 0 mod 8 ). Thus the number of operations are
 * independent of the exponent of the input.
 *
 * (2/pi) is represented by an array of 24-bit integers in ipio2[].
 *
 * Input parameters:
 * 	x[]	The input value (must be positive) is broken into nx
 *		pieces of 24-bit integers in double precision format.
 *		x[i] will be the i-th 24 bit of x. The scaled exponent
 *		of x[0] is given in input parameter e0 (i.e., x[0]*2^e0
 *		match x's up to 24 bits.
 *
 *		Example of breaking a double positive z into x[0]+x[1]+x[2]:
 *			e0 = ilogb(z)-23
 *			z  = scalbn(z,-e0)
 *		for i = 0,1,2
 *			x[i] = floor(z)
 *			z    = (z-x[i])*2**24
 *
 *
 *	y[]	ouput result in an array of double precision numbers.
 *		The dimension of y[] is:
 *			24-bit  precision	1
 *			53-bit  precision	2
 *			64-bit  precision	2
 *			113-bit precision	3
 *		The actual value is the sum of them. Thus for 113-bit
 *		precison, one may have to do something like:
 *
 *		long double t,w,r_head, r_tail;
 *		t = (long double)y[2] + (long double)y[1];
 *		w = (long double)y[0];
 *		r_head = t+w;
 *		r_tail = w - (r_head - t);
 *
 *	e0	The exponent of x[0]. Must be <= 16360 or you need to
 *              expand the ipio2 table.
 *
 *	nx	dimension of x[]
 *
 *  	prec	an integer indicating the precision:
 *			0	24  bits (single)
 *			1	53  bits (double)
 *			2	64  bits (extended)
 *			3	113 bits (quad)
 *
 * External function:
 *	double scalbn(), floor();
 *
 *
 * Here is the description of some local variables:
 *
 * 	jk	jk+1 is the initial number of terms of ipio2[] needed
 *		in the computation. The minimum and recommended value
 *		for jk is 3,4,4,6 for single, double, extended, and quad.
 *		jk+1 must be 2 larger than you might expect so that our
 *		recomputation test works. (Up to 24 bits in the integer
 *		part (the 24 bits of it that we compute) and 23 bits in
 *		the fraction part may be lost to cancelation before we
 *		recompute.)
 *
 * 	jz	local integer variable indicating the number of
 *		terms of ipio2[] used.
 *
 *	jx	nx - 1
 *
 *	jv	index for pointing to the suitable ipio2[] for the
 *		computation. In general, we want
 *			( 2^e0*x[0] * ipio2[jv-1]*2^(-24jv) )/8
 *		is an integer. Thus
 *			e0-3-24*jv >= 0 or (e0-3)/24 >= jv
 *		Hence jv = max(0,(e0-3)/24).
 *
 *	jp	jp+1 is the number of terms in PIo2[] needed, jp = jk.
 *
 * 	q[]	double array with integral value, representing the
 *		24-bits chunk of the product of x and 2/pi.
 *
 *	q0	the corresponding exponent of q[0]. Note that the
 *		exponent for q[i] would be q0-24*i.
 *
 *	PIo2[]	double precision array, obtained by cutting pi/2
 *		into 24 bits chunks.
 *
 *	f[]	ipio2[] in floating point
 *
 *	iq[]	integer array by breaking up q[] in 24-bits chunk.
 *
 *	fq[]	final product of x*(2/pi) in fq[0],..,fq[jk]
 *
 *	ih	integer. If >0 it indicates q[] is >= 0.5, hence
 *		it also indicates the *sign* of the result.
 *
 */


/*
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 */

static int sun_rem_pio2_aux (double *x, double *y, int e0, int nx, int prec) {
  int32_t jz, jx, jv, jp, jk, carry, n, iq[20], i, j, k, m, q0, ih;
  double z, fw, f[20], fq[20], q[20];
  /* initialize jk*/
  jk = init_jk[prec];
  jp = jk;
  /* determine jx,jv,q0, note that 3>q0 */
  jx =  nx - 1;
  jv = (e0 - 3)/24; if (jv < 0) jv = 0;
  q0 =  e0 - 24*(jv + 1);
  /* set up f[0] to f[jx+jk] where f[jx+jk] = ipio2[jv+jk] */
  j = jv - jx; m = jx + jk;
  for (i=0; i <= m; i++, j++) f[i] = (j < 0) ? zero : (double)ipio2[j];
  /* compute q[0],q[1],...q[jk] */
  for (i=0; i <= jk; i++) {
    for (j=0, fw=0.0; j <= jx; j++) { fw += x[j]*f[jx + i - j]; }
    q[i] = fw;
  }
  jz = jk;
recompute:
  /* distill q[] into iq[] reversingly */
  for (i=0, j=jz, z=q[jz]; j > 0; i++, j--) {
    fw    =  (double)((int32_t)(twon24*z));
    iq[i] =  (int32_t)(z - two24*fw);
    z     =  q[j - 1] + fw;
  }
  /* compute n */
  z  = sun_scalbn(z, q0);       /* actual value of z */
  z -= 8.0*sun_floor(z*0.125);  /* trim off integer >= 8 */
  n  = (int32_t)z;
  z -= (double)n;
  ih = 0;
  if (q0 > 0) {  /* need iq[jz-1] to determine n */
    i  = (iq[jz - 1] >> (24 - q0)); n += i;
	iq[jz - 1] -= i << (24 - q0);
    ih = iq[jz - 1] >> (23 - q0);
  }
  else if (q0 == 0) ih = iq[jz - 1] >> 23;
  else if (z >= 0.5) ih = 2;
  if (ih > 0) {	/* q > 0.5 */
    n += 1; carry = 0;
    for (i=0; i < jz; i++) {  /* compute 1-q */
      j = iq[i];
      if (carry == 0) {
        if (j != 0) {
          carry = 1; iq[i] = 0x1000000 - j;
        }
      } else  iq[i] = 0xffffff - j;
    }
    if (q0 > 0) {  /* rare case: chance is 1 in 12 */
      switch(q0) {
        case 1:
          iq[jz - 1] &= 0x7fffff; break;
        case 2:
          iq[jz - 1] &= 0x3fffff; break;
        }
    }
    if (ih == 2) {
      z = one - z;
      if (carry != 0) z -= sun_scalbn(one, q0);
    }
  }
  /* check if recomputation is needed */
  if (z == zero) {
    j = 0;
    for (i=jz - 1; i >= jk; i--) j |= iq[i];
    if (j == 0) {  /* need recomputation */
      for (k=1; iq[jk - k] == 0; k++);    /* k = no. of terms needed */
      for (i=jz + 1; i <= jz + k; i++) {  /* add q[jz+1] to q[jz+k] */
        f[jx + i] = (double) ipio2[jv + i];
        for (j=0, fw=0.0; j <= jx; j++) fw += x[j]*f[jx + i - j];
        q[i] = fw;
      }
      jz += k;
      goto recompute;
    }
  }
  /* chop off zero terms */
  if (z == 0.0) {
    jz -= 1; q0 -= 24;
    while (iq[jz] == 0) { jz--; q0 -= 24; }
  } else {  /* break z into 24-bit if necessary */
    z = sun_scalbn(z, -q0);
    if (z >= two24) {
      fw = (double)((int32_t)(twon24*z));
      iq[jz] = (int32_t)(z - two24*fw);
      jz += 1; q0 += 24;
      iq[jz] = (int32_t)fw;
    } else iq[jz] = (int32_t)z;
  }
  /* convert integer "bit" chunk to floating-point value */
  fw = sun_scalbn(one, q0);
  for (i=jz; i >= 0; i--) {
    q[i] = fw*(double)iq[i]; fw *= twon24;
  }
  /* compute PIo2[0,...,jp]*q[jz,...,0] */
  for (i=jz; i >= 0; i--) {
    for (fw = 0.0, k = 0; k <= jp && k <= jz - i; k++) fw += PIo2[k]*q[i + k];
    fq[jz - i] = fw;
  }
  /* compress fq[] into y[] */
  switch (prec) {
    case 0:
      fw = 0.0;
      for (i=jz; i >= 0; i--) fw += fq[i];
      y[0] = (ih == 0)? fw : -fw;
      break;
    case 1:
    case 2:
      fw = 0.0;
      for (i=jz; i >=0; i--) fw += fq[i];
      __STRICT_ASSIGN(double, fw, fw);
      y[0] = (ih == 0) ? fw : -fw;
      fw = fq[0] - fw;
      for (i=1; i <= jz; i++) fw += fq[i];
      y[1] = (ih == 0) ? fw : -fw;
      break;
    case 3:	/* painful */
      for (i=jz; i > 0; i--) {
        fw      = fq[i - 1] + fq[i];
        fq[i]  += fq[i - 1] - fw;
        fq[i - 1] = fw;
      }
      for (i=jz; i > 1; i--) {
        fw      = fq[i - 1] + fq[i];
        fq[i]  += fq[i - 1] - fw;
        fq[i - 1] = fw;
      }
      for (fw=0.0, i=jz; i >= 2; i--) fw += fq[i];
      if (ih == 0) {
        y[0] =  fq[0]; y[1] =  fq[1]; y[2] =  fw;
      } else {
        y[0] = -fq[0]; y[1] = -fq[1]; y[2] = -fw;
      }
  }
  return n & 7;
}

static const double
invpio2 =  6.36619772367581382433e-01, /* 0x3FE45F30, 0x6DC9C883 */
pio2_1  =  1.57079632673412561417e+00, /* 0x3FF921FB, 0x54400000 */
pio2_1t =  6.07710050650619224932e-11, /* 0x3DD0B461, 0x1A626331 */
pio2_2  =  6.07710050630396597660e-11, /* 0x3DD0B461, 0x1A600000 */
pio2_2t =  2.02226624879595063154e-21, /* 0x3BA3198A, 0x2E037073 */
pio2_3  =  2.02226624871116645580e-21, /* 0x3BA3198A, 0x2E000000 */
pio2_3t =  8.47842766036889956997e-32; /* 0x397B839A, 0x252049C1 */

LUALIB_API int sun_rem_pio2 (double x, double *y) {  /* argument reduction */
  double z, w, t, r, fn;
  double tx[3], ty[2];
  int32_t e0, i, j, nx, n, ix, hx;
  uint32_t low;
  __GET_HI(hx, x);		/* high word of x */
  ix = hx & 0x7fffffff;
#if 0 /* Must be handled in caller. */
  if (ix <= 0x3fe921fb)   /* |x| ~<= pi/4 , no need for reduction */
    { y[0] = x; y[1] = 0; return 0; }
#endif
  if (ix <= 0x400f6a7a) {		/* |x| ~<= 5pi/4 */
    if ((ix & 0xfffff) == 0x921fb)  /* |x| ~= pi/2 or 2pi/2 */
      goto medium;		/* cancellation -- use medium case */
    if (ix <= 0x4002d97c) {	/* |x| ~<= 3pi/4 */
      if (hx > 0) {
        z = x - pio2_1;	/* one round good to 85 bits */
        y[0] = z - pio2_1t;
        y[1] = (z - y[0]) - pio2_1t;
        return 1;
      } else {
        z = x + pio2_1;
        y[0] = z + pio2_1t;
        y[1] = (z - y[0]) + pio2_1t;
        return -1;
      }
    } else {
      if (hx > 0) {
        z = x - 2*pio2_1;
        y[0] = z - 2*pio2_1t;
        y[1] = (z - y[0]) - 2*pio2_1t;
        return 2;
      } else {
        z = x + 2*pio2_1;
        y[0] = z + 2*pio2_1t;
        y[1] = (z - y[0]) + 2*pio2_1t;
        return -2;
      }
    }
  }
  if (ix <= 0x401c463b) {		/* |x| ~<= 9pi/4 */
    if (ix <= 0x4015fdbc) {	/* |x| ~<= 7pi/4 */
      if (ix == 0x4012d97c)	/* |x| ~= 3pi/2 */
        goto medium;
      if (hx > 0) {
        z = x - 3*pio2_1;
        y[0] = z - 3*pio2_1t;
        y[1] = (z - y[0]) - 3*pio2_1t;
        return 3;
      } else {
        z = x + 3*pio2_1;
        y[0] = z + 3*pio2_1t;
        y[1] = (z - y[0]) + 3*pio2_1t;
        return -3;
      }
    } else {
      if (ix == 0x401921fb)	/* |x| ~= 4pi/2 */
        goto medium;
      if (hx > 0) {
        z = x - 4*pio2_1;
        y[0] = z - 4*pio2_1t;
        y[1] = (z - y[0]) - 4*pio2_1t;
        return 4;
      } else {
        z = x + 4*pio2_1;
        y[0] = z + 4*pio2_1t;
        y[1] = (z - y[0]) + 4*pio2_1t;
        return -4;
      }
    }
  }
  if (ix < 0x413921fb) {	/* |x| ~< 2^20*(pi/2), medium size */
medium:
    /* Use a specialized rint() to get fn.  Assume round-to-nearest. */
    __STRICT_ASSIGN(double, fn, x*invpio2 + 0x1.8p52);
    fn = fn - 0x1.8p52;
    if (tools_isfloat(fn)) fn = rint(fn);  /* 2.11.0 extension for MinGW, Linux GCC does not need it,
    for there fn is already rounded to an integer */
    n  = rint(fn);
    r  = x - fn*pio2_1;
    w  = fn*pio2_1t;	/* 1st round good to 85 bit */
    {
      uint32_t high;
      j  = ix >> 20;
      y[0] = r - w;
      __GET_HI(high, y[0]);
      i = j - ((high >> 20) & 0x7ff);
      if (i > 16) {  /* 2nd iteration needed, good to 118 */
        t = r;
        w = fn*pio2_2;
        r = t - w;
        w = fn*pio2_2t - ((t - r) - w);
        y[0] = r - w;
        __GET_HI(high, y[0]);
        i = j - ((high >> 20) & 0x7ff);
        if (i > 49)  {	/* 3rd iteration need, 151 bits acc */
          t  = r;	/* will cover all possible cases */
          w  = fn*pio2_3;
          r  = t - w;
          w  = fn*pio2_3t - ((t - r) - w);
          y[0] = r - w;
        }
      }
    }
    y[1] = (r - y[0]) - w;
    return n;
  }
  /*
   * all other (large) arguments
   */
  if (ix >= 0x7ff00000) {		/* x is inf or NaN */
    y[0] = y[1] = x - x; return 0;
  }
  /* set z = scalbn(|x|,ilogb(x)-23) */
  __GET_LO(low, x);
  e0 	= (ix >> 20) - 1046;	/* e0 = ilogb(z)-23; */
  __INSERT_WORDS(z, ix - ((int32_t)(e0 << 20)), low);
  for (i=0; i < 2; i++) {
    tx[i] = (double)((int32_t)(z));
    z = (z - tx[i])*two24;
  }
  tx[2] = z;
  nx = 3;
  while (tx[nx - 1] == zero) nx--;	/* skip zero term */
  n = sun_rem_pio2_aux(tx, ty, e0, nx, 1);
  if (hx < 0) { y[0] = -ty[0]; y[1] = -ty[1]; return -n; }
  y[0] = ty[0]; y[1] = ty[1]; return n;
}


/* Constants used in polynomial approximation of sin/cos */
static const double
half =  5.00000000000000000000e-01, /* 0x3FE00000, 0x00000000 */
S1  = -1.66666666666666324348e-01, /* 0xBFC55555, 0x55555549 */
S2  =  8.33333333332248946124e-03, /* 0x3F811111, 0x1110F8A6 */
S3  = -1.98412698298579493134e-04, /* 0xBF2A01A0, 0x19C161D5 */
S4  =  2.75573137070700676789e-06, /* 0x3EC71DE3, 0x57B1FE7D */
S5  = -2.50507602534068634195e-08, /* 0xBE5AE5E6, 0x8A2B9CEB */
S6  =  1.58969099521155010221e-10, /* 0x3DE5D93A, 0x5ACFD57C */
C1  =  4.16666666666666019037e-02, /* 0x3FA55555, 0x5555554C */
C2  = -1.38888888888741095749e-03, /* 0xBF56C16C, 0x16C15177 */
C3  =  2.48015872894767294178e-05, /* 0x3EFA01A0, 0x19CB1590 */
C4  = -2.75573143513906633035e-07, /* 0xBE927E4F, 0x809C52AD */
C5  =  2.08757232129817482790e-09, /* 0x3E21EE9E, 0xBDB4B1C4 */
C6  = -1.13596475577881948265e-11; /* 0xBDA8FAE9, 0xBE8838D4 */

static INLINE void tools_sincos_aux (double x, double y, int iy, double *k_s, double *k_c ) {
  /* Inline calculation of sin/cos, as we can save
     some work, and we will always need to calculate
     both values, no matter the result of switch */
  double z, w, r, v, hz;
  z = x*x;
  w = z*z;
  /* cos-specific computation; equivalent to calling __kernel_cos(x,y) and storing in k_c*/
  r   = z*(C1 + z*(C2 + z*C3)) + w*w*(C4 + z*(C5 + z*C6));
  hz  = 0.5*z;
  v   = one - hz;
  *k_c = v + (((one - v) - hz) + (z*r - x*y));
  /* sin-specific computation; equivalent to calling __kernel_sin(x,y,1) and storing in k_s*/
  r = S2 + z*(S3 + z*S4) + z*w*(S5 + z*S6);
  v = z*x;
  if (iy == 0)
    *k_s = x + v*(S1 + z*r);
  else
    *k_s = x - ((z*(half*y - v*r) - y) - v*S1);
}

LUALIB_API void tools_sincos (double x, double *s, double *c) {  /* 35 % faster than calling sin and cos individually */
  double y[2];
  int32_t ix;
  /* Store high word of x in ix */
  __GET_HI(ix, x);
  /* |x| ~< pi/4 */
  ix &= 0x7fffffff;
  if (ix <= 0x3fe921fb) {
    /* Check for small x for sin and cos */
    if (ix < 0x3e46a09e) {
      /* Check for exact zero */
      if ( (int)x == 0 ) {
        *s = x;
        *c = 1.0;
        return;
      }
    }
    /* Call kernel function with 0 extra */
    tools_sincos_aux(x, 0.0, 0, s, c);
  } else if (ix >= 0x7ff00000) {
    /* sincos(Inf or NaN) is NaN */
    *s = x - x;
    *c = x - x;
  }
  /* argument reduction needed */
  else {
    double k_c, k_s;
    /* Calculate remainer, then sub out to kernel */
    int32_t n = sun_rem_pio2(x, y);
    tools_sincos_aux( y[0], y[1], 1, &k_s, &k_c );
    /* Figure out permutation of sin/cos outputs to true outputs */
    switch (n & 3) {
      case 0:
        *c = k_c;
        *s = k_s;
        break;
      case 1:
        *c = -k_s;
        *s = k_c;
        break;
      case 2:
        *c = -k_c;
        *s = -k_s;
        break;
      default:
        *c =  k_s;
        *s = -k_c;
        break;
    }
  }
}


LUALIB_API void tools_sinhcosh (double x, double *sih, double *coh) {
  *sih = sun_sinh(x);
  /* see page 85, Abramowitz and Stegun, Handbook of Mathematical Functions */
  *coh = sun_hypot2(*sih);  /* much faster than calling cosh */
}


/* Taken from: c-SSS package by Fletcher T. Penney, file GLibFacade.c, MIT licence.
 * The following section came from:
 *
 *	http://lists-archives.org/mingw-users/12649-asprintf-missing-vsnprintf-
 *		behaving-differently-and-_vsncprintf-undefined.html
 *
 * and
 *
 *	http://groups.google.com/group/jansson-users/browse_thread/thread/
 *		76a88d63d9519978/041a7d0570de2d48?lnk=raot
 */

/* Solaris and Windows do not provide vasprintf() or asprintf(). */
#if defined(_WIN32) || defined(__SOLARIS)
LUALIB_API int tools_vasprintf (char **sptr, char *fmt, va_list argv) {
  int wanted = vsnprintf( *sptr = NULL, 0, fmt, argv );
  if ((wanted > 0) && ((*sptr = malloc( 1 + wanted )) != NULL))
    return vsprintf( *sptr, fmt, argv );
  return wanted;
}

LUALIB_API int tools_asprintf (char **sptr, char *fmt, ... ) {
  int retval;
  va_list argv;
  va_start(argv, fmt);
  retval = tools_vasprintf(sptr, fmt, argv);
  va_end(argv);
  return retval;
}
#endif


/* Convert a Julian Date to Gregorian Date, 2.10.0 */
LUALIB_API int tools_auxcdate (double x, int *iy, int *im, int *id, double *fd, int *deg, int *min, int *sec) {
  long long int degrees, minutes, seconds;  /* do not use int or long int due to possible overflows */
  /* fd = decimal fractional day */
  if (iauJd2cal(x, 0, iy, im, id, fd) == -1) return -1;
  /* iauJd2cal returns fd with round-off errors, so ignore it */
  seconds = sun_floor(24*3600.0*(*fd) + 0.5);
  degrees = seconds / 3600;
  seconds = fabs(fmod(seconds, 3600.0));
  minutes = seconds / 60;
  seconds = fmod(seconds, 60);
  *deg = (int)degrees;
  *min = (int)minutes;
  *sec = (int)seconds;
  return (sun_floor(x - 0.5) == 60 + 2415018) ? -2 : 0;
}


/* https://stackoverflow.com/questions/18662261/fastest-implementation-of-sine-cosine-and-square-root-in-c-doesnt-need-to-b,
   answer by nimig18 */
static float cosOff4LUT[] =
  {  0x1.000000p+00,  0x1.6A09E6p-01, 0x0.000000p+00, -0x1.6A09E6p-01,
    -0x1.000000p+00, -0x1.6A09E6p-01, 0x0.000000p+00,  0x1.6A09E6p-01 };
LUALIB_API void tools_sincosfast (float x, float *pS, float *pC) {  /* 12 % faster than math.sincos */
  int m, ms, mc;
  float xI, xR, xR2, c, s, cy, sy;
  /* Cody & Waite's range reduction Algorithm, [-pi/4, pi/4] */
  xI  = floorf(x * 0x1.45F306p+00 + 0.5);  /* casting is not faster */
  xR  = (x - xI * 0x1.920000p-01) - xI*0x1.FB5444p-13;
  m   = (int)xI;
  xR2 = xR*xR;
  /* Find cosine & sine index for angle offsets indices */
  mc = (  m  ) & 0x7;     /* two's complement permits upper modulus for negative numbers =P */
  ms = (m + 6) & 0x7;     /* two's complement permits upper modulus for negative numbers =P, note phase correction for sine. */
  /* Find cosine & sine */
  cy = cosOff4LUT[mc];     /* Load angle offset neighborhood cosine value */
  sy = cosOff4LUT[ms];     /* Load angle offset neighborhood sine value */
  /* c = 0xf.ff79fp-4 + xR2 * (-0x7.e58e9p-4); */               /* TOL = 1.2786e-4 */
  c = 0xf.ffffdp-4 + xR2 * (-0x7.ffebep-4 + xR2 * 0xa.956a9p-8);  /* TOL = 1.7882e-7 */
  /* s = xR * (0xf.ffbf7p-4 + xR2 * (-0x2.a41d0cp-4)); */  /* TOL = 4.835251e-6 */
  s = xR * (0xf.fffffp-4 + xR2 * (-0x2.aaa65cp-4 + xR2 * 0x2.1ea25p-8));  /* TOL = 1.1841e-8 */
  *pC = c*cy - s*sy;
  *pS = c*sy + s*cy;
}


/* Returns an approximation of the sine of the given number x, in radians. Depending on the CPU used, it is 7 to 58 percent faster
  than sun_sin. 2.14.6; in the range -1000 .. 1000 with step size 0.1, mean error is 0.000444, median error is 0.000474. */

LUALIB_API double tools_sinfast (double x) {
  double_cast p, q;
  uint64_t sign;
  double half, apx;
  p.f = 0.22308510060189463;
  half = (x < 0) ? -0.5 : 0.5;
  x = (half + (int)(INVPI2*x))*PI2 - x;
  q.f = x;
  sign = q.i & 0x8000000000000000ULL;
  q.i &= 0x7FFFFFFFFFFFFFFFULL;  /* absolute value, only ULL hex will work with uint64_t */
  apx = x*(INVPIO4 - q.f*INVPISQO4);
  p.i |= sign;
  return apx*(0.77633023248007499 + apx*p.f);
}


/* Returns an approximation of the cosine of the given number x, in radians. Depending on the CPU used, it is 7 to 58 percent faster
  than sun_cos. 2.14.6 */

LUALIB_API double tools_cosfast (double x) {
  double_cast p, q;
  uint64_t sign;
  double half, apx;
  p.f = 0.22308510060189463;
  x += PIO2;  /* the rest is the same as in tools_sinfast, avoid extra function call */
  half = (x < 0) ? -0.5 : 0.5;
  x = (half + (int)(INVPI2*x))*PI2 - x;
  q.f = x;
  sign = q.i & 0x8000000000000000ULL;
  q.i &= 0x7FFFFFFFFFFFFFFFULL;  /* absolute value, only ULL hex will work with uint64_t */
  apx = x*(INVPIO4 - q.f*INVPISQO4);
  p.i |= sign;
  return apx*(0.77633023248007499 + apx*p.f);
}


LUALIB_API double tools_tanfast (double x) {  /* 2.14.6, depending on CPU, 7 to 58 percent faster than calling tools_sunscos */
  return tools_sinfast(x)/tools_cosfast(x);
}


/* https://stackoverflow.com/questions/18662261/fastest-implementation-of-sine-cosine-and-square-root-in-c-doesnt-need-to-b,
   answer by nimig18. In the range 0 .. 1000 with step size 0.1, median error is 0.00791 and mean error is 0.00829. */
LUALIB_API float tools_sqrtfast (float x) {  /* 10 percent _slower_ than GCC's sqrt builtin on Aspire V3-331. */
  ieee_float_shape_signedint X, Y;
  float ScOff;
  uint8_t e;
  X.f = x;
  e = (X.i >> 23);  /* f.SFPbits.e; */
  if (x < 0) return AGN_NAN;  /* 2.12.0 RC 3 change */
  else if (x == 0) return 0;
  ScOff = ((e & 1) != 0) ? 1.0f : 0x1.6a09e6p0;    /* NOTE: If exp=EVEN, b/c (exp-127) a (EVEN - ODD) := ODD; but a (ODD - ODD) := EVEN!! */
  e = (e + 127) >> 1;                              /* NOTE: If exp=ODD,  b/c (exp-127) then flr((exp-127)/2) */
  X.i = (X.i & ((1uL << 23) - 1)) | (0x7F << 23);  /* Mask mantissa, force exponent to zero. */
  Y.i = ((uint32_t) e) << 23;
  /* Error grows with square root of the exponent. Unfortunately no work around like inverse square root... :( */
  /* Y.f *= ScOff * (0x9.5f61ap-4 + X.f*(0x6.a09e68p-4));        // Error = +-1.78e-2 * 2^(flr(log2(x)/2)) */
  Y.f *= ScOff * (0x7.2181d8p-4 + X.f*(0xa.05406p-4 + X.f*(-0x1.23a14cp-4)));  /* Error = +-7.64e-5 * 2^(flr(log2(x)/2)) */
  /* Y.f *= ScOff * (0x5.f10e7p-4 + X.f*(0xc.8f2p-4 +X.f*(-0x2.e41a4cp-4 + X.f*(0x6.441e6p-8)))); */  /* Error =  8.21e-5 * 2^(flr(log2(x)/2)) */
  /* Y.f *= ScOff * (0x5.32eb88p-4 + X.f*(0xe.abbf5p-4 + X.f*(-0x5.18ee2p-4 + X.f*(0x1.655efp-4 + X.f*(-0x2.b11518p-8)))));  Error = +-9.92e-6 * 2^(flr(log2(x)/2)) */
  /* Y.f *= ScOff * (0x4.adde5p-4 + X.f*(0x1.08448cp0 + X.f*(-0x7.ae1248p-4 + X.f*(0x3.2cf7a8p-4 + X.f*(-0xc.5c1e2p-8 + X.f*(0x1.4b6dp-8))))));  Error = +-1.38e-6 * 2^(flr(log2(x)/2)) */
  // Y.f *= ScOff * (0x4.4a17fp-4 + X.f*(0x1.22d44p0 + X.f*(-0xa.972e8p-4 + X.f*(0x5.dd53fp-4 + X.f*(-0x2.273c08p-4 + X.f*(0x7.466cb8p-8 + X.f*(-0xa.ac00ep-12)))))));  Error = +-2.9e-7 * 2^(flr(log2(x)/2)) */
  /* Y.f *= ScOff * (0x3.fbb3e8p-4 + X.f*(0x1.3b2a3cp0 + X.f*(-0xd.cbb39p-4 + X.f*(0x9.9444ep-4 + X.f*(-0x4.b5ea38p-4 + X.f*(0x1.802f9ep-4 + X.f*(-0x4.6f0adp-8 + X.f*(0x5.c24a28p-12 ))))))));   Error = +-2.7e-6 * 2^(flr(log2(x)/2)) */
  return Y.f;
}


LUALIB_API float tools_sqrtapx (float x) {  /* 4 % off, taken from http://bits.stephan-brumme.com */
  ieee_float_shape_type u = { x };
  u.word  += 127 << 23;  /* adjust bias */
  u.word >>= 1;          /* approximation of square root */
  return u.value;
}


/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
*/
/*
FUNCTION
<<round>>, <<roundf>>---round to integer, to nearest
INDEX
        round
INDEX
        roundf

ANSI_SYNOPSIS
        #include <math.h>
        double round(double <[x]>);
        float roundf(float <[x]>);

DESCRIPTION
        The <<round>> functions round their argument to the nearest integer
        value in floating-point format, rounding halfway cases away from zero,
        regardless of the current rounding direction.  (While the "inexact"
        floating-point exception behaviour is unspecified by the C standard, the
        <<round>> functions are written so that "inexact" is not raised if the
        result does not equal the argument, which behaviour is as recommended
        by IEEE 754 for its related functions.)

RETURNS
		<[x]> rounded to an integral value.

PORTABILITY
		ANSI C, POSIX

SEEALSO
		<<nearbyint>>, <<rint>>

REMARK
		The openlibm version is slower. */

LUALIB_API double sun_round (double x) {
  /* Most significant word, least significant word. */
  int32_t msw, exponent_less_1023;
  uint32_t lsw;
  __EXTRACT_WORDS(msw, lsw, x);
  /* Extract exponent field. */
  exponent_less_1023 = ((msw & 0x7ff00000) >> 20) - 1023;
  if (exponent_less_1023 < 20) {
    if (exponent_less_1023 < 0) {
      msw &= 0x80000000;
      if (exponent_less_1023 == -1)
        /* Result is +1.0 or -1.0. */
        msw |= (1023 << 20);
      lsw = 0;
    } else {
      uint32_t exponent_mask = 0x000fffff >> exponent_less_1023;
      if ((msw & exponent_mask) == 0 && lsw == 0)
        /* x in an integral value. */
        return x;
      msw += 0x00080000 >> exponent_less_1023;
      msw &= ~exponent_mask;
      lsw = 0;
    }
  } else if (exponent_less_1023 > 51) {
    if (exponent_less_1023 == 1024)
      /* x is NaN or infinite. */
      return x + x;
    else
      return x;
  } else {
    uint32_t tmp, exponent_mask;
    exponent_mask = 0xffffffff >> (exponent_less_1023 - 20);
    if ((lsw & exponent_mask) == 0)
      /* x is an integral value. */
      return x;
    tmp = lsw + (1 << (51 - exponent_less_1023));
    if (tmp < lsw)
      msw += 1;
    lsw = tmp;
    lsw &= ~exponent_mask;
  }
  __INSERT_WORDS(x, msw, lsw);
  return x;
}


/* @(#)e_fmod.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/*
 * __ieee754_fmod(x,y)
 * Return x mod y in exact arithmetic
 * Method: shift and subtract
 */

static const double Zero[] = { 0.0, -0.0, };

LUALIB_API double sun_fmod (double x, double y) {  /* SLOWER than built-in fmod */
  int n, hx, hy, hz, ix, iy, sx, i;
  uint32_t lx, ly, lz;
  __EXTRACT_WORDS (hx, lx, x);
  __EXTRACT_WORDS (hy, ly, y);
  sx = hx & 0x80000000;  /* sign of x */
  hx ^= sx;          /* |x| */
  hy &= 0x7fffffff;  /* |y| */
  /* purge off exception values */
  if ((hy | ly) == 0 || (hx >= 0x7ff00000) ||	/* y=0, or x not finite */
     ((hy | ((ly | -ly) >> 31)) > 0x7ff00000))	/* or y is NaN */
     return (x*y)/(x*y);
  if (hx <= hy) {
    if ((hx < hy) || (lx < ly)) return x;  /* |x| < |y| return x */
    if (lx == ly)
      return Zero[(uint32_t)sx >> 31];     /* |x| = |y| return x*0*/
  }
  /* determine ix = ilogb(x) */
  if (unlikely(hx < 0x00100000)) {  /* subnormal x, 2.5.15 optimisation */
    if (hx == 0) {
      for (ix = -1043, i=lx; i > 0; i <<= 1) ix -=1;
	} else {
      for (ix=-1022, i=(hx << 11); i > 0; i <<= 1) ix -=1;
    }
  } else ix = (hx >> 20) - 1023;
  /* determine iy = ilogb(y) */
  if (unlikely(hy < 0x00100000)) {  /* subnormal y, 2.5.15 optimisation */
    if (hy == 0) {
      for (iy=-1043, i=ly; i > 0; i <<= 1) iy -=1;
    } else {
      for (iy=-1022, i=(hy << 11); i > 0; i <<= 1) iy -=1;
	}
  } else iy = (hy >> 20) - 1023;
  /* set up {hx,lx}, {hy,ly} and align y to x */
  if (likely(ix >= -1022))
    hx = 0x00100000 | (0x000fffff & hx);
  else {  /* subnormal x, shift x to normal */
    n = -1022 - ix;
    if (n <= 31) {
      hx = (hx << n) | (lx >> (32 - n));
      lx <<= n;
    } else {
      hx = lx << (n - 32);
      lx = 0;
    }
  }
  if (likely(iy >= -1022))
    hy = 0x00100000 | (0x000fffff & hy);
  else {  /* subnormal y, shift y to normal */
    n = -1022 - iy;
    if (n <= 31) {
      hy = (hy << n) | (ly >> (32 - n));
      ly <<= n;
    } else {
      hy = ly << (n - 32);
      ly = 0;
    }
  }
  /* fix point fmod */
  n = ix - iy;
  while (n--) {
    hz = hx - hy; lz = lx - ly; if (lx < ly) hz -= 1;
    if (hz < 0) { hx = hx + hx + (lx >> 31); lx = lx + lx; }
    else {
      if ((hz|lz) == 0) 		/* return sign(x)*0 */
        return Zero[(uint32_t)sx >> 31];
      hx = hz + hz + (lz >> 31); lx = lz + lz;
    }
  }
  hz = hx - hy; lz = lx - ly; if (lx < ly) hz -= 1;
  if (hz >= 0) { hx = hz; lx = lz; }
  /* convert back to floating value and restore the sign */
  if ((hx | lx) == 0) 			/* return sign(x)*0 */
    return Zero[(uint32_t)sx >> 31];
  while (hx < 0x00100000) {		/* normalize x */
    hx = hx + hx + (lx >> 31); lx = lx + lx;
    iy -= 1;
  }
  if (iy >= -1022) {	/* normalize output */
    hx = ((hx - 0x00100000) | ((iy + 1023) << 20));
    __INSERT_WORDS(x, hx | sx, lx);
  } else {		/* subnormal output */
    n = -1022 - iy;
    if (n <= 20) {
      lx = (lx >> n) | ((uint32_t)hx << (32 - n));
      hx >>= n;
    } else if (n <= 31) {
      lx = (hx << (32 - n)) | (lx >> n); hx = sx;
    } else {
      lx = hx >> (n - 32); hx = sx;
    }
    __INSERT_WORDS(x, hx | sx, lx);
    x *= one;		/* create necessary signal */
  }
  return x;		/* exact output */
}


/* @(#)s_modf.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/*
 * modf(double x, double *iptr)
 * return fraction part of x, and return x's integral part in *iptr.
 * Method:
 *	Bit twiddling.
 *
 * Exception:
 *	No exception.
 */
LUALIB_API double sun_modf (double x, double *iptr) {  /* the function sometimes is not faster than built-in modf */
  int32_t i0, i1, j0;
  uint32_t i;
  __EXTRACT_WORDS(i0, i1, x);
  j0 = ((i0 >> 20) & 0x7ff) - 0x3ff;  /* exponent of x */
  if (j0 < 20) {   /* integer part in high x */
    if (j0 < 0) {  /* |x| < 1 */
      __INSERT_WORDS(*iptr, i0 & 0x80000000, 0);  /* *iptr = +-0 */
      return x;
    } else {
      i = (0x000fffff) >> j0;
      if (((i0 & i) | i1) == 0) {  /* x is integral */
        uint32_t high;
        *iptr = x;
        __GET_HI(high, x);
        __INSERT_WORDS(x, high & 0x80000000, 0);  /* return +-0 */
        return x;
      } else {
        __INSERT_WORDS(*iptr, i0 & (~i), 0);
        return x - *iptr;
      }
    }
  } else if (j0 > 51) {  /* no fraction part */
    uint32_t high;
    if (unlikely(j0 == 0x400)) {   /* inf/NaN, 2.5.15 optimisation */
      *iptr = x;
      return 0.0 / x;
    }
    *iptr = x*one;
    __GET_HI(high, x);
    __INSERT_WORDS(x, high & 0x80000000, 0);  /* return +-0 */
    return x;
  } else {  /* fraction part in low x */
    i = ((uint32_t)(0xffffffff)) >> (j0 - 20);
    if ((i1 & i) == 0) {  /* x is integral */
      uint32_t high;
      *iptr = x;
      __GET_HI(high, x);
      __INSERT_WORDS(x, high & 0x80000000, 0);  /* return +-0 */
      return x;
    } else {
      __INSERT_WORDS(*iptr, i0, i1 & (~i));
      return x - *iptr;
    }
  }
}


LUALIB_API double sun_int (double x) {  /* get integral part in a double; depending on x, the result is either
  positive or negative; based on sun_modf, 2.14.9 */
  int32_t i0, i1, j0;
  uint32_t i;
  double iptr;
  __EXTRACT_WORDS(i0, i1, x);
  j0 = ((i0 >> 20) & 0x7ff) - 0x3ff;  /* exponent of x */
  if (j0 < 20) {   /* integer part in high x */
    if (j0 < 0) {  /* |x| < 1 */
      __INSERT_WORDS(iptr, i0 & 0x80000000, 0);  /* *iptr = +-0 */
      return iptr;
    } else {
      i = (0x000fffff) >> j0;
      if (((i0 & i) | i1) == 0) return x;  /* x is integral */
      __INSERT_WORDS(iptr, i0 & (~i), 0);
      return iptr;
    }
  } else if (j0 > 51) {  /* no fraction part */
    if (unlikely(j0 == 0x400)) return x;  /* inf/NaN, 2.5.15 optimisation */
    return x*one;
  } else {  /* fraction part in low x */
    i = ((uint32_t)(0xffffffff)) >> (j0 - 20);
    if ((i1 & i) == 0) return x;  /* x is integral */
    __INSERT_WORDS(iptr, i0, i1 & (~i));
    return iptr;
  }
}


#define treatintegral(x)  ( (x < 0) ? -(x) : (x) )

static INLINE double treatdecimal (double x, int32_t a, int32_t b) {
  double iptr;
  __INSERT_WORDS(iptr, a, b);
  return fabs(x - iptr);
}

LUALIB_API double sun_frac (double x) {  /* get fractional part in a double, will always be positive
  even if x is negative; based on sun_modf, 2.14.9 */
  uint32_t i;
  int32_t i0, i1, j0;
  __EXTRACT_WORDS(i0, i1, x);
  j0 = ((i0 >> 20) & 0x7ff) - 0x3ff;  /* exponent of x */
  if (j0 < 20) {  /* integer part in high x */
    if (j0 < 0) return treatintegral(x);  /* |x| < 1 */
    i = 0x000fffff >> j0;
    return (((i0 & i) | i1) == 0) ? treatintegral(x) : treatdecimal(x, i0 & (~i), 0);
  } else if (j0 < 52) {  /* fraction part in low x, j0 in 20 .. 51 */
    i = ((uint32_t)(0xffffffff)) >> (j0 - 20);
    return ((i1 & i) == 0) ? treatintegral(x) : treatdecimal(x, i0, i1 & (~i));
  } else  /* no fraction part, j0 > 51 */
    return (j0 != 0x400) ? treatintegral(x) : 0.0/x;
}


/* @(#)e_hypot.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */
LUALIB_API INLINE double sun_hypot (double x, double y) {  /* the function only sometimes is faster than built-in hypot; 1 % slower than tools_hypot when inlined */
  double a, b, t1, t2, y1, y2, w;
  int32_t j, k, ha, hb;
  __GET_HI(ha, x);
  ha &= 0x7fffffff;
  __GET_HI(hb, y);
  hb &= 0x7fffffff;
  if (hb > ha) { a = y; b = x; j = ha; ha = hb; hb = j; } else { a = x; b = y; }
  a = fabs(a);  /* fabs is faster than __SET_HI(a, ha) */
  b = fabs(b);  /* dito */
  if ((ha - hb) > 0x3c00000) return a + b;  /* x/y > 2**60 */
  k = 0;
  if (ha > 0x5f300000) {     /* a>2**500 */
    if (unlikely(ha >= 0x7ff00000)) {  /* Inf or NaN, 2.5.15 optimisation */
      uint32_t low;
      /* Use original arg order iff result is NaN; quieten sNaNs. */
      w = fabs(x + 0.0) - fabs(y + 0.0);
      __GET_LO(low, a);
      if (((ha & 0xfffff) | low) == 0) w = a;
        __GET_LO(low, b);
        if (((hb ^ 0x7ff00000) | low) == 0) w = b;
        return w;
    }
    /* scale a and b by 2**-600 */
    ha -= 0x25800000; hb -= 0x25800000; k += 600;
    __SET_HI(a, ha);
    __SET_HI(b, hb);
  }
  if (unlikely(hb < 0x20b00000)) {	/* b < 2**-500, 2.5.15 optimisation */
    if (hb <= 0x000fffff) {	/* subnormal b or 0 */
      uint32_t low;
      __GET_LO(low, b);
      if ((hb | low) == 0) return a;
      t1 = 0;
      __SET_HI(t1, 0x7fd00000);  /* t1=2^1022 */
      b *= t1;
      a *= t1;
      k -= 1022;
    } else {             /* scale a and b by 2^600 */
      ha += 0x25800000;  /* a *= 2^600 */
      hb += 0x25800000;  /* b *= 2^600 */
      k -= 600;
      __SET_HI(a, ha);
      __SET_HI(b, hb);
    }
  }
  /* medium size a and b */
  w = a - b;
  if (w > b) {
    t1 = 0;
    __SET_HI(t1, ha);
    t2 = a - t1;
    w  = sqrt(t1*t1 - (b*(-b) - t2*(a + t1)));
  } else {
    a  = a + a;
    y1 = 0;
    __SET_HI(y1, hb);
    y2 = b - y1;
    t1 = 0;
    __SET_HI(t1, ha + 0x00100000);
    t2 = a - t1;
    w  = sqrt(t1*y1 - (w*(-w) - (t1*y2 + t2*b)));
  }
  if (k != 0) {
    uint32_t high;
    t1 = 1.0;
    __GET_HI(high, t1);
    __SET_HI(t1, high + (k << 20));
    return t1*w;
  } else
    return w;
}


LUALIB_API double sun_hypot2 (double y) {  /* 2.14.13, sqrt(1 + x^2) */
  double a, b, t1, t2, y1, y2, w;
  int32_t j, k, ha, hb;
  ha = 0x3ff00000;
  __GET_HI(hb, y);
  hb &= 0x7fffffff;
  if (hb > ha) { a = fabs(y); b = 1; j = ha; ha = hb; hb = j; } else { a = 1; b = fabs(y); }
  if ((ha - hb) > 0x3c00000) return a + b;  /* x/y > 2**60 */
  k = 0;
  if (unlikely(ha > 0x5f300000)) {     /* a>2**500, 2.5.15 optimisation */
    if (ha >= 0x7ff00000) {  /* Inf or NaN */
      uint32_t low;
      /* Use original arg order iff result is NaN; quieten sNaNs. */
      w = 1 - fabs(y + 0.0);
      __GET_LO(low, a);
      if (((ha & 0xfffff) | low) == 0) w = a;
        __GET_LO(low, b);
        if (((hb ^ 0x7ff00000) | low) == 0) w = b;
        return w;
    }
    /* scale a and b by 2**-600 */
    ha -= 0x25800000; hb -= 0x25800000; k += 600;
    __SET_HI(a, ha);
    __SET_HI(b, hb);
  }
  if (unlikely(hb < 0x20b00000)) {     /* b < 2**-500, 2.5.15 optimisation */
    if (hb <= 0x000fffff) {  /* subnormal b or 0 */
      uint32_t low;
      __GET_LO(low, b);
      if ((hb | low) == 0) return a;
      t1 = 0;
      __SET_HI(t1, 0x7fd00000);  /* t1=2^1022 */
      b *= t1;
      a *= t1;
      k -= 1022;
    } else {             /* scale a and b by 2^600 */
      ha += 0x25800000;  /* a *= 2^600 */
      hb += 0x25800000;  /* b *= 2^600 */
      k -= 600;
      __SET_HI(a, ha);
      __SET_HI(b, hb);
    }
  }
  /* medium size a and b */
  w = a - b;
  if (w > b) {
    t1 = 0;
    __SET_HI(t1, ha);
    t2 = a - t1;
    w  = sqrt(t1*t1 - (b*(-b) - t2*(a + t1)));
  } else {
    a  = a + a;
    y1 = 0;
    __SET_HI(y1, hb);
    y2 = b - y1;
    t1 = 0;
    __SET_HI(t1, ha + 0x00100000);
    t2 = a - t1;
    w  = sqrt(t1*y1 - (w*(-w) - (t1*y2 + t2*b)));
  }
  if (k != 0) {
    uint32_t high;
    t1 = 1.0;
    __GET_HI(high, t1);
    __SET_HI(t1, high + (k << 20));
    return t1*w;
  } else
    return w;
}


LUALIB_API double sun_hypot3 (double y) {  /* 2.14.13, sqrt(1 - x^2), 0.5 % slower than tools_hypot3 */
  double a, b, t1, t2, w;
  int32_t k, ha, hb;
  uint32_t lb;
  ha = 0x3ff00000;
  __EXTRACT_WORDS(hb, lb, y);
  hb &= 0x7fffffff;
  if ((hb | ((lb | (-lb)) >> 31)) > 0x3ff00000) return AGN_NAN;  /* |x| > 1 */
  if (hb > ha) { a = fabs(y); b = 1; ha = hb; } else { a = 1; b = fabs(y); }
  if ((ha - hb) > 0x3c00000) return a + b;  /* x/y > 2**60 */
  k = 0;
  if (unlikely(ha > 0x5f300000)) {     /* a>2**500, 2.5.15 optimisation */
    if (ha >= 0x7ff00000) {  /* Inf or NaN */
      uint32_t low;
      /* Use original arg order iff result is NaN; quieten sNaNs. */
      w = 1 - fabs(y + 0.0);
      __GET_LO(low, a);
      if (((ha & 0xfffff) | low) == 0) w = a;
        __GET_LO(low, b);
        if (((hb ^ 0x7ff00000) | low) == 0) w = b;
        return w;
    }
    /* scale a and b by 2**-600 */
    ha -= 0x25800000; hb -= 0x25800000; k += 600;
    __SET_HI(a, ha);
    __SET_HI(b, hb);
  }
  if (unlikely(hb < 0x20b00000)) {     /* b < 2**-500, 2.5.15 optimisation */
    if (hb <= 0x000fffff) {  /* subnormal b or 0 */
      uint32_t low;
      __GET_LO(low, b);
      if ((hb | low) == 0) return a;
      t1 = 0;
      __SET_HI(t1, 0x7fd00000);	/* t1=2^1022 */
      b *= t1;
      a *= t1;
      k -= 1022;
    } else {             /* scale a and b by 2^600 */
      ha += 0x25800000;  /* a *= 2^600 */
      hb += 0x25800000;  /* b *= 2^600 */
      k -= 600;
      __SET_HI(a, ha);
      __SET_HI(b, hb);
    }
  }
  /* medium size a and b */
  w = a - b;
  t1 = 0;
  __SET_HI(t1, ha);
  t2 = a - t1;
  return sqrt(t1*t1 + (b*(-b) - t2*(a + t1)));
}


/* @(#)e_atan2.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 *
 */

/* __ieee754_atan2(y,x)
 * Method :
 *	1. Reduce y to positive by atan2(y,x)=-atan2(-y,x).
 *	2. Reduce x to positive by (if x and y are unexceptional):
 *		ARG (x+iy) = arctan(y/x)   	   ... if x > 0,
 *		ARG (x+iy) = pi - arctan[y/(-x)]   ... if x < 0,
 *
 * Special cases:
 *
 *	ATAN2((anything), NaN ) is NaN;
 *	ATAN2(NAN , (anything) ) is NaN;
 *	ATAN2(+-0, +(anything but NaN)) is +-0  ;
 *	ATAN2(+-0, -(anything but NaN)) is +-pi ;
 *	ATAN2(+-(anything but 0 and NaN), 0) is +-pi/2;
 *	ATAN2(+-(anything but INF and NaN), +INF) is +-0 ;
 *	ATAN2(+-(anything but INF and NaN), -INF) is +-pi;
 *	ATAN2(+-INF,+INF ) is +-pi/4 ;
 *	ATAN2(+-INF,-INF ) is +-3pi/4;
 *	ATAN2(+-INF, (anything but,0,NaN, and INF)) is +-pi/2;
 *
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 */

static const double
pi_o_4  = 7.8539816339744827900E-01, /* 0x3FE921FB, 0x54442D18 */
pi_o_2  = 1.5707963267948965580E+00, /* 0x3FF921FB, 0x54442D18 */
pi      = 3.1415926535897931160E+00, /* 0x400921FB, 0x54442D18 */
huge    = 1.0e300;

/* @(#)s_atan.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* atan(x)
 * Method
 *   1. Reduce x to positive by atan(x) = -atan(-x).
 *   2. According to the integer k=4t+0.25 chopped, t=x, the argument
 *      is further reduced to one of the following intervals and the
 *      arctangent of t is evaluated by the corresponding formula:
 *
 *      [0,7/16]      atan(x) = t-t^3*(a1+t^2*(a2+...(a10+t^2*a11)...)
 *      [7/16,11/16]  atan(x) = atan(1/2) + atan( (t-0.5)/(1+t/2) )
 *      [11/16.19/16] atan(x) = atan( 1 ) + atan( (t-1)/(1+t) )
 *      [19/16,39/16] atan(x) = atan(3/2) + atan( (t-1.5)/(1+1.5t) )
 *      [39/16,INF]   atan(x) = atan(INF) + atan( -1/t )
 *
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 *
 * The function is half as fast as GCC's atanh.
 *
 */

static const double atanhi[] = {
  4.63647609000806093515e-01, /* atan(0.5)hi 0x3FDDAC67, 0x0561BB4F */
  7.85398163397448278999e-01, /* atan(1.0)hi 0x3FE921FB, 0x54442D18 */
  9.82793723247329054082e-01, /* atan(1.5)hi 0x3FEF730B, 0xD281F69B */
  1.57079632679489655800e+00, /* atan(inf)hi 0x3FF921FB, 0x54442D18 */
};

static const double atanlo[] = {
  2.26987774529616870924e-17, /* atan(0.5)lo 0x3C7A2B7F, 0x222F65E2 */
  3.06161699786838301793e-17, /* atan(1.0)lo 0x3C81A626, 0x33145C07 */
  1.39033110312309984516e-17, /* atan(1.5)lo 0x3C700788, 0x7AF0CBBD */
  6.12323399573676603587e-17, /* atan(inf)lo 0x3C91A626, 0x33145C07 */
};

static const double aT[] = {
  3.33333333333329318027e-01, /* 0x3FD55555, 0x5555550D */
 -1.99999999998764832476e-01, /* 0xBFC99999, 0x9998EBC4 */
  1.42857142725034663711e-01, /* 0x3FC24924, 0x920083FF */
 -1.11111104054623557880e-01, /* 0xBFBC71C6, 0xFE231671 */
  9.09088713343650656196e-02, /* 0x3FB745CD, 0xC54C206E */
 -7.69187620504482999495e-02, /* 0xBFB3B0F2, 0xAF749A6D */
  6.66107313738753120669e-02, /* 0x3FB10D66, 0xA0D03D51 */
 -5.83357013379057348645e-02, /* 0xBFADDE2D, 0x52DEFD9A */
  4.97687799461593236017e-02, /* 0x3FA97B4B, 0x24760DEB */
 -3.65315727442169155270e-02, /* 0xBFA2B444, 0x2C6A6C2F */
  1.62858201153657823623e-02, /* 0x3F90AD3A, 0xE322DA11 */
};

LUALIB_API double sun_atan (double x) {
  double w, s1, s2, z;
  int32_t ix, hx, id;
  __GET_HI(hx, x);
  ix = hx & 0x7fffffff;
  if (ix >= 0x44100000) {  /* if |x| >= 2^66 */
    uint32_t low;
    __GET_LO(low, x);
    if (ix > 0x7ff00000 ||
       (ix == 0x7ff00000 && (low != 0)))
       return x + x;  /* NaN */
    if (hx > 0) return atanhi[3] + *(volatile double *)&atanlo[3];
    else return -atanhi[3] - *(volatile double *)&atanlo[3];
  }
  if (ix < 0x3fdc0000) {	/* |x| < 0.4375 */
    if (ix < 0x3e400000) {	/* |x| < 2^-27 */
      if (huge + x > one) return x;	/* raise inexact */
    }
    id = -1;
  } else {
    x = fabs(x);
    if (ix < 0x3ff30000) {    /* |x| < 1.1875 */
      if (ix < 0x3fe60000) {  /* 7/16 <=|x|<11/16 */
        id = 0; x = (2.0*x - one)/(2.0 + x);
      } else {  /* 11/16<=|x|< 19/16 */
        id = 1; x = (x - one)/(x + one);
      }
    } else {
      if (ix < 0x40038000) {  /* |x| < 2.4375 */
        id = 2; x = (x - 1.5)/(one + 1.5*x);
      } else {  /* 2.4375 <= |x| < 2^66 */
        id = 3; x = -1.0/x;
      }
    }
  }
  /* end of argument reduction */
  z = x*x;
  w = z*z;
  /* break sum from i=0 to 10 aT[i]z**(i+1) into odd and even poly */
  s1 = z*(aT[0] + w*(aT[2] + w*(aT[4] + w*(aT[6] + w*(aT[8] + w*aT[10])))));
  s2 = w*(aT[1] + w*(aT[3] + w*(aT[5] + w*(aT[7] + w*aT[9]))));
  if (id < 0) return x - x*(s1 + s2);
  else {
    z = atanhi[id] - ((x*(s1 + s2) - atanlo[id]) - x);
    return (hx < 0) ? -z : z;
  }
}


LUALIB_API double sun_atan2 (double y, double x) {
  double z;
  int32_t k, m, hx, hy, ix, iy;
  uint32_t lx, ly;
  __EXTRACT_WORDS(hx, lx, x);
  ix = hx & 0x7fffffff;
  __EXTRACT_WORDS(hy, ly, y);
  iy = hy & 0x7fffffff;
  if (unlikely(((ix | ((lx | -lx) >> 31)) > 0x7ff00000) ||  /* 2.5.15 optimisation */
     ((iy | ((ly | -ly) >> 31)) > 0x7ff00000)))	/* x or y is NaN */
    return x + y;
  if (((hx - 0x3ff00000) | lx) == 0) return sun_atan(y);   /* x=1.0 */
  m = ((hy >> 31) & 1) | ((hx >> 30) & 2);	/* 2*sign(x)+sign(y) */
  /* when y = 0 */
  if ((iy | ly) == 0) {
    switch(m) {
      case 0:
      case 1: return y; 	      /* atan(+-0,+anything)=+-0 */
      case 2: return  pi + tiny;  /* atan(+0,-anything) = pi */
      case 3: return -pi - tiny;  /* atan(-0,-anything) =-pi */
    }
  }
  /* when x = 0 */
  if ((ix | lx) == 0) return (hy < 0) ? -pi_o_2 - tiny : pi_o_2 + tiny;
  /* when x is INF */
  if (unlikely(ix == 0x7ff00000)) {  /* 2.5.15 optimisation */
    if (iy == 0x7ff00000) {
      switch(m) {
        case 0: return  pi_o_4 + tiny;      /* atan(+INF,+INF) */
        case 1: return -pi_o_4 - tiny;      /* atan(-INF,+INF) */
        case 2: return  3.0*pi_o_4 + tiny;  /*atan(+INF,-INF)*/
        case 3: return -3.0*pi_o_4 - tiny;  /*atan(-INF,-INF)*/
      }
    } else {
      switch(m) {
        case 0: return  zero  ;	/* atan(+...,+INF) */
        case 1: return -zero  ;	/* atan(-...,+INF) */
        case 2: return  pi + tiny  ;	/* atan(+...,-INF) */
        case 3: return -pi - tiny  ;	/* atan(-...,-INF) */
      }
    }
  }
  /* when y is INF */
  if (unlikely(iy == 0x7ff00000)) return (hy < 0) ? -pi_o_2 - tiny : pi_o_2 + tiny;  /* 2.5.15 optimisation */
  /* compute y/x */
  k = (iy - ix) >> 20;
  if (k > 60) {		 	/* |y/x| >  2**60 */
    z = pi_o_2 + 0.5*pi_lo;
    m &= 1;
  }
  else if (hx < 0 && k <- 60) z = 0.0; 	/* 0 > |y|/x > -2**-60 */
  else z = sun_atan(fabs(y/x));		/* safe to do y/x */
  switch (m) {
    case 0: return z;	/* atan(+,+) */
    case 1: return -z;	/* atan(-,+) */
    case 2: return pi - (z - pi_lo);/* atan(+,-) */
    default: /* case 3 */
      return (z - pi_lo) - pi;/* atan(-,-) */
  }
}


/* @(#)k_sin.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* __kernel_sin( x, y, iy)
 * kernel sin function on ~[-pi/4, pi/4] (except on -0), pi/4 ~ 0.7854
 * Input x is assumed to be bounded by ~pi/4 in magnitude.
 * Input y is the tail of x.
 * Input iy indicates whether y is 0. (if iy=0, y assume to be 0).
 *
 * Algorithm
 *	1. Since sin(-x) = -sin(x), we need only to consider positive x.
 *	2. Callers must return sin(-0) = -0 without calling here since our
 *	   odd polynomial is not evaluated in a way that preserves -0.
 *	   Callers may do the optimisation sin(x) ~ x for tiny x.
 *	3. sin(x) is approximated by a polynomial of degree 13 on
 *	   [0,pi/4]
 *		  	         3            13
 *	   	sin(x) ~ x + S1*x + ... + S6*x
 *	   where
 *
 * 	|sin(x)         2     4     6     8     10     12  |     -58
 * 	|----- - (1+S1*x +S2*x +S3*x +S4*x +S5*x  +S6*x   )| <= 2
 * 	|  x 					           |
 *
 *	4. sin(x+y) = sin(x) + sin'(x')*y
 *		    ~ sin(x) + (1-x*x/2)*y
 *	   For better accuracy, let
 *		     3      2      2      2      2
 *		r = x *(S2+x *(S3+x *(S4+x *(S5+x *S6))))
 *	   then                   3    2
 *		sin(x) = x + (S1*x + (x *(r-y/2)+y))
 */

static INLINE double kernel_sin (double x, double y, int iy) {  /* inlined 2.16.13 */
  double z, r, v, w;
  z	=  x*x;
  w	=  z*z;
  r	=  S2 + z*(S3 + z*S4) + z*w*(S5 + z*S6);
  v	=  z*x;
  if (iy == 0) return x + v*(S1 + z*r);
  else return x - ((z*(half*y - v*r) - y) - v*S1);
}

/* @(#)k_cos.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/*
 * __kernel_cos( x,  y )
 * kernel cos function on [-pi/4, pi/4], pi/4 ~ 0.785398164
 * Input x is assumed to be bounded by ~pi/4 in magnitude.
 * Input y is the tail of x.
 *
 * Algorithm
 *	1. Since cos(-x) = cos(x), we need only to consider positive x.
 *	2. if x < 2^-27 (hx<0x3e400000 0), return 1 with inexact if x!=0.
 *	3. cos(x) is approximated by a polynomial of degree 14 on
 *	   [0,pi/4]
 *		  	                 4            14
 *	   	cos(x) ~ 1 - x*x/2 + C1*x + ... + C6*x
 *	   where the remez error is
 *
 * 	|              2     4     6     8     10    12     14 |     -58
 * 	|cos(x)-(1-.5*x +C1*x +C2*x +C3*x +C4*x +C5*x  +C6*x  )| <= 2
 * 	|    					               |
 *
 * 	               4     6     8     10    12     14
 *	4. let r = C1*x +C2*x +C3*x +C4*x +C5*x  +C6*x  , then
 *	       cos(x) ~ 1 - x*x/2 + r
 *	   since cos(x+y) ~ cos(x) - sin(x)*y
 *			  ~ cos(x) - x*y,
 *	   a correction term is necessary in cos(x) and hence
 *		cos(x+y) = 1 - (x*x/2 - (r - x*y))
 *	   For better accuracy, rearrange to
 *		cos(x+y) ~ w + (tmp + (r-x*y))
 *	   where w = 1 - x*x/2 and tmp is a tiny correction term
 *	   (1 - x*x/2 == w + tmp exactly in infinite precision).
 *	   The exactness of w + tmp in infinite precision depends on w
 *	   and tmp having the same precision as x.  If they have extra
 *	   precision due to compiler bugs, then the extra precision is
 *	   only good provided it is retained in all terms of the final
 *	   expression for cos().  Retention happens in all cases tested
 *	   under FreeBSD, so don't pessimize things by forcibly clipping
 *	   any extra precision in w.
 */

static INLINE double kernel_cos (double x, double y) {  /* inlined 2.16.13 */
  double hz, z, r, w;
  z  = x*x;
  w  = z*z;
  r  = z*(C1 + z*(C2 + z*C3)) + w*w*(C4 + z*(C5 + z*C6));
  hz = 0.5*z;
  w  = one - hz;
  return w + (((one - w) - hz) + (z*r - x*y));
}

/* @(#)s_sin.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* sin(x)
 * Return sine function of x.
 *
 * kernel function:
 *	__kernel_sin		... sine function on [-pi/4,pi/4]
 *	__kernel_cos		... cose function on [-pi/4,pi/4]
 *	__ieee754_rem_pio2	... argument reduction routine
 *
 * Method.
 *      Let S,C and T denote the sin, cos and tan respectively on
 *	[-PI/4, +PI/4]. Reduce the argument x to y1+y2 = x-k*pi/2
 *	in [-pi/4 , +pi/4], and let n = k mod 4.
 *	We have
 *
 *          n        sin(x)      cos(x)        tan(x)
 *     ----------------------------------------------------------
 *	    0	       S	   C		 T
 *	    1	       C	  -S		-1/T
 *	    2	      -S	  -C		 T
 *	    3	      -C	   S		-1/T
 *     ----------------------------------------------------------
 *
 * Special cases:
 *      Let trig be any of sin, cos, or tan.
 *      trig(+-INF)  is NaN, with signals;
 *      trig(NaN)    is that NaN;
 *
 * Accuracy:
 *	TRIG(x) returns trig(x) nearly rounded
 */

LUALIB_API double sun_sin (double x) {
  double y[2], z = 0.0;
  int32_t n, ix;
  __GET_HI(ix, x);   /* High word of x. */
  ix &= 0x7fffffff;  /* |x| ~< pi/4 */
  if (ix <= 0x3fe921fb) {
    if (ix < 0x3e500000) {  /* |x| < 2**-26 */
      if ((int) x == 0) return x;
    }	/* generate inexact */
    return kernel_sin(x, z, 0);
  }
  /* sin(Inf or NaN) is NaN */
  else if (unlikely(ix >= 0x7ff00000)) return x - x;  /* 2.5.15 optimisation */
  /* argument reduction needed */
  else {
    n = sun_rem_pio2(x, y);
    switch (n & 3) {
      case 0: return kernel_sin(y[0], y[1], 1);
      case 1: return kernel_cos(y[0], y[1]);
      case 2: return -kernel_sin(y[0], y[1], 1);
      default:
        return -kernel_cos(y[0], y[1]);
    }
  }
}

/* @(#)s_cos.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* cos(x)
 * Return cosine function of x.
 *
 * kernel function:
 *	__kernel_sin		... sine function on [-pi/4,pi/4]
 *	__kernel_cos		... cosine function on [-pi/4,pi/4]
 *	__ieee754_rem_pio2	... argument reduction routine
 *
 * Method.
 *      Let S,C and T denote the sin, cos and tan respectively on
 *	[-PI/4, +PI/4]. Reduce the argument x to y1+y2 = x-k*pi/2
 *	in [-pi/4 , +pi/4], and let n = k mod 4.
 *	We have
 *
 *          n        sin(x)      cos(x)        tan(x)
 *     ----------------------------------------------------------
 *	    0	       S	   C		 T
 *	    1	       C	  -S		-1/T
 *	    2	      -S	  -C		 T
 *	    3	      -C	   S		-1/T
 *     ----------------------------------------------------------
 *
 * Special cases:
 *      Let trig be any of sin, cos, or tan.
 *      trig(+-INF)  is NaN, with signals;
 *      trig(NaN)    is that NaN;
 *
 * Accuracy:
 *	TRIG(x) returns trig(x) nearly rounded
 */

LUALIB_API double sun_cos (double x) {
  double y[2], z = 0.0;
  int32_t n, ix;
  /* High word of x. */
  __GET_HI(ix, x);
  /* |x| ~< pi/4 */
  ix &= 0x7fffffff;
  if (ix <= 0x3fe921fb) {
    if (ix < 0x3e46a09e)			/* if x < 2**-27 * sqrt(2) */
      if (((int)x) == 0) return 1.0;	/* generate inexact */
    return kernel_cos(x, z);
  }
  /* cos(Inf or NaN) is NaN */
  else if (unlikely(ix >= 0x7ff00000)) return x - x;  /* 2.5.15 optimisation */
  /* argument reduction needed */
  else {
    n = sun_rem_pio2(x, y);
    switch(n & 3) {
      case 0: return  kernel_cos(y[0], y[1]);
      case 1: return -kernel_sin(y[0], y[1], 1);
      case 2: return -kernel_cos(y[0], y[1]);
      default:
        return kernel_sin(y[0], y[1], 1);
    }
  }
}


/* @(#)e_log.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* __ieee754_log(x)
 * Return the logrithm of x
 *
 * Method :
 *   1. Argument Reduction: find k and f such that
 *			x = 2^k * (1+f),
 *	   where  sqrt(2)/2 < 1+f < sqrt(2) .
 *
 *   2. Approximation of log(1+f).
 *	Let s = f/(2+f) ; based on log(1+f) = log(1+s) - log(1-s)
 *		 = 2s + 2/3 s**3 + 2/5 s**5 + .....,
 *	     	 = 2s + s*R
 *      We use a special Reme algorithm on [0,0.1716] to generate
 * 	a polynomial of degree 14 to approximate R The maximum error
 *	of this polynomial approximation is bounded by 2**-58.45. In
 *	other words,
 *		        2      4      6      8      10      12      14
 *	    R(z) ~ Lg1*s +Lg2*s +Lg3*s +Lg4*s +Lg5*s  +Lg6*s  +Lg7*s
 *  	(the values of Lg1 to Lg7 are listed in the program)
 *	and
 *	    |      2          14          |     -58.45
 *	    | Lg1*s +...+Lg7*s    -  R(z) | <= 2
 *	    |                             |
 *	Note that 2s = f - s*f = f - hfsq + s*hfsq, where hfsq = f*f/2.
 *	In order to guarantee error in log below 1ulp, we compute log
 *	by
 *		log(1+f) = f - s*(f - R)	(if f is not too large)
 *		log(1+f) = f - (hfsq - s*(hfsq+R)).	(better accuracy)
 *
 *	3. Finally,  log(x) = k*ln2 + log(1+f).
 *			    = k*ln2_hi+(f-(hfsq-(s*(hfsq+R)+k*ln2_lo)))
 *	   Here ln2 is split into two floating point number:
 *			ln2_hi + ln2_lo,
 *	   where n*ln2_hi is always exact for |n| < 2000.
 *
 * Special cases:
 *	log(x) is NaN with signal if x < 0 (including -INF) ;
 *	log(+INF) is +INF; log(0) is -INF with signal;
 *	log(NaN) is that NaN with no signal.
 *
 * Accuracy:
 *	according to an error analysis, the error is always less than
 *	1 ulp (unit in the last place).
 *
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 */

static const double
ln2_hi  =  6.93147180369123816490e-01,	/* 3fe62e42 fee00000 */
ln2_lo  =  1.90821492927058770002e-10,	/* 3dea39ef 35793c76 */
two54   =  1.80143985094819840000e+16,  /* 43500000 00000000 */
Lg1 = 6.666666666666735130e-01,  /* 3FE55555 55555593 */
Lg2 = 3.999999999940941908e-01,  /* 3FD99999 9997FA04 */
Lg3 = 2.857142874366239149e-01,  /* 3FD24924 94229359 */
Lg4 = 2.222219843214978396e-01,  /* 3FCC71C5 1D8E78AF */
Lg5 = 1.818357216161805012e-01,  /* 3FC74664 96CB03DE */
Lg6 = 1.531383769920937332e-01,  /* 3FC39A09 D078C69F */
Lg7 = 1.479819860511658591e-01;  /* 3FC2F112 DF3E5244 */

LUALIB_API INLINE double sun_log (double x) {
  double hfsq, f, s, z, R, w, t1, t2, dk;
  int32_t k, hx, i, j;
  uint32_t lx;
  __EXTRACT_WORDS(hx, lx, x);
  k = 0;
  if (hx < 0x00100000) {  /* x < 2**-1022  */
    if (unlikely( (((hx & 0x7fffffff) | lx) == 0) || (hx < 0)))  /* 2.14.13 change, do not return -inf for 0 args any longer, 2.5.15 optimisation */
      return AGN_NAN; /* AGN_NAN is much faster than: (x - x)/zero;	*/ /* log(+-0 && log(-#) = NaN */
    k -= 54; x *= two54; /* subnormal number, scale up x */
    __GET_HI(hx, x);
  }
  if (hx >= 0x7ff00000) return x + x;
  k += (hx >> 20) - 1023;
  hx &= 0x000fffff;
  i = (hx + 0x95f64) & 0x100000;
  __SET_HI(x, hx |(i ^ 0x3ff00000));	/* normalize x or x/2 */
  k += (i >> 20);
  f = x - 1.0;
  if ((0x000fffff & (2 + hx)) < 3) {	/* -2**-20 <= f < 2**-20 */
    if (f == zero) {
      if (k == 0) {
        return zero;
      } else {
        dk = (double)k;
        return dk*ln2_hi + dk*ln2_lo;
      }
    }
    R = f*f*(0.5 - 0.33333333333333333*f);
    if (k == 0) return f - R;
    else {
      dk = (double)k;
      return dk*ln2_hi - ((R - dk*ln2_lo) - f);
    }
  }
  s = f/(2.0 + f);
  dk = (double)k;
  z = s*s;
  i = hx - 0x6147a;
  w = z*z;
  j = 0x6b851 - hx;
  t1= w*(Lg2 + w*(Lg4 + w*Lg6));
  t2= z*(Lg1 + w*(Lg3 + w*(Lg5 + w*Lg7)));
  i |= j;
  R = t2 + t1;
  if (i > 0) {
    hfsq = 0.5*f*f;
    if (k == 0) return f - (hfsq - s*(hfsq + R)); else
      return dk*ln2_hi - ((hfsq - (s*(hfsq + R) + dk*ln2_lo)) - f);
  } else {
    if (k == 0) return f-s*(f - R); else
      return dk*ln2_hi - ((s*(f - R) - dk*ln2_lo) - f);
  }
}


LUALIB_API double tools_logbase (double x, double b) {  /* 2.14.2 */
  if (x == 1 && b < 0) return 0;
  else if (x == b && (x < 0 || x > 1)) return 1;
  else if (x <= 0 || b <= 0 || b == 1) return AGN_NAN;
  else return sun_log(x)/sun_log(b);
}


LUALIB_API float tools_lnfast (float x) {  /* 2.14.6
  18 percent faster than sun_log on Aspire V3-331 (Intel(R) Pentium(R) 3556U @ 1.70GHz); the function is quite `crude` at
  integral x and its vicinity. Mean and median error is 1.35e-2 in the range 0.1 .. 1000, step 0.1 */
  ieee_float_shape_type u;
  u.value = x;
  x = 8.2629582881927490e-8f * (float)u.word;
  return x - 87.989971088f;
}


/* @(#)e_exp.c 1.6 04/04/22 */
/*
 * ====================================================
 * Copyright (C) 2004 by Sun Microsystems, Inc. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* __ieee754_exp(x)
 * Returns the exponential of x.
 *
 * Method
 *   1. Argument reduction:
 *      Reduce x to an r so that |r| <= 0.5*ln2 ~ 0.34658.
 *	Given x, find r and integer k such that
 *
 *               x = k*ln2 + r,  |r| <= 0.5*ln2.
 *
 *      Here r will be represented as r = hi-lo for better
 *	accuracy.
 *
 *   2. Approximation of exp(r) by a special rational function on
 *	the interval [0,0.34658]:
 *	Write
 *	    R(r**2) = r*(exp(r)+1)/(exp(r)-1) = 2 + r*r/6 - r**4/360 + ...
 *      We use a special Remes algorithm on [0,0.34658] to generate
 * 	a polynomial of degree 5 to approximate R. The maximum error
 *	of this polynomial approximation is bounded by 2**-59. In
 *	other words,
 *	    R(z) ~ 2.0 + P1*z + P2*z**2 + P3*z**3 + P4*z**4 + P5*z**5
 *  	(where z=r*r, and the values of P1 to P5 are listed below)
 *	and
 *	    |                  5          |     -59
 *	    | 2.0+P1*z+...+P5*z   -  R(z) | <= 2
 *	    |                             |
 *	The computation of exp(r) thus becomes
 *                             2*r
 *		exp(r) = 1 + -------
 *		              R - r
 *                                 r*R1(r)
 *		       = 1 + r + ----------- (for better accuracy)
 *		                  2 - R1(r)
 *	where
 *			         2       4             10
 *		R1(r) = r - (P1*r  + P2*r  + ... + P5*r   ).
 *
 *   3. Scale back to obtain exp(x):
 *	From step 1, we have
 *	   exp(x) = 2^k * exp(r)
 *
 * Special cases:
 *	exp(INF) is INF, exp(NaN) is NaN;
 *	exp(-INF) is 0, and
 *	for finite argument, only exp(0)=1 is exact.
 *
 * Accuracy:
 *	according to an error analysis, the error is always less than
 *	1 ulp (unit in the last place).
 *
 * Misc. info.
 *	For IEEE double
 *	    if x >  7.09782712893383973096e+02 then exp(x) overflow
 *	    if x < -7.45133219101941108420e+02 then exp(x) underflow
 *
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 */

static const double
halF[2]	= {0.5,-0.5,},
o_threshold=  7.09782712893383973096e+02,  /* 0x40862E42, 0xFEFA39EF */
u_threshold= -7.45133219101941108420e+02,  /* 0xc0874910, 0xD52D3051 */
ln2HI[2]   ={ 6.93147180369123816490e-01,  /* 0x3fe62e42, 0xfee00000 */
             -6.93147180369123816490e-01,},/* 0xbfe62e42, 0xfee00000 */
ln2LO[2]   ={ 1.90821492927058770002e-10,  /* 0x3dea39ef, 0x35793c76 */
             -1.90821492927058770002e-10,},/* 0xbdea39ef, 0x35793c76 */
invln2 =  1.44269504088896338700e+00, /* 0x3ff71547, 0x652b82fe */
P1   =  1.66666666666666019037e-01, /* 0x3FC55555, 0x5555553E */
P2   = -2.77777777770155933842e-03, /* 0xBF66C16C, 0x16BEBD93 */
P3   =  6.61375632143793436117e-05, /* 0x3F11566A, 0xAF25DE2C */
P4   = -1.65339022054652515390e-06, /* 0xBEBBBD41, 0xC5D26BF1 */
P5   =  4.13813679705723846039e-08; /* 0x3E663769, 0x72BEA4D0 */

LUALIB_API INLINE double sun_exp (double x) {	/* default IEEE double exp */
  double y, hi = 0.0, lo = 0.0, c, t, twopk;
  int32_t k = 0, xsb;
  uint32_t hx;
  __GET_HI(hx, x);
  xsb = (hx >> 31) & 1;  /* sign bit of x */
  hx &= 0x7fffffff;      /* high word of |x| */
  /* filter out non-finite argument */
  if (hx >= 0x40862E42) {  /* if |x|>=709.78... */
    if (hx >= 0x7ff00000) {
      uint32_t lx;
      __GET_LO(lx, x);
      if (((hx & 0xfffff)|lx) != 0)
        return x + x; 		/* NaN */
      else
        return (xsb == 0) ? x : 0.0;	/* exp(+-inf)={inf,0} */
    }
    if (x > o_threshold) return huge*huge; /* overflow */
    if (x < u_threshold) return twom1000*twom1000; /* underflow */
  }
  /* this implementation gives 2.7182818284590455 for exp(1.0),
     which is well within the allowable error. however,
     2.718281828459045 is closer to the true value so we prefer that
     answer, given that 1.0 is such an important argument value. */
  if (x == 1.0)
    return 2.718281828459045235360;
  /* argument reduction */
  if (hx > 0x3fd62e42) {    /* if  |x| > 0.5 ln2 */
    if (hx < 0x3FF0A2B2) {  /* and |x| < 1.5 ln2 */
      hi = x - ln2HI[xsb]; lo = ln2LO[xsb]; k = 1 - xsb - xsb;
    } else {
      k  = (int)(invln2*x + halF[xsb]);
      t  = k;
      hi = x - t*ln2HI[0];  /* t*ln2HI is exact here */
      lo = t*ln2LO[0];
    }
    __STRICT_ASSIGN(double, x, hi - lo);
  }
  else if (unlikely(hx < 0x3e300000))  {	/* when |x|<2**-28, 2.5.15 optimisation */
    if (huge + x > one) return one + x;/* trigger inexact */
  }
  else k = 0;
  /* x is now in primary range */
  t  = x*x;
  if (k >= -1021)
    __INSERT_WORDS(twopk, 0x3ff00000 + (k << 20), 0);
  else
    __INSERT_WORDS(twopk, 0x3ff00000 + ((k + 1000) << 20), 0);
  c  = x - t*(P1 + t*(P2 + t*(P3 + t*(P4 + t*P5))));
  if (k == 0) return one - ((x*c)/(c - 2.0) - x);
  else y = one - ((lo - (x*c)/(2.0 - c)) - hi);
  if (k >= -1021) {
    if (k == 1024) return y*2.0*0x1p1023;
    return y*twopk;
  } else {
    return y*twopk*twom1000;
  }
}


/* Approximates 2^x with enough precission (mean error = 3.2037e-7) for -infinity < x < 10.
   For x > 10, accuracy is quite bad. */
LUALIB_API float tools_pow2fast (float x) {  /* 2.14.6 */
  float clipx, y;
  ieee_float_shape_type u;
  clipx = (x < -126) ? -126.0f : x;
  y = clipx - (int)clipx + (float)(x < 0);
  u.word = (uint32_t)((1 << 23)*(clipx + 121.2740575f + 27.7280233f/(4.84252568f - y) - 1.49012907f * y));
  return u.value;
}


/* @(#)k_tan.c 1.5 04/04/22 SMI */

/*
 * ====================================================
 * Copyright 2004 Sun Microsystems, Inc.  All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* __kernel_tan( x, y, k )
 * kernel tan function on ~[-pi/4, pi/4] (except on -0), pi/4 ~ 0.7854
 * Input x is assumed to be bounded by ~pi/4 in magnitude.
 * Input y is the tail of x.
 * Input k indicates whether tan (if k = 1) or -1/tan (if k = -1) is returned.
 *
 * Algorithm
 *	1. Since tan(-x) = -tan(x), we need only to consider positive x.
 *	2. Callers must return tan(-0) = -0 without calling here since our
 *	   odd polynomial is not evaluated in a way that preserves -0.
 *	   Callers may do the optimisation tan(x) ~ x for tiny x.
 *	3. tan(x) is approximated by a odd polynomial of degree 27 on
 *	   [0,0.67434]
 *		  	         3             27
 *	   	tan(x) ~ x + T1*x + ... + T13*x
 *	   where
 *
 * 	        |tan(x)         2     4            26   |     -59.2
 * 	        |----- - (1+T1*x +T2*x +.... +T13*x    )| <= 2
 * 	        |  x 					|
 *
 *	   Note: tan(x+y) = tan(x) + tan'(x)*y
 *		          ~ tan(x) + (1+x*x)*y
 *	   Therefore, for better accuracy in computing tan(x+y), let
 *		     3      2      2       2       2
 *		r = x *(T2+x *(T3+x *(...+x *(T12+x *T13))))
 *	   then
 *		 		    3    2
 *		tan(x+y) = x + (T1*x + (x *(r+y)+y))
 *
 *      4. For x in [0.67434,pi/4],  let y = pi/4 - x, then
 *		tan(x) = tan(pi/4-y) = (1-tan(y))/(1+tan(y))
 *		       = 1 - 2*(tan(y) - (tan(y)^2)/(1+tan(y)))
 */

static const double xxx[] = {
 3.33333333333334091986e-01,	/* 3FD55555, 55555563 */
 1.33333333333201242699e-01,	/* 3FC11111, 1110FE7A */
 5.39682539762260521377e-02,	/* 3FABA1BA, 1BB341FE */
 2.18694882948595424599e-02,	/* 3F9664F4, 8406D637 */
 8.86323982359930005737e-03,	/* 3F8226E3, E96E8493 */
 3.59207910759131235356e-03,	/* 3F6D6D22, C9560328 */
 1.45620945432529025516e-03,	/* 3F57DBC8, FEE08315 */
 5.88041240820264096874e-04,	/* 3F4344D8, F2F26501 */
 2.46463134818469906812e-04,	/* 3F3026F7, 1A8D1068 */
 7.81794442939557092300e-05,	/* 3F147E88, A03792A6 */
 7.14072491382608190305e-05,	/* 3F12B80F, 32F0A7E9 */
-1.85586374855275456654e-05,	/* BEF375CB, DB605373 */
 2.59073051863633712884e-05,	/* 3EFB2A70, 74BF7AD4 */
/* one */	 1.00000000000000000000e+00,	/* 3FF00000, 00000000 */
/* pio4 */	 7.85398163397448278999e-01,	/* 3FE921FB, 54442D18 */
/* pio4lo */	 3.06161699786838301793e-17	/* 3C81A626, 33145C07 */
};
/* #define	one	xxx[13] */
#define	pio4	xxx[14]
#define	pio4lo	xxx[15]
#define T	xxx

static INLINE double kernel_tan (double x, double y, int iy) {  /* inlined 2.16.13 */
  double z, r, v, w, s;
  int32_t ix, hx;
  __GET_HI(hx, x);
  ix = hx & 0x7fffffff;	   /* high word of |x| */
  if (ix >= 0x3FE59428) {  /* |x| >= 0.6744 */
    if (hx < 0) {
      x = -x;
      y = -y;
    }
    z = pio4 - x;
    w = pio4lo - y;
    x = z + w;
    y = 0.0;
  }
  z = x * x;
  w = z * z;
  /*
   * Break x^5*(T[1]+x^2*T[2]+...) into
   * x^5(T[1]+x^4*T[3]+...+x^20*T[11]) +
   * x^5(x^2*(T[2]+x^4*T[4]+...+x^22*[T12]))
   */
  r = T[1] + w * (T[3] + w * (T[5] + w * (T[7] + w * (T[9] +
      w * T[11]))));
  v = z * (T[2] + w * (T[4] + w * (T[6] + w * (T[8] + w * (T[10] +
      w * T[12])))));
  s = z * x;
  r = y + z * (s * (r + v) + y);
  r += T[0] * s;
  w = x + r;
  if (ix >= 0x3FE59428) {
    v = (double) iy;
    return (double) (1 - ((hx >> 30) & 2)) *
      (v - 2.0 * (x - (w * w / (w + v) - r)));
  }
  if (iy == 1)
    return w;
  else {
    /*
     * if allow error up to 2 ulp, simply return
     * -1.0 / (x+r) here
     */
    /* compute -1.0 / (x+r) accurately */
    double a, t;
    z = w;
    __SET_LO(z, 0);
    v = r - (z - x);	/* z+v = r+x */
    t = a = -1.0 / w;	/* a = -1.0/w */
    __SET_LO(t, 0);
    s = 1.0 + t * z;
    return t + a * (s + t * v);
  }
}

/* @(#)s_tan.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */


/* tan(x)
 * Return tangent function of x.
 *
 * kernel function:
 *	__kernel_tan		... tangent function on [-pi/4,pi/4]
 *	__ieee754_rem_pio2	... argument reduction routine
 *
 * Method.
 *      Let S,C and T denote the sin, cos and tan respectively on
 *	[-PI/4, +PI/4]. Reduce the argument x to y1+y2 = x-k*pi/2
 *	in [-pi/4 , +pi/4], and let n = k mod 4.
 *	We have
 *
 *          n        sin(x)      cos(x)        tan(x)
 *     ----------------------------------------------------------
 *	    0	       S	   C		 T
 *	    1	       C	  -S		-1/T
 *	    2	      -S	  -C		 T
 *	    3	      -C	   S		-1/T
 *     ----------------------------------------------------------
 *
 * Special cases:
 *      Let trig be any of sin, cos, or tan.
 *      trig(+-INF)  is NaN, with signals;
 *      trig(NaN)    is that NaN;
 *
 * Accuracy:
 *	TRIG(x) returns trig(x) nearly rounded
 */

LUALIB_API double sun_tan (double x) {
  double y[2], z = 0.0;
  int32_t n, ix;
  /* High word of x. */
  __GET_HI(ix, x);
  /* |x| ~< pi/4 */
  ix &= 0x7fffffff;
  if (ix <= 0x3fe921fb) {
    if (ix < 0x3e400000)			/* x < 2**-27 */
      if ((int)x == 0) return x;		/* generate inexact */
    return kernel_tan(x, z, 1);
  }
  /* tan(Inf or NaN) is NaN */
  else if (ix >= 0x7ff00000) return x - x;		/* NaN */
  /* argument reduction needed */
  else {
    n = sun_rem_pio2(x, y);
    return kernel_tan(y[0], y[1], 1 - ((n & 1) << 1)); /*   1 -- n even
							-1 -- n odd */
  }
}

/* @(#)s_fabs.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/*
 * fabs(x) returns the absolute value of x.
 */

LUALIB_API double sun_fabs (double x) {  /* as fast as GCC's fabs */
  uint32_t high;
  __GET_HI(high, x);
  __SET_HI(x, high & 0x7fffffff);
  return x;
}


/* @(#)s_logb.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/*
 * double logb(x)
 * IEEE 754 logb. Included to pass IEEE test suite. Not recommend.
 * Use ilogb instead.
 */

LUALIB_API double sun_logb (double x) {  /* NOT tested, not faster than GCC's logb */
  int32_t lx, ix;
  __EXTRACT_WORDS(ix, lx, x);
  ix &= 0x7fffffff;  /* high |x| */
  if ((ix | lx) == 0) return -1.0/fabs(x);     /* x == 0 */
  if (unlikely(ix >= 0x7ff00000)) return x*x;  /* inf, nan, 2.5.15 optimisation */
  if (unlikely(ix < 0x00100000)) {  /* subnormal, 2.5.15 optimisation */
    x *= two54;           /* convert to normal */
    __GET_HI(ix, x);
    ix &= 0x7fffffff;
    return (double) ((ix >> 20) - 1023 - 54);
  } else
    return (double) ((ix >> 20) - 1023);
}


/* @(#)e_sinh.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* __ieee754_sinh(x)
 * Method :
 * mathematically sinh(x) if defined to be (exp(x)-exp(-x))/2
 *	1. Replace x by |x| (sinh(-x) = -sinh(x)).
 *	2.
 *		                                    E + E/(E+1)
 *	    0        <= x <= 22     :  sinh(x) := --------------, E=expm1(x)
 *			       			        2
 *
 *	    22       <= x <= lnovft :  sinh(x) := exp(x)/2
 *	    lnovft   <= x <= ln2ovft:  sinh(x) := exp(x/2)/2 * exp(x/2)
 *	    ln2ovft  <  x	    :  sinh(x) := x*shuge (overflow)
 *
 * Special cases:
 *	sinh(x) is |x| if x is +INF, -INF, or NaN.
 *	only sinh(0)=0 is exact for finite x.
 */

static const double shuge = 1.0e307;

LUALIB_API double sun_sinh (double x) {
  double t, h, w;
  int32_t ix, jx;
  /* High word of |x|. */
  __GET_HI(jx, x);
  ix = jx & 0x7fffffff;
  /* x is INF or NaN */
  if (unlikely(ix >= 0x7ff00000)) return x + x;  /* 2.5.15 optimisation */
  h = 0.5;
  if (jx < 0) h = -h;
  /* |x| in [0,22], return sign(x)*0.5*(E+E/(E+1))) */
  if (ix < 0x40360000) {		/* |x|<22 */
    if (ix < 0x3e300000) 		/* |x|<2**-28 */
      if (shuge + x > one) return x;/* sinh(tiny) = tiny with inexact */
    t = sun_expm1(fabs(x));
    if (ix < 0x3ff00000) return h*(2.0*t - t*t/(t + one));
    return h*(t + t/(t + one));
  }
  /* |x| in [22, log(maxdouble)] return 0.5*exp(|x|) */
  if (ix < 0x40862E42) return h*sun_exp(fabs(x));
  /* |x| in [log(maxdouble), overflowthresold] */
  if (ix <= 0x408633CE) {
    /* return h*2.0*__ldexp_exp(fabs(x), -1); */
    /* taken from Sun source file e_sinh.c 1.3 95/01/18 */
    w = sun_exp(0.5*fabs(x));
    t = h*w;
    return t*w;
  }
  /* |x| > overflowthresold, sinh(x) overflow */
  return x*shuge;
}

/* @(#)e_cosh.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* __ieee754_cosh(x)
 * Method :
 * mathematically cosh(x) if defined to be (exp(x)+exp(-x))/2
 *	1. Replace x by |x| (cosh(x) = cosh(-x)).
 *	2.
 *		                                        [ exp(x) - 1 ]^2
 *	    0        <= x <= ln2/2  :  cosh(x) := 1 + -------------------
 *			       			           2*exp(x)
 *
 *		                                  exp(x) +  1/exp(x)
 *	    ln2/2    <= x <= 22     :  cosh(x) := -------------------
 *			       			          2
 *	    22       <= x <= lnovft :  cosh(x) := exp(x)/2
 *	    lnovft   <= x <= ln2ovft:  cosh(x) := exp(x/2)/2 * exp(x/2)
 *	    ln2ovft  <  x	    :  cosh(x) := huge*huge (overflow)
 *
 * Special cases:
 *	cosh(x) is |x| if x is +INF, -INF, or NaN.
 *	only cosh(0)=1 is exact for finite x.
 */

LUALIB_API double sun_cosh (double x) {
  double t, w;
  int32_t ix;
  /* High word of |x|. */
  __GET_HI(ix, x);
  ix &= 0x7fffffff;
  /* x is INF or NaN */
  if (ix >= 0x7ff00000) return x*x;
  /* |x| in [0,0.5*ln2], return 1+expm1(|x|)^2/(2*exp(|x|)) */
  if (ix < 0x3fd62e43) {
    t = sun_expm1(fabs(x));
    w = one + t;
    if (ix < 0x3c800000) return w;	/* cosh(tiny) = 1 */
    return one + (t*t)/(w + w);
  }
  /* |x| in [0.5*ln2,22], return (exp(|x|)+1/exp(|x|)/2; */
  if (ix < 0x40360000) {
    t = sun_exp(fabs(x));
    return half*t + half/t;
  }
  /* |x| in [22, log(maxdouble)] return half*exp(|x|) */
  if (ix < 0x40862E42) return half*sun_exp(fabs(x));
  /* |x| in [log(maxdouble), overflowthresold] */
  if (ix <= 0x408633CE) {
    /* return __ldexp_exp(fabs(x), -1); */
    /* taken from Sun source file e_cosh.c 1.3 95/01/18 */
    w = sun_exp(half*fabs(x));
    t = half*w;
    return t*w;
  }
  /* |x| > overflowthresold, cosh(x) overflow */
  return huge*huge;
}


/* @(#)s_tanh.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* Tanh(x)
 * Return the Hyperbolic Tangent of x
 *
 * Method :
 *				       x    -x
 *				      e  - e
 *	0. tanh(x) is defined to be -----------
 *				       x    -x
 *				      e  + e
 *	1. reduce x to non-negative by tanh(-x) = -tanh(x).
 *	2.  0      <= x <= 2**-55 : tanh(x) := x*(one+x)
 *					        -t
 *	    2**-55 <  x <=  1     : tanh(x) := -----; t = expm1(-2x)
 *					       t + 2
 *						     2
 *	    1      <= x <=  22.0  : tanh(x) := 1-  ----- ; t=expm1(2x)
 *						   t + 2
 *	    22.0   <  x <= INF    : tanh(x) := 1.
 *
 * Special cases:
 *	tanh(NaN) is NaN;
 *	only tanh(0)=0 is exact for finite argument.
 */

static const double two = 2.0;

LUALIB_API double sun_tanh (double x) {
  double t, z;
  int32_t jx, ix;
  /* High word of |x|. */
  __GET_HI(jx, x);
  ix = jx & 0x7fffffff;
  /* x is INF or NaN */
  if (unlikely(ix >= 0x7ff00000)) {  /* 2.5.15 optimisation */
    if (jx >= 0) return one/x + one;    /* tanh(+-inf)=+-1 */
    else return one/x - one;    /* tanh(NaN) = NaN */
  }
  /* |x| < 22 */
  if (ix < 0x40360000) {		/* |x|<22 */
    if (ix < 0x3c800000) 		/* |x|<2**-55 */
      return x*(one + x);    	/* tanh(small) = small */
    if (ix >= 0x3ff00000) {	/* |x|>=1  */
      t = sun_expm1(two*fabs(x));
      z = one - two/(t + two);
    } else {
      t = sun_expm1(-two*fabs(x));
      z = -t/(t + two);
    }
    /* |x| > 22, return +-1 */
  } else {
    z = one - tiny;		/* raised inexact flag */
  }
  return (jx >= 0) ? z : -z;
}


/* @(#)s_expm1.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* expm1(x)
 * Returns exp(x)-1, the exponential of x minus 1.
 *
 * Method
 *   1. Argument reduction:
 *	Given x, find r and integer k such that
 *
 *               x = k*ln2 + r,  |r| <= 0.5*ln2 ~ 0.34658
 *
 *      Here a correction term c will be computed to compensate
 *	the error in r when rounded to a floating-point number.
 *
 *   2. Approximating expm1(r) by a special rational function on
 *	the interval [0,0.34658]:
 *	Since
 *	    r*(exp(r)+1)/(exp(r)-1) = 2+ r^2/6 - r^4/360 + ...
 *	we define R1(r*r) by
 *	    r*(exp(r)+1)/(exp(r)-1) = 2+ r^2/6 * R1(r*r)
 *	That is,
 *	    R1(r**2) = 6/r *((exp(r)+1)/(exp(r)-1) - 2/r)
 *		     = 6/r * ( 1 + 2.0*(1/(exp(r)-1) - 1/r))
 *		     = 1 - r^2/60 + r^4/2520 - r^6/100800 + ...
 *      We use a special Reme algorithm on [0,0.347] to generate
 * 	a polynomial of degree 5 in r*r to approximate R1. The
 *	maximum error of this polynomial approximation is bounded
 *	by 2**-61. In other words,
 *	    R1(z) ~ 1.0 + Q1*z + Q2*z**2 + Q3*z**3 + Q4*z**4 + Q5*z**5
 *	where 	Q1  =  -1.6666666666666567384E-2,
 * 		Q2  =   3.9682539681370365873E-4,
 * 		Q3  =  -9.9206344733435987357E-6,
 * 		Q4  =   2.5051361420808517002E-7,
 * 		Q5  =  -6.2843505682382617102E-9;
 *		z   =  r*r,
 *	with error bounded by
 *	    |                  5           |     -61
 *	    | 1.0+Q1*z+...+Q5*z   -  R1(z) | <= 2
 *	    |                              |
 *
 *	expm1(r) = exp(r)-1 is then computed by the following
 * 	specific way which minimize the accumulation rounding error:
 *			       2     3
 *			      r     r    [ 3 - (R1 + R1*r/2)  ]
 *	      expm1(r) = r + --- + --- * [--------------------]
 *		              2     2    [ 6 - r*(3 - R1*r/2) ]
 *
 *	To compensate the error in the argument reduction, we use
 *		expm1(r+c) = expm1(r) + c + expm1(r)*c
 *			   ~ expm1(r) + c + r*c
 *	Thus c+r*c will be added in as the correction terms for
 *	expm1(r+c). Now rearrange the term to avoid optimisation
 * 	screw up:
 *		        (      2                                    2 )
 *		        ({  ( r    [ R1 -  (3 - R1*r/2) ]  )  }    r  )
 *	 expm1(r+c)~r - ({r*(--- * [--------------------]-c)-c} - --- )
 *	                ({  ( 2    [ 6 - r*(3 - R1*r/2) ]  )  }    2  )
 *                      (                                             )
 *
 *		   = r - E
 *   3. Scale back to obtain expm1(x):
 *	From step 1, we have
 *	   expm1(x) = either 2^k*[expm1(r)+1] - 1
 *		    = or     2^k*[expm1(r) + (1-2^-k)]
 *   4. Implementation notes:
 *	(A). To save one multiplication, we scale the coefficient Qi
 *	     to Qi*2^i, and replace z by (x^2)/2.
 *	(B). To achieve maximum accuracy, we compute expm1(x) by
 *	  (i)   if x < -56*ln2, return -1.0, (raise inexact if x!=inf)
 *	  (ii)  if k=0, return r-E
 *	  (iii) if k=-1, return 0.5*(r-E)-0.5
 *        (iv)	if k=1 if r < -0.25, return 2*((r+0.5)- E)
 *	       	       else	     return  1.0+2.0*(r-E);
 *	  (v)   if (k<-2||k>56) return 2^k(1-(E-r)) - 1 (or exp(x)-1)
 *	  (vi)  if k <= 20, return 2^k((1-2^-k)-(E-r)), else
 *	  (vii) return 2^k(1-((E+2^-k)-r))
 *
 * Special cases:
 *	expm1(INF) is INF, expm1(NaN) is NaN;
 *	expm1(-INF) is -1, and
 *	for finite argument, only expm1(0)=0 is exact.
 *
 * Accuracy:
 *	according to an error analysis, the error is always less than
 *	1 ulp (unit in the last place).
 *
 * Misc. info.
 *	For IEEE double
 *	    if x >  7.09782712893383973096e+02 then expm1(x) overflow
 *
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 */

static const double
/* Scaled Q's: Qn_here = 2**n * Qn_above, for R(2*z) where z = hxs = x*x/2: */
Q1  =  -3.33333333333331316428e-02, /* BFA11111 111110F4 */
Q2  =   1.58730158725481460165e-03, /* 3F5A01A0 19FE5585 */
Q3  =  -7.93650757867487942473e-05, /* BF14CE19 9EAADBB7 */
Q4  =   4.00821782732936239552e-06, /* 3ED0CFCA 86E65239 */
Q5  =  -2.01099218183624371326e-07; /* BE8AFDB7 6E09C32D */

LUALIB_API double sun_expm1 (double x) {  /* more than twice as fast as GCC's implementation */
  double y, hi, lo, c, t, e, hxs, hfx, r1, twopk;
  int32_t k, xsb;
  uint32_t hx;
  c = 0;
  __GET_HI(hx, x);
  xsb = hx & 0x80000000;  /* sign bit of x */
  hx &= 0x7fffffff;       /* high word of |x| */
  /* filter out huge and non-finite argument */
  if (hx >= 0x4043687A) {    /* if |x|>=56*ln2 */
    if (hx >= 0x40862E42) {    /* if |x|>=709.78... */
      if (unlikely(hx >= 0x7ff00000)) {  /* 2.5.15 optimisation */
        uint32_t low;
        __GET_LO(low, x);
        if (((hx & 0xfffff) | low) != 0)
          return x + x; 	 /* NaN */
        else return (xsb == 0) ? x : -1.0;    /* exp(+-inf)={inf,-1} */
      }
      if (x > o_threshold) return huge*huge;  /* overflow */
    }
    if (xsb != 0) {  /* x < -56*ln2, return -1.0 with inexact */
      if (x + tiny < 0.0)   /* raise inexact */
		return tiny - one;  /* return -1 */
    }
  }
  /* argument reduction */
  if (hx > 0x3fd62e42) {    /* if  |x| > 0.5 ln2 */
    if (hx < 0x3FF0A2B2) {  /* and |x| < 1.5 ln2 */
      if (xsb == 0) {
        hi = x - ln2_hi; lo =  ln2_lo; k = 1;
      } else {
        hi = x + ln2_hi; lo = -ln2_lo; k = -1;
      }
    } else {
      k  = invln2*x + ((xsb == 0) ? 0.5 : -0.5);
      t  = k;
      hi = x - t*ln2_hi;	/* t*ln2_hi is exact here */
      lo = t*ln2_lo;
    }
    __STRICT_ASSIGN(double, x, hi - lo);
    c = (hi - x) - lo;
  }
  else if (hx < 0x3c900000) {  	/* when |x|<2**-54, return x */
    t = huge + x;	/* return x with inexact flags when x!=0 */
    return x - (t - (huge + x));
  }
  else k = 0;
  /* x is now in primary range */
  hfx = 0.5*x;
  hxs = x*hfx;
  r1 = one + hxs*(Q1 + hxs*(Q2 + hxs*(Q3 + hxs*(Q4 + hxs*Q5))));
  t  = 3.0 - r1*hfx;
  e  = hxs*((r1 - t)/(6.0 - x*t));
  if (k == 0) return x - (x*e - hxs);		/* c is 0 */
  else {
    __INSERT_WORDS(twopk, 0x3ff00000 + (k << 20), 0);	/* 2^k */
    e  = (x*(e - c) - c);
    e -= hxs;
    if (k == -1) return 0.5*(x - e)-0.5;
    if (k == 1) {
      if (x < -0.25) return -2.0*(e - (x + 0.5));
      else return  one + 2.0*(x - e);
    }
    if (k <= -2 || k > 56) {   /* suffice to return exp(x)-1 */
      y = one - (e - x);
      if (k == 1024) y = y*2.0*0x1p1023;
      else y = y*twopk;
      return y - one;
    }
    t = one;
    if (k < 20) {
      __SET_HI(t, 0x3ff00000 - (0x200000 >> k));  /* t=1-2^-k */
      y = t - (e - x);
      y = y*twopk;
    } else {
      __SET_HI(t, ((0x3ff - k) << 20));	/* 2^-k */
      y = x - (e + t);
      y += one;
      y = y*twopk;
    }
  }
  return y;
}

/* @(#)s_log1p.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* double log1p(double x)
 *
 * Method :
 *   1. Argument Reduction: find k and f such that
 *			1+x = 2^k * (1+f),
 *	   where  sqrt(2)/2 < 1+f < sqrt(2) .
 *
 *      Note. If k=0, then f=x is exact. However, if k!=0, then f
 *	may not be representable exactly. In that case, a correction
 *	term is need. Let u=1+x rounded. Let c = (1+x)-u, then
 *	log(1+x) - log(u) ~ c/u. Thus, we proceed to compute log(u),
 *	and add back the correction term c/u.
 *	(Note: when x > 2**53, one can simply return log(x))
 *
 *   2. Approximation of log1p(f).
 *	Let s = f/(2+f) ; based on log(1+f) = log(1+s) - log(1-s)
 *		 = 2s + 2/3 s**3 + 2/5 s**5 + .....,
 *	     	 = 2s + s*R
 *      We use a special Reme algorithm on [0,0.1716] to generate
 * 	a polynomial of degree 14 to approximate R The maximum error
 *	of this polynomial approximation is bounded by 2**-58.45. In
 *	other words,
 *		        2      4      6      8      10      12      14
 *	    R(z) ~ Lp1*s +Lp2*s +Lp3*s +Lp4*s +Lp5*s  +Lp6*s  +Lp7*s
 *  	(the values of Lp1 to Lp7 are listed in the program)
 *	and
 *	    |      2          14          |     -58.45
 *	    | Lp1*s +...+Lp7*s    -  R(z) | <= 2
 *	    |                             |
 *	Note that 2s = f - s*f = f - hfsq + s*hfsq, where hfsq = f*f/2.
 *	In order to guarantee error in log below 1ulp, we compute log
 *	by
 *		log1p(f) = f - (hfsq - s*(hfsq+R)).
 *
 *	3. Finally, log1p(x) = k*ln2 + log1p(f).
 *		 	     = k*ln2_hi+(f-(hfsq-(s*(hfsq+R)+k*ln2_lo)))
 *	   Here ln2 is split into two floating point number:
 *			ln2_hi + ln2_lo,
 *	   where n*ln2_hi is always exact for |n| < 2000.
 *
 * Special cases:
 *	log1p(x) is NaN with signal if x < -1 (including -INF) ;
 *	log1p(+INF) is +INF; log1p(-1) is -INF with signal;
 *	log1p(NaN) is that NaN with no signal.
 *
 * Accuracy:
 *	according to an error analysis, the error is always less than
 *	1 ulp (unit in the last place).
 *
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 *
 * Note: Assuming log() return accurate answer, the following
 * 	 algorithm can be used to compute log1p(x) to within a few ULP:
 *
 *		u = 1+x;
 *		if(u==1.0) return x ; else
 *			   return log(u)*(x/(u-1.0));
 *
 *	 See HP-15C Advanced Functions Handbook, p.193.
 */
/*
 * k_log1p(f):
 * Return log(1+f) - f for 1+f in ~[sqrt(2)/2, sqrt(2)].
 *
 * The following describes the overall strategy for computing
 * logarithms in base e.  The argument reduction and adding the final
 * term of the polynomial are done by the caller for increased accuracy
 * when different bases are used.
 *
 * Method :
 *   1. Argument Reduction: find k and f such that
 *			x = 2^k * (1+f),
 *	   where  sqrt(2)/2 < 1+f < sqrt(2) .
 *
 *   2. Approximation of log(1+f).
 *	Let s = f/(2+f) ; based on log(1+f) = log(1+s) - log(1-s)
 *		 = 2s + 2/3 s**3 + 2/5 s**5 + .....,
 *	     	 = 2s + s*R
 *      We use a special Reme algorithm on [0,0.1716] to generate
 * 	a polynomial of degree 14 to approximate R The maximum error
 *	of this polynomial approximation is bounded by 2**-58.45. In
 *	other words,
 *		        2      4      6      8      10      12      14
 *	    R(z) ~ Lg1*s +Lg2*s +Lg3*s +Lg4*s +Lg5*s  +Lg6*s  +Lg7*s
 *  	(the values of Lg1 to Lg7 are listed in the program)
 *	and
 *	    |      2          14          |     -58.45
 *	    | Lg1*s +...+Lg7*s    -  R(z) | <= 2
 *	    |                             |
 *	Note that 2s = f - s*f = f - hfsq + s*hfsq, where hfsq = f*f/2.
 *	In order to guarantee error in log below 1ulp, we compute log
 *	by
 *		log(1+f) = f - s*(f - R)	(if f is not too large)
 *		log(1+f) = f - (hfsq - s*(hfsq+R)).	(better accuracy)
 *
 *	3. Finally,  log(x) = k*ln2 + log(1+f).
 *			    = k*ln2_hi+(f-(hfsq-(s*(hfsq+R)+k*ln2_lo)))
 *	   Here ln2 is split into two floating point number:
 *			ln2_hi + ln2_lo,
 *	   where n*ln2_hi is always exact for |n| < 2000.
 *
 * Special cases:
 *	log(x) is NaN with signal if x < 0 (including -INF) ;
 *	log(+INF) is +INF; log(0) is -INF with signal;
 *	log(NaN) is that NaN with no signal.
 *
 * Accuracy:
 *	according to an error analysis, the error is always less than
 *	1 ulp (unit in the last place).
 *
 * Constants:
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 */
/*
 * We always inline k_log1p(), since doing so produces a
 * substantial performance improvement (~40% on amd64).
 */
static INLINE double kernel_log1p (double f) {  /* inlined 2.16.13 */
  double hfsq, s, z, R, w, t1, t2;
  s = f/(2.0 + f);
  z = s*s;
  w = z*z;
  t1 = w*(Lg2 + w*(Lg4 + w*Lg6));
  t2 = z*(Lg1 + w*(Lg3 + w*(Lg5 + w*Lg7)));
  R = t2 + t1;
  hfsq = 0.5*f*f;
  return s*(hfsq + R);
}

/* @(#)e_log2.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 *
 * Return the base 2 logarithm of x.  See e_log.c and k_log.h for most
 * comments.
 *
 * This reduces x to {k, 1+f} exactly as in e_log.c, then calls the kernel,
 * then does the combining and scaling steps
 *    log2(x) = (f - 0.5*f*f + k_log1p(f)) / ln2 + k
 * in not-quite-routine extra precision.
 *
 * There is an alternative implementation in uClibc-0.9.33/libm/e_log2.c, it is slower, however.
 *
 */

static const double
ivln2hi    =  1.44269504072144627571e+00, /* 0x3ff71547, 0x65200000 ~~ log2(e) */
ivln2lo    =  1.67517131648865118353e-10; /* 0x3de705fc, 0x2eefa200 */

LUALIB_API INLINE double sun_log2 (double x) {
  double f, hfsq, hi, lo, r, val_hi, val_lo, w, y;
  int32_t i, k, hx;
  uint32_t lx;
  __EXTRACT_WORDS(hx, lx, x);
  k = 0;
  if (hx < 0x00100000) {  /* x < 2**-1022  */
    if ((((hx & 0x7fffffff) | lx) == 0) || hx < 0)
      return AGN_NAN;  /* =(x - x)/zero; */ /* log(0-#) = NaN, change, we do not return -inf for x=0 */
    k -= 54; x *= two54;    /* subnormal number, scale up x */
    __GET_HI(hx, x);
  }
  if (unlikely(hx >= 0x7ff00000)) return x + x;  /* inf, nan, 2.5.15 optimisation */
  if (hx == 0x3ff00000 && lx == 0)
    return zero;  /* log(1) = +0 */
  k += (hx >> 20) - 1023;
  hx &= 0x000fffff;
  i = (hx + 0x95f64) & 0x100000;
  __SET_HI(x, hx | (i ^ 0x3ff00000));  /* normalize x or x/2 */
  k += (i >> 20);
  y = (double)k;
  f = x - 1.0;
  hfsq = 0.5*f*f;
  r = kernel_log1p(f);
  /*
   * f-hfsq must (for args near 1) be evaluated in extra precision
   * to avoid a large cancellation when x is near sqrt(2) or 1/sqrt(2).
   * This is fairly efficient since f-hfsq only depends on f, so can
   * be evaluated in parallel with R.  Not combining hfsq with R also
   * keeps R small (though not as small as a true `lo' term would be),
   * so that extra precision is not needed for terms involving R.
   *
   * Compiler bugs involving extra precision used to break Dekker's
   * theorem for spitting f-hfsq as hi+lo, unless double_t was used
   * or the multi-precision calculations were avoided when double_t
   * has extra precision.  These problems are now automatically
   * avoided as a side effect of the optimization of combining the
   * Dekker splitting step with the clear-low-bits step.
   *
   * y must (for args near sqrt(2) and 1/sqrt(2)) be added in extra
   * precision to avoid a very large cancellation when x is very near
   * these values.  Unlike the above cancellations, this problem is
   * specific to base 2.  It is strange that adding +-1 is so much
   * harder than adding +-ln2 or +-log10_2.
   *
   * This uses Dekker's theorem to normalize y+val_hi, so the
   * compiler bugs are back in some configurations, sigh.  And I
   * don't want to used double_t to avoid them, since that gives a
   * pessimization and the support for avoiding the pessimization
   * is not yet available.
   *
   * The multi-precision calculations for the multiplications are
   * routine.
  */
  hi = f - hfsq;
  __SET_LO(hi, 0);
  lo = (f - hi) - hfsq + r;
  val_hi = hi*ivln2hi;
  val_lo = (lo + hi)*ivln2lo + lo*ivln2hi;
  /* spadd(val_hi, val_lo, y), except for not using double_t: */
  w = y + val_hi;
  val_lo += (y - w) + val_hi;
  val_hi = w;
  return val_lo + val_hi;
}


/* @(#)e_log10.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/*
 * Return the base 10 logarithm of x.  See e_log.c and k_log.h for most
 * comments.
 *
 *    log10(x) = (f - 0.5*f*f + k_log1p(f)) / ln10 + k * log10(2)
 * in not-quite-routine extra precision.
 */

static const double
ivln10hi   =  4.34294481878168880939e-01, /* 0x3fdbcb7b, 0x15200000 */
ivln10lo   =  2.50829467116452752298e-11, /* 0x3dbb9438, 0xca9aadd5 */
log10_2hi  =  3.01029995663611771306e-01, /* 0x3FD34413, 0x509F6000 */
log10_2lo  =  3.69423907715893078616e-13; /* 0x3D59FEF3, 0x11F12B36 */

LUALIB_API INLINE double sun_log10 (double x) {  /* four percent faster than GCC's log10 implementation */
  double f, hfsq, hi, lo, r, val_hi, val_lo, w, y, y2;
  int32_t i, k, hx;
  uint32_t lx;
  __EXTRACT_WORDS(hx, lx, x);
  k = 0;
  if (hx < 0x00100000) {  /* x < 2**-1022  */
    /* changed 2.11.1 : log(+-0) = NaN && log(-#) = NaN */
    if (unlikely(hx < 0 || ((hx & 0x7fffffff) | lx) == 0)) return AGN_NAN;  /* = (x - x)/zero; 2.5.15 optimisation */
    k -= 54; x *= two54; /* subnormal number, scale up x */
    __GET_HI(hx, x);
  }
  if (unlikely(hx >= 0x7ff00000)) return x + x;  /* nan, inf, 2.5.15 optimisation */
  if (hx == 0x3ff00000 && lx == 0)
    return zero;  /* log(1) = +0 */
  k += (hx >> 20) - 1023;
  hx &= 0x000fffff;
  i = (hx + 0x95f64) & 0x100000;
  __SET_HI(x, hx | (i ^ 0x3ff00000));  /* normalize x or x/2 */
  k += (i >> 20);
  y = (double)k;
  f = x - 1.0;
  hfsq = 0.5*f*f;
  r = kernel_log1p(f);
  /* See e_log2.c for most details. */
  hi = f - hfsq;
  __SET_LO(hi, 0);
  lo = (f - hi) - hfsq + r;
  val_hi = hi*ivln10hi;
  y2 = y*log10_2hi;
  val_lo = y*log10_2lo + (lo + hi)*ivln10lo + lo*ivln10hi;
  /*
   * Extra precision in for adding y*log10_2hi is not strictly needed
   * since there is no very large cancellation near x = sqrt(2) or
   * x = 1/sqrt(2), but we do it anyway since it costs little on CPUs
   * with some parallelism and it reduces the error for many args.
   */
  w = y2 + val_hi;
  val_lo += (y2 - w) + val_hi;
  val_hi = w;
  return val_lo + val_hi;
}


/* see: https://bitbucket.org/zlwu/cegcc-mingw/src/0d9e81454bfa3231299396c92d4b019bf283cd2b/mingwex/complex/ctan.c?at=master&fileviewer=file-view-default
  ctan(a + I*b) = (sin(2*a) + I*sinh(2*b)) / (cos(2*a) + cosh(2*b)) */
#ifndef PROPCMPLX
LUALIB_API double complex tools_ctan (double complex z) {
  double re, im, den;
  double complex r;
  re = 2*(__real__ z);
  im = 2*(__imag__ z);
  den = sun_cos(re) + sun_cosh(im);
  switch (tools_fpclassify(den)) {
    case FP_ZERO:
      __real__ r = HUGE_VAL;
      __imag__ r = HUGE_VAL;
      break;
    case FP_INFINITE:
      __real__ r = 0;
      __imag__ r = im > 0 ? 1 : -1;
      break;
    default: {
      __real__ r = sun_sin(re)/den;
      __imag__ r = sun_sinh(im)/den;
    }
  }
  return r;
}
#endif

/* openlibm-0.4.1/src/s_scalbn.c, do not use fdbmlib's implementation as the latter is inaccurate
 *
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */
/*
 * scalbn/ldexp (double x, int n)
 * scalbn/ldexp(x,n) returns x* 2**n  computed by  exponent
 * manipulation rather than by actually performing an
 * exponentiation or a multiplication.
 */

static const double
twom54  =  5.55111512312578270212e-17; /* 0x3C900000, 0x00000000 */

LUALIB_API double sun_ldexp (double x, int n) {
  int32_t k, hx, lx;
  if (!isfinite(x) || x == 0.0) return x;  /* work like GCC's scalbn/ldexp */
  __EXTRACT_WORDS(hx, lx, x);
  k = (hx & 0x7ff00000) >> 20; /* extract exponent */
  if (k == 0) { /* 0 or subnormal x */
    if ((lx | (hx & 0x7fffffff)) == 0)
      return x; /* +-0 */
    x *= two54;
    __GET_HI(hx, x);
    k = ((hx & 0x7ff00000) >> 20) - 54;
  }
  if (unlikely(k == 0x7ff))  /* 2.5.15 optimisation */
    return x + x; /* NaN or Inf */
  k = k + n;
  if (k > 0x7fe)
    return huge * copysign(huge, x); /* overflow */
  if (n < -50000)
    return tiny * copysign(tiny, x); /* underflow */
  if (k > 0) { /* normal result */
    __SET_HI(x, (hx & 0x800fffff) | (k << 20));
    return x;
  }
  if (k <= -54) {
    if (n > 50000) /* in case integer overflow in n+k */
      return huge * copysign(huge, x); /* overflow */
    return tiny * copysign(tiny, x); /* underflow */
    }
  k += 54; /* subnormal result */
  __SET_HI(x, (hx & 0x800fffff) | (k << 20));
  return x * twom54;
}


/* @(#)s_frexp.c 1.4 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */
/*
 * for non-zero x
 *	x = frexp(arg,&exp);
 * return a double fp quantity x such that 0.5 <= |x| <1.0
 * and the corresponding binary exponent "exp". That is
 *	arg = x*2^exp.
 * If arg is inf, 0.0, or NaN, then frexp(arg,&exp) returns arg
 * with *exp=0.
 */

LUALIB_API double sun_frexp (double x, int *eptr) {
  int32_t hx, ix, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ix = 0x7fffffff & hx;
  *eptr = 0;
  if (ix >= 0x7ff00000 || ((ix | lx) == 0)) return x;  /* 0,inf,nan */
  if (ix < 0x00100000) {  /* subnormal */
    x *= two54;
    __GET_HI(hx, x);
    ix = hx & 0x7fffffff;
    *eptr = -54;
  }
  *eptr += (ix >> 20) - 1022;
  hx = (hx & 0x800fffff) | 0x3fe00000;
  __SET_HI(x, hx);
  return x;
}


/* Source: openlibm-0.4.1/src/k_exp.c

 * Copyright (c) 2011 David Schultz <das@FreeBSD.ORG>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * __ldexp_exp(x, expt) and __ldexp_cexp(x, expt) compute exp(x) * 2**expt.
 * They are intended for large arguments (real part >= ln(DBL_MAX))
 * where care is needed to avoid overflow.
 *
 * The present implementation is narrowly tailored for our hyperbolic and
 * exponential functions.  We assume expt is small (0 or -1), and the caller
 * has filtered out very large x, for which overflow would be inevitable.
 */
static const uint32_t k = 1799;		/* constant for reduction */
static const double kln2 =  1246.97177782734161156;	/* k * ln2 */

/* Source: openlibm-0.4.1/src/k_exp.c
 * Compute exp(x), scaled to avoid spurious overflow.  An exponent is
 * returned separately in 'expt'.
 *
 * Input:  ln(DBL_MAX) <= x < ln(2 * DBL_MAX / DBL_MIN_DENORM) ~= 1454.91
 * Output: 2**1023 <= y < 2**1024
 */
LUALIB_API double ds_frexp_exp (double x, int *expt) {
  double exp_x;
  uint32_t hx;
  /*
   * We use exp(x) = exp(x - kln2) * 2**k, carefully chosen to
   * minimize |exp(kln2) - 2**k|.  We also scale the exponent of
   * exp_x to MAX_EXP so that the result can be multiplied by
   * a tiny number without losing accuracy due to denormalisation.
  */
  exp_x = sun_exp(x - kln2);
  __GET_HI(hx, exp_x);
  *expt = (hx >> 20) - (0x3ff + 1023) + k;
  __SET_HI(exp_x, (hx & 0xfffff) | ((0x3ff + 1023) << 20));
  return exp_x;
}

/* Source: openlibm-0.4.1/src/k_exp.c
 * __ldexp_exp(x, expt) and __ldexp_cexp(x, expt) compute exp(x) * 2**expt.
 * They are intended for large arguments (real part >= ln(DBL_MAX))
 * where care is needed to avoid overflow.
 *
 * The present implementation is narrowly tailored for our hyperbolic and
 * exponential functions.  We assume expt is small (0 or -1), and the caller
 * has filtered out very large x, for which overflow would be inevitable.
 */

LUALIB_API double ds_ldexp_exp (double x, int expt) {
  double exp_x, scale;
  int ex_expt;
  exp_x = ds_frexp_exp(x, &ex_expt);
  expt += ex_expt;
  __INSERT_WORDS(scale, (0x3ff + expt) << 20, 0);
  return (exp_x * scale);
}


#ifndef PROPCMPLX
LUALIB_API double complex ds_ldexp_cexp (double complex z, int expt) {
  double x, y, exp_x, scale1, scale2;
  int ex_expt, half_expt;
  x = creal(z);
  y = cimag(z);
  exp_x = ds_frexp_exp(x, &ex_expt);
  expt += ex_expt;
  /*
   * Arrange so that scale1 * scale2 == 2**expt.  We use this to
   * compensate for scalbn being horrendously slow.
  */
  half_expt = 0.5*expt;
  __INSERT_WORDS(scale1, (0x3ff + half_expt) << 20, 0);
  half_expt = expt - half_expt;
  __INSERT_WORDS(scale2, (0x3ff + half_expt) << 20, 0);
  return (cpack(cos(y) * exp_x * scale1 * scale2,
    sin(y) * exp_x * scale1 * scale2));
}
#endif


/*	$OpenBSD: s_cpow.c,v 1.6 2013/07/03 04:46:36 espie Exp $	*/
/*
 * Copyright (c) 2008 Stephen L. Moshier <steve@moshier.net>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*							cpow
 *
 *	Complex power function
 *
 *
 *
 * SYNOPSIS:
 *
 * double complex cpow();
 * double complex a, z, w;
 *
 * w = cpow (a, z);
 *
 *
 *
 * DESCRIPTION:
 *
 * Raises complex A to the complex Zth power.
 * Definition is per AMS55 # 4.2.8,
 * analytically equivalent to cpow(a,z) = cexp(z clog(a)).
 *
 * ACCURACY:
 *
 *                      Relative error:
 * arithmetic   domain     # trials      peak         rms
 *    IEEE      -10,+10     30000       9.4e-15     1.5e-15
 *
 */
#ifndef PROPCMPLX
LUALIB_API double complex tools_cpow (double complex a, double complex z) {  /* 2.11.1, 13 % faster than GCC's implementation */
  double complex w;
  double x, y, r, theta, absa, arga, si, co, x0, y0;
  x0 = creal(a); y0 = cimag(a);
  x = creal(z); y = cimag(z);
  if (x0 == 0 && y0 == 0 && x <= 0) {  /* modified */
    return (y == 0) ? AGN_NAN : 0;
  }
  absa = sun_hypot(x0, y0);
  arga = sun_atan2(y0, x0);
  r = sun_pow(absa, x);
  theta = x*arga;
  if (y != 0.0) {
    r = r*sun_exp(-y*arga);
    theta = theta + y*sun_log(absa);
  }
  tools_sincos(theta, &si, &co);
  __real__ w = r*co;
  __imag__ w = r*si;  /* we deliberately do not zero values close to 0 ! */
  return w;
}


LUALIB_API double complex tools_cexp (double complex z) {  /* 2.11.1 */
  double complex r;
  lua_Number a, b, t1;
  a = __real__ z;
  b = __imag__ z;
  t1 = sun_exp(a);
  tools_sincos(b, &a, &b);
  __real__ r = t1*b;
  __imag__ r = t1*a;
  return r;
}


LUALIB_API double complex tools_clog (double complex z) {  /* 2.11.1, 25 % faster than GCC's clog */
  agn_Complex r;
  lua_Number a, b, t;
  a = __real__ z;
  b = __imag__ z;
  if (a == 0 && b == 0) {
    __real__ r = AGN_NAN;
    __imag__ r = 0;
  } else {
    t = a*a + b*b;
    __real__ r = sun_ldexp(sun_log(t), -1);
    __imag__ r = sun_atan2(b, a);
  }
  return r;
}


LUALIB_API double tools_cabs (double complex z) {  /* 2.11.1 */
  lua_Number a, b, t;
  a = creal(z); b = cimag(z);
  if ( isnan(a) || isnan(b) ) return AGN_NAN;
  if ( isinf(a) != 0 || isinf(b) != 0 ) return HUGE_VAL;
  if (a < 0) a = -a;
  if (b < 0) b = -b;
  if (b > a) { SWAP(a, b, t) }
  if (a + b == a) return (a);  /* -> a == 0 || b == 0 */
  t = b/a;  /* minimum divided by maximum */
  return a * sqrt(1 + t*t);
}


LUALIB_API double complex tools_cdiv (double complex x, double complex y) {  /* 2.11.1 */
  double complex r;
  lua_Number a, b, c, d, t2, t3, t5;
  a = creal(x); b = cimag(x); c = creal(y); d = cimag(y);
  if ((c) == 0 && (d) == 0) (r) = AGN_NAN;
  else {
    t2 = c*c;
    t3 = d*d;
    t5 = 1/(t2 + t3);
    __real__ r = a*c*t5 + b*d*t5;
    __imag__ r = b*c*t5 - a*d*t5;
  }
  return r;
}


/* Source: https://sourceforge.net/p/mingw-w64/mingw-w64/ci/2bdcae2a1af9ee60c5e9ee1c5d80553295c092b4/tree/mingw-w64-crt/complex/casin.c
 *
 * This file has no copyright assigned and is placed in the Public Domain.
 *
 * casin.c
 * Contributed by Danny Smith
 * 2003-10-20

 MapleV R4:

 > z := a + I*b;

                             z := a + I b

 > convert(arcsin(x), ln);

                                   2 1/2
                       -I ln((1 - x )    + I x)
*/
LUALIB_API double complex tools_casin (double complex z) {  /* 2.11.2 */
  double complex r;
  double x, y;
  x = __real__ z;
  y = __imag__ z;
  if (y == 0) {
#if (defined(__SOLARIS) || defined(__OS2__))
    __real__ r = sun_asin(x);  /* GCC 3.4.6's a sin is buggy with 0 + I*b */
#else
    __real__ r = asin(x);
#endif
    __imag__ r = 0;
  } else {
    double complex zz;
    __real__ zz = 1 - (x - y)*(x + y);
    __imag__ zz = -2*x*y;
    zz = tools_csqrt(zz);
    __real__ zz -= y;
    __imag__ zz += x;
    zz = tools_clog(zz);
    __real__ r =   __imag__ zz;
    __imag__ r = - __real__ zz;
  }
  return r;
}


/* Source: https://sourceforge.net/p/mingw-w64/mingw-w64/ci/2bdcae2a1af9ee60c5e9ee1c5d80553295c092b4/tree/mingw-w64-crt/complex/catan.c */
LUALIB_API double complex tools_catan (double complex z) {  /* 2.11.2, 15 percent faster than GCC's catan */
  double complex r, t;
  double x, y;
  x = __real__ z;
  y = __imag__ z;
  if (y == 0) {  /* 2.11.2, added to prevent round-off errors */
    __real__ r = sun_atan(x);
    __imag__ r = 0;
  } else if (x == 0 && (fabs(y)) == 1) {
    __real__ r = AGN_NAN;  /* modified */
    __imag__ r = AGN_NAN;  /* modified */
  } else if (isinf(sun_hypot(x, y))) {
    __real__ r = (x > 0 ? PIO2 : -PIO2);
    __imag__ r = 0.0;
  } else {
    __real__ t = -x;
    __imag__ t = 1 - y;
    __real__ r = x;
    __imag__ r = y + 1;
    t = tools_clog(tools_cdiv(r, t));
    __real__ r  = -0.5 * __imag__ t;
    __imag__ r =   0.5 * __real__ t;
    if (x == 0 && y > 1)  /* 2.11.2, added to prevent wrong results */
      __real__ r = - __real__ r;
  }
  return r;
}


/* 2.11.2, this Maple-based implementation is 12 % faster than
   https://sourceforge.net/p/mingw-w64/mingw-w64/ci/2bdcae2a1af9ee60c5e9ee1c5d80553295c092b4/tree/mingw-w64-crt/complex/csqrt.c#l60
   the latter 8 % faster than David Schultz's 2007 FreeBSD version */
LUALIB_API double complex tools_csqrt (double complex z) {
  agn_Complex r;
  lua_Number a, b, t1, t2, t4, t6, t10;
  a = creal(z); b = cimag(z);
  if (b == 0) {
    if (a < 0) {
      __real__ r = 0;
      __imag__ r = sqrt(-a);
    } else {
      __real__ r = sqrt(a);
      __imag__ r = 0;
    }
  } else {
    t1 = a*a; t2 = b*b; t4 = sqrt(t1 + t2);
    t6 = sqrt(2.0*t4 + 2.0*a); t10 = sqrt(2.0*t4 - 2.0*a);
    __real__ r = t6/2;
    __imag__ r = tools_csgn(b, -a)*t10/2;
  }
  return r;
}


LUALIB_API double complex tools_ctanh (double complex z) {  /* 2.11.2, 30 % faster than GCC's ctanh */
  double complex r;
  double rtwo, itwo, denom;
  rtwo = 2.0 * __real__ z;
  itwo = 2.0 * __imag__ z;
  denom = sun_cosh(rtwo) + sun_cos(itwo);
  if (denom == 0.0) {
    __real__ r = HUGE_VAL;
    __imag__ r = HUGE_VAL;
  } else if (isinf(denom)) {
    __real__ r = rtwo > 0 ? 1.0 : -1.0;
    __imag__ r = 0.0;
  } else {
    __real__ r = sun_sinh(rtwo)/denom;
    __imag__ r = sun_sin(itwo)/denom;
  }
  return r;
}
#endif


/* @(#)e_acos.c 1.3 95/01/18 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* __ieee754_acos(x)
 * Method :
 *	acos(x)  = pi/2 - asin(x)
 *	acos(-x) = pi/2 + asin(x)
 * For |x|<=0.5
 *	acos(x) = pi/2 - (x + x*x^2*R(x^2))	(see asin.c)
 * For x>0.5
 * 	acos(x) = pi/2 - (pi/2 - 2asin(sqrt((1-x)/2)))
 *		= 2asin(sqrt((1-x)/2))
 *		= 2s + 2s*z*R(z) 	...z=(1-x)/2, s=sqrt(z)
 *		= 2f + (2c + 2s*z*R(z))
 *     where f=hi part of s, and c = (z-f*f)/(s+f) is the correction term
 *     for f so that f+c ~ sqrt(z).
 * For x<-0.5
 *	acos(x) = pi - 2asin(sqrt((1-|x|)/2))
 *		= pi - 0.5*(s+s*z*R(z)), where z=(1-|x|)/2,s=sqrt(z)
 *
 * Special cases:
 *	if x is NaN, return x itself;
 *	if |x|>1, return NaN with invalid signal.
 *
 * Function needed: sqrt
 */

static const double
pio2_hi =  1.57079632679489655800e+00, /* 0x3FF921FB, 0x54442D18 */
pio2_lo =  6.12323399573676603587e-17, /* 0x3C91A626, 0x33145C07 */
pS0 =  1.66666666666666657415e-01, /* 0x3FC55555, 0x55555555 */
pS1 = -3.25565818622400915405e-01, /* 0xBFD4D612, 0x03EB6F7D */
pS2 =  2.01212532134862925881e-01, /* 0x3FC9C155, 0x0E884455 */
pS3 = -4.00555345006794114027e-02, /* 0xBFA48228, 0xB5688F3B */
pS4 =  7.91534994289814532176e-04, /* 0x3F49EFE0, 0x7501B288 */
pS5 =  3.47933107596021167570e-05, /* 0x3F023DE1, 0x0DFDF709 */
qS1 = -2.40339491173441421878e+00, /* 0xC0033A27, 0x1C8A2D4B */
qS2 =  2.02094576023350569471e+00, /* 0x40002AE5, 0x9C598AC8 */
qS3 = -6.88283971605453293030e-01, /* 0xBFE6066C, 0x1B8D0159 */
qS4 =  7.70381505559019352791e-02; /* 0x3FB3B8C5, 0xB12E9282 */

LUALIB_API double sun_acos (double x) {  /* as fast as GCC's acos implementation */
  double z, p, q, r, w, s, c, df;
  int32_t hx, lx, ix;
  __GET_HI(hx, x);
  ix = hx & 0x7fffffff;
  if (ix > 0x3ff00000) return AGN_NAN;  /* |x| > 1 ? modified */
  __GET_LO(lx, x);
  if (((ix - 0x3ff00000) | lx) == 0) {	/* |x|==1 */
    if (hx > 0) return 0.0;		/* acos(1) = 0  */
    else return pi + 2.0*pio2_lo;	/* acos(-1)= pi */
  }
  if (ix < 0x3fe00000) {	/* |x| < 0.5 */
    if (ix <= 0x3c600000) return pio2_hi + pio2_lo;/*if|x|<2**-57*/
    z = x*x;
    p = z*(pS0 + z*(pS1 + z*(pS2 + z*(pS3 + z*(pS4 + z*pS5)))));
    q = one + z*(qS1 + z*(qS2 + z*(qS3 + z*qS4)));
    r = p/q;
    return pio2_hi - (x - (pio2_lo - x*r));
  } else if (hx < 0) {		/* x < -0.5 */
    z = (one + x)*0.5;
    p = z*(pS0 + z*(pS1 + z*(pS2 + z*(pS3 + z*(pS4 + z*pS5)))));
    q = one + z*(qS1 + z*(qS2 + z*(qS3 + z*qS4)));
    s = sqrt(z);
    r = p/q;
    w = r*s - pio2_lo;
    return pi - 2.0*(s+w);
  } else {			/* x > 0.5 */
    z = (one - x)*0.5;
    s = sqrt(z);
    df = s;
    __SET_LO(df, 0);
    c  = (z - df*df)/(s + df);
    p = z*(pS0 + z*(pS1 + z*(pS2 + z*(pS3 + z*(pS4 + z*pS5)))));
    q = one + z*(qS1 + z*(qS2 + z*(qS3 + z*qS4)));
    r = p/q;
    w = r*s + c;
    return 2.0*(df + w);
  }
}


/* @(#)e_asin.c 1.4 96/03/07 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunSoft, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */
/* __ieee754_asin(x)
 * Method :
 *	Since  asin(x) = x + x^3/6 + x^5*3/40 + x^7*15/336 + ...
 *	we approximate asin(x) on [0,0.5] by
 *		asin(x) = x + x*x^2*R(x^2)
 *	where
 *		R(x^2) is a rational approximation of (asin(x)-x)/x^3
 *	and its Remes error is bounded by
 *		|(asin(x)-x)/x^3 - R(x^2)| < 2^(-58.75)
 *
 *	For x in [0.5,1]
 *		asin(x) = pi/2-2*asin(sqrt((1-x)/2))
 *	Let y = (1-x), z = y/2, s := sqrt(z), and pio2_hi+pio2_lo=pi/2;
 *	then for x>0.98
 *		asin(x) = pi/2 - 2*(s+s*z*R(z))
 *			= pio2_hi - (2*(s+s*z*R(z)) - pio2_lo)
 *	For x<=0.98, let pio4_hi = pio2_hi/2, then
 *		f = hi part of s;
 *		c = sqrt(z) - f = (z-f*f)/(s+f) 	...f+c=sqrt(z)
 *	and
 *		asin(x) = pi/2 - 2*(s+s*z*R(z))
 *			= pio4_hi+(pio4-2s)-(2s*z*R(z)-pio2_lo)
 *			= pio4_hi+(pio4-2f)-(2s*z*R(z)-(pio2_lo+2c))
 *
 * Special cases:
 *	if x is NaN, return x itself;
 *	if |x|>1, return NaN with invalid signal.
 *
 */
static const double
pio4_hi =  7.85398163397448278999e-01; /* 0x3FE921FB, 0x54442D18 */

LUALIB_API double sun_asin (double x) {  /* as fast as GCC's acos implementation */
  double t, w, p, q, c, r, s;
  int32_t hx, lx, ix;
  __GET_HI(hx, x);
  ix = hx & 0x7fffffff;
  if (ix > 0x3ff00000) return AGN_NAN;  /* |x| > 1 ? modified */
  __GET_LO(lx, x);
  if (((ix - 0x3ff00000) | lx) == 0) {  /* |x|==1 */
    /* asin(1)=+-pi/2 with inexact */
    return x*pio2_hi + x*pio2_lo;
  } else if (ix < 0x3fe00000) {	/* |x|<0.5 */
    if (ix < 0x3e400000) {		/* if |x| < 2**-27 */
      t = 0;  /* avoid compiler warning */
      if (huge + x > one) return x;/* return x with inexact if x!=0*/
	} else {
	  t = x*x;
	}
    p = t*(pS0 + t*(pS1 + t*(pS2 + t*(pS3 + t*(pS4 + t*pS5)))));
    q = one + t*(qS1 + t*(qS2 + t*(qS3 + t*qS4)));
    w = p/q;
    return x + x*w;
  }
  /* 1> |x|>= 0.5 */
  w = one - fabs(x);
  t = w*0.5;
  p = t*(pS0 + t*(pS1 + t*(pS2 + t*(pS3 + t*(pS4 + t*pS5)))));
  q = one + t*(qS1 + t*(qS2 + t*(qS3 + t*qS4)));
  s = sqrt(t);
  if (ix >= 0x3FEF3333) { 	/* if |x| > 0.975 */
    w = p/q;
    t = pio2_hi - (2.0*(s + s*w) - pio2_lo);
  } else {
    w  = s;
    __SET_LO(w, 0);
    c  = (t - w*w)/(s + w);
    r  = p/q;
    p  = 2.0*s*r - (pio2_lo - 2.0*c);
    q  = pio4_hi - 2.0*w;
    t  = pio4_hi - (p - q);
  }
  if (hx > 0) return t; else return -t;
}


/* @(#)e_pow.c 1.5 04/04/22 SMI */
/*
 * ====================================================
 * Copyright (C) 2004 by Sun Microsystems, Inc. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */
/* __ieee754_pow(x,y) return x**y
 *
 *		      n
 * Method:  Let x =  2   * (1+f)
 *	1. Compute and return log2(x) in two pieces:
 *		log2(x) = w1 + w2,
 *	   where w1 has 53-24 = 29 bit trailing zeros.
 *	2. Perform y*log2(x) = n+y' by simulating muti-precision
 *	   arithmetic, where |y'|<=0.5.
 *	3. Return x**y = 2**n*exp(y'*log2)
 *
 * Special cases:
 *	1.  (anything) ** 0  is 1
 *	2.  (anything) ** 1  is itself
 *	3.  (anything) ** NAN is NAN
 *	4.  NAN ** (anything except 0) is NAN
 *	5.  +-(|x| > 1) **  +INF is +INF
 *	6.  +-(|x| > 1) **  -INF is +0
 *	7.  +-(|x| < 1) **  +INF is +0
 *	8.  +-(|x| < 1) **  -INF is +INF
 *	9.  +-1         ** +-INF is NAN
 *	10. +0 ** (+anything except 0, NAN)               is +0
 *	11. -0 ** (+anything except 0, NAN, odd integer)  is +0
 *	12. +0 ** (-anything except 0, NAN)               is +INF
 *	13. -0 ** (-anything except 0, NAN, odd integer)  is +INF
 *	14. -0 ** (odd integer) = -( +0 ** (odd integer) )
 *	15. +INF ** (+anything except 0,NAN) is +INF
 *	16. +INF ** (-anything except 0,NAN) is +0
 *	17. -INF ** (anything)  = -0 ** (-anything)
 *	18. (-anything) ** (integer) is (-1)**(integer)*(+anything**integer)
 *	19. (-anything except 0 and inf) ** (non-integer) is NAN
 *
 * Accuracy:
 *	pow(x,y) returns x**y nearly rounded. In particular
 *			pow(integer,integer)
 *	always returns the correct integer provided it is
 *	representable.
 *
 * Constants :
 * The hexadecimal values are the intended ones for the following
 * constants. The decimal values may be used, provided that the
 * compiler will convert from decimal to binary accurately enough
 * to produce the hexadecimal values shown.
 */

static const double
bp[] = {1.0, 1.5,},
dp_h[] = { 0.0, 5.84962487220764160156e-01,}, /* 0x3FE2B803, 0x40000000 */
dp_l[] = { 0.0, 1.35003920212974897128e-08,}, /* 0x3E4CFDEB, 0x43CFD006 */
two53	=  9007199254740992.0,	/* 0x43400000, 0x00000000 */
	/* poly coefs for (3/2)*(log(x)-2s-2/3*s**3 */
L1  =  5.99999999999994648725e-01, /* 0x3FE33333, 0x33333303 */
L2  =  4.28571428578550184252e-01, /* 0x3FDB6DB6, 0xDB6FABFF */
L3  =  3.33333329818377432918e-01, /* 0x3FD55555, 0x518F264D */
L4  =  2.72728123808534006489e-01, /* 0x3FD17460, 0xA91D4101 */
L5  =  2.30660745775561754067e-01, /* 0x3FCD864A, 0x93C9DB65 */
L6  =  2.06975017800338417784e-01, /* 0x3FCA7E28, 0x4A454EEF */
lg2  =  6.93147180559945286227e-01, /* 0x3FE62E42, 0xFEFA39EF */
lg2_h  =  6.93147182464599609375e-01, /* 0x3FE62E43, 0x00000000 */
lg2_l  = -1.90465429995776804525e-09, /* 0xBE205C61, 0x0CA86C39 */
ovt =  8.0085662595372944372e-0017, /* -(1024-log2(ovfl+.5ulp)) */
cp    =  9.61796693925975554329e-01, /* 0x3FEEC709, 0xDC3A03FD =2/(3ln2) */
cp_h  =  9.61796700954437255859e-01, /* 0x3FEEC709, 0xE0000000 =(float)cp */
cp_l  = -7.02846165095275826516e-09, /* 0xBE3E2FE0, 0x145B01F5 =tail of cp_h*/
ivln2    =  1.44269504088896338700e+00, /* 0x3FF71547, 0x652B82FE =1/ln2 */
ivln2_h  =  1.44269502162933349609e+00, /* 0x3FF71547, 0x60000000 =24b 1/ln2*/
ivln2_l  =  1.92596299112661746887e-08; /* 0x3E54AE0B, 0xF85DDF44 =1/ln2 tail*/

LUALIB_API INLINE double sun_pow (double x, double y) {  /* 1.3 times faster */
  double z, ax, z_h, z_l, p_h, p_l, y1, t1, t2, r, s, t, u, v, w;
  int32_t i, j, k, yisint, n, hx, hy, ix, iy;
  uint32_t lx, ly;
  if (x == 0 && y <= 0 && trunc(y) == y) return AGN_NAN;  /* modified 2.11.1 */
  __EXTRACT_WORDS(hx, lx, x);
  __EXTRACT_WORDS(hy, ly, y);
  ix = hx & 0x7fffffff;  iy = hy & 0x7fffffff;
  /* y==zero: x**0 = 1 */
  if ((iy | ly) == 0) return one;
  /* x==1: 1**y = 1, even if y is NaN */
  if (hx == 0x3ff00000 && lx == 0) return one;
    /* y!=zero: result is NaN if either arg is NaN */
  if (ix > 0x7ff00000 || ((ix == 0x7ff00000) && (lx != 0)) ||
      iy > 0x7ff00000 || ((iy == 0x7ff00000) && (ly != 0)))
    return (x + 0.0) + (y + 0.0);
    /* determine if y is an odd int when x < 0
     * yisint = 0	... y is not an integer
     * yisint = 1	... y is an odd int
     * yisint = 2	... y is an even int
     */
  yisint  = 0;
  if (hx < 0) {
    if (iy >= 0x43400000) yisint = 2; /* even integer y */
    else if (iy >= 0x3ff00000) {
      k = (iy >> 20) - 0x3ff;	   /* exponent */
      if (k > 20) {
        j = ly >> (52 - k);
        if ((j << (52 - k)) == ly) yisint = 2 - (j & 1);
      } else if (ly == 0) {
        j = iy >> (20 - k);
        if((j << (20 - k)) == iy) yisint = 2 - (j & 1);
      }
    }
  }
  /* special value of y */
  if (ly == 0) {
    if (unlikely(iy == 0x7ff00000)) {	/* y is +-inf, 2.5.15 optimisation */
      if (((ix - 0x3ff00000) | lx) == 0)
        return one;	/* (-1)**+-inf is NaN */
      else if (ix >= 0x3ff00000)/* (|x|>1)**+-inf = inf,0 */
        return (hy >= 0) ? y : zero;
      else			/* (|x|<1)**-,+inf = inf,0 */
        return (hy < 0) ? -y : zero;
    }
    if (iy == 0x3ff00000) {	/* y is  +-1 */
      if (hy < 0) return one/x; else return x;
    }
    if (hy == 0x40000000) return x*x;   /* y is  2 */
    if (hy == 0x40080000) return x*x*x; /* y is  3 */
    if (hy == 0x40100000) {             /* y is  4 */
      u = x*x;
      return u*u;
    }
    if (hy == 0x3fe00000) {             /* y is  0.5 */
      if (hx >= 0)	/* x >= +0 */
        return sqrt(x);
    }
  }
  ax = fabs(x);
  /* special value of x */
  if (lx == 0) {
    if (ix == 0x7ff00000 || ix == 0|| ix == 0x3ff00000) {
      z = ax;			/*x is +-0,+-inf,+-1*/
      if (hy < 0) z = one/z;	/* z = (1/|x|) */
      if (hx < 0) {
        if (((ix - 0x3ff00000) | yisint) == 0) {
          z = (z - z)/(z - z);  /* (-1)**non-int is NaN */
        } else if (yisint == 1)
          z = -z;		/* (x<0)**odd = -(|x|**odd) */
      }
      return z;
    }
  }
  /* CYGNUS LOCAL + fdlibm-5.3 fix: This used to be
	n = (hx>>31)+1;
       but ANSI C says a right shift of a signed negative quantity is
       implementation defined.  */
  n = ((uint32_t)hx >> 31) - 1;
  /* (x<0)**(non-int) is NaN */
  if ((n | yisint) == 0) return (x - x)/(x - x);
  s = one; /* s (sign of result -ve**odd) = -1 else = 1 */
  if ((n | (yisint - 1)) == 0) s = -one; /* (-ve)**(odd int) */
  /* |y| is huge */
  if (iy > 0x41e00000) { /* if |y| > 2**31 */
    if (iy > 0x43f00000) {	/* if |y| > 2**64, must o/uflow */
      if (ix <= 0x3fefffff) return (hy < 0) ? huge*huge : tiny*tiny;
      if (ix >= 0x3ff00000) return (hy > 0) ? huge*huge : tiny*tiny;
    }
    /* over/underflow if x is not close to one */
    if (ix < 0x3fefffff) return (hy < 0) ? s*huge*huge : s*tiny*tiny;
    if (ix > 0x3ff00000) return (hy > 0) ? s*huge*huge : s*tiny*tiny;
    /* now |1-x| is tiny <= 2**-20, suffice to compute
	   log(x) by x-x^2/2+x^3/3-x^4/4 */
    t = ax - one;		/* t has 20 trailing zeros */
    w = (t*t)*(0.5 - t*(0.3333333333333333333333 - t*0.25));
    u = ivln2_h*t;	/* ivln2_h has 21 sig. bits */
    v = t*ivln2_l - w*ivln2;
    t1 = u + v;
    __SET_LO(t1, 0);
    t2 = v - (t1 - u);
  } else {
    double ss, s2, s_h, s_l, t_h, t_l;
    n = 0;
    /* take care subnormal number */
    if (ix < 0x00100000) {
      ax *= two53; n -= 53; __GET_HI(ix, ax);
    }
    n  += ((ix) >> 20) - 0x3ff;
    j  = ix & 0x000fffff;
    /* determine interval */
    ix = j | 0x3ff00000;		/* normalize ix */
    if (j <= 0x3988E) k = 0;		/* |x|<sqrt(3/2) */
    else if (j < 0xBB67A) k = 1;	/* |x|<sqrt(3)   */
    else { k=0; n += 1; ix -= 0x00100000; }
    __SET_HI(ax, ix);
    /* compute ss = s_h+s_l = (x-1)/(x+1) or (x-1.5)/(x+1.5) */
    u = ax - bp[k];		/* bp[0]=1.0, bp[1]=1.5 */
    v = one/(ax + bp[k]);
    ss = u*v;
    s_h = ss;
    __SET_LO(s_h, 0);
    /* t_h=ax+bp[k] High */
    t_h = zero;
    __SET_HI(t_h, ((ix >> 1) | 0x20000000) + 0x00080000 + (k << 18));
    t_l = ax - (t_h - bp[k]);
    s_l = v*((u - s_h*t_h) - s_h*t_l);
    /* compute log(ax) */
    s2 = ss*ss;
    r = s2*s2*(L1 + s2*(L2 + s2*(L3 + s2*(L4 + s2*(L5 + s2*L6)))));
    r += s_l*(s_h + ss);
    s2  = s_h*s_h;
    t_h = 3.0 + s2 + r;
    __SET_LO(t_h, 0);
    t_l = r - ((t_h - 3.0) - s2);
    /* u+v = ss*(1+...) */
    u = s_h*t_h;
    v = s_l*t_h + t_l*ss;
    /* 2/(3log2)*(ss+...) */
    p_h = u + v;
    __SET_LO(p_h, 0);
    p_l = v - (p_h - u);
    z_h = cp_h*p_h;		/* cp_h+cp_l = 2/(3*log2) */
    z_l = cp_l*p_h + p_l*cp + dp_l[k];
    /* log2(ax) = (ss+..)*2/(3*log2) = n + dp_h + z_h + z_l */
    t = (double)n;
    t1 = (((z_h + z_l) + dp_h[k]) + t);
    __SET_LO(t1, 0);
    t2 = z_l-(((t1 - t) - dp_h[k]) - z_h);
  }
  /* split up y into y1+y2 and compute (y1+y2)*(t1+t2) */
  y1  = y;
  __SET_LO(y1, 0);
  p_l = (y - y1)*t1 + y*t2;
  p_h = y1*t1;
  z = p_l + p_h;
  __EXTRACT_WORDS(j, i, z);
  if (j >= 0x40900000) {				/* z >= 1024 */
    if (((j - 0x40900000) | i) != 0)			/* if z > 1024 */
      return s*huge*huge;			/* overflow */
    else {
      if (p_l + ovt > z - p_h) return s*huge*huge;	/* overflow */
    }
  } else if ((j & 0x7fffffff) >= 0x4090cc00 ) {	/* z <= -1075 */
    if (((j - 0xc090cc00) | i) != 0) 		/* z < -1075 */
      return s*tiny*tiny;		/* underflow */
    else {
      if (p_l <= z - p_h) return s*tiny*tiny;	/* underflow */
    }
  }
  /*
   * compute 2**(p_h+p_l)
   */
  i = j & 0x7fffffff;
  k = (i >> 20) - 0x3ff;
  n = 0;
  if (i > 0x3fe00000) {		/* if |z| > 0.5, set n = [z+0.5] */
    n = j + (0x00100000 >> (k + 1));
    k = ((n & 0x7fffffff) >> 20) - 0x3ff;	/* new k for n */
    t = zero;
    __SET_HI(t, n &~ (0x000fffff >> k));
    n = ((n & 0x000fffff) | 0x00100000) >> (20 - k);
    if (j < 0) n = -n;
    p_h -= t;
  }
  t = p_l + p_h;
  __SET_LO(t, 0);
  u = t*lg2_h;
  v = (p_l - (t - p_h))*lg2 + t*lg2_l;
  z = u + v;
  w = v - (z - u);
  t  = z*z;
  t1  = z - t*(P1 + t*(P2 + t*(P3 + t*(P4 + t*P5))));
  r  = (z*t1)/(t1 - two) - (w + z*w);
  z  = one - (r - z);
  __GET_HI(j, z);
  j += (n << 20);
  if ((j >> 20) <= 0) z = sun_ldexp(z, n);	/* subnormal output */
  else __SET_HI(z, j);
  return s*z;
}


/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/*
 * ceil(x)
 * Return x rounded toward -inf to integral value
 * Method:
 *	Bit twiddling.
 * Exception:
 *	Inexact flag raised if x not equal to ceil(x).
 */

LUALIB_API double sun_ceil (double x) {
  int32_t i0, i1, j0;
  uint32_t i, j;
  __EXTRACT_WORDS(i0, i1, x);
  j0 = ((i0 >> 20) & 0x7ff) - 0x3ff;
  if (j0 < 20) {
    if (j0 < 0) { 	/* raise inexact if x != 0 */
      if (huge + x > 0.0) {/* return 0*sign(x) if |x|<1 */
        if (i0 < 0) { i0 = 0x80000000; i1 = 0; }
        else if ((i0 | i1) != 0) { i0 = 0x3ff00000; i1 = 0; }
	  }
    } else {
      i = (0x000fffff) >> j0;
      if (((i0 & i) | i1) == 0) return x; /* x is integral */
      if (huge + x > 0.0) {	/* raise inexact flag */
        if (i0 > 0) i0 += (0x00100000) >> j0;
        i0 &= (~i); i1 = 0;
      }
    }
  } else if (j0 > 51) {
    if (j0 == 0x400) return x + x;	/* inf or NaN */
    else return x;		/* x is integral */
  } else {
    i = ((uint32_t)(0xffffffff)) >> (j0 - 20);
	if ((i1 & i) == 0) return x;	/* x is integral */
	if (huge + x > 0.0) { 		/* raise inexact flag */
      if (i0 > 0) {
        if (j0 == 20) i0 += 1;
        else {
          j = i1 + (1 << (52 - j0));
          if (j < i1) i0 += 1;	/* got a carry */
          i1 = j;
        }
      }
      i1 &= (~i);
    }
  }
  __INSERT_WORDS(x, i0, i1);
  return x;
}


/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */
/*
 * floor(x)
 * Return x rounded toward -inf to integral value
 * Method:
 *	Bit twiddling.
 * Exception:
 *	Inexact flag raised if x not equal to floor(x).
 */

LUALIB_API double sun_floor (double x) {
  int32_t i0, i1, j0;
  uint32_t i, j;
  __EXTRACT_WORDS(i0, i1, x);
  j0 = ((i0 >> 20) & 0x7ff) - 0x3ff;
  if (j0 < 20) {
    if (j0 < 0) { 	/* raise inexact if x != 0 */
      if (huge + x > 0.0) {  /* return 0*sign(x) if |x|<1 */
        if (i0 >= 0)
          { i0 = i1 = 0; }
        else if (((i0 & 0x7fffffff) | i1) != 0)
          { i0 = 0xbff00000; i1 = 0; }
      }
    } else {
      i = (0x000fffff) >> j0;
      if (((i0 & i) | i1) == 0) return x;  /* x is integral */
      if (huge + x > 0.0) {  /* raise inexact flag */
        if (i0 < 0) i0 += (0x00100000 >> j0);
        i0 &= (~i); i1 = 0;
      }
    }
  } else if (j0 > 51) {
    if (j0 == 0x400) return x + x;	/* inf or NaN */
    else return x;		/* x is integral */
  } else {
    i = ((uint32_t)(0xffffffff)) >> (j0 - 20);
    if ((i1 & i) == 0) return x;  /* x is integral */
    if (huge + x > 0.0) {         /* raise inexact flag */
      if (i0 < 0) {
        if (j0 == 20) i0 += 1;
        else {
          j = i1 + (1 << (52 - j0));
          if (j < i1) i0 += 1;    /* got a carry */
          i1 = j;
        }
      }
      i1 &= (~i);
    }
  }
  __INSERT_WORDS(x, i0, i1);
  return x;
}


/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */
/* ilogb(double x)
 * return the binary exponent of x
 * ilogb(+-0) = undefined (formerly FP_ILOGB0)
 * ilogb(+-inf) = HUGE_VAL (formerly INT_MAX)
 * ilogb(NaN) = AGN_NAN (formerly FP_ILOGBNAN, no signal is raised)
 */
LUALIB_API double sun_ilogb (double x) {
  int32_t hx, lx, ix;
  __GET_HI(hx, x);
  hx &= 0x7fffffff;
  if (hx < 0x00100000) {
    __GET_LO(lx, x);
    if ((hx | lx) == 0) return AGN_NAN;  /* 2.11.1 fix, +-0, ilogb(0) = undefined */
    /* subnormal x */
    ix = -1043;
    if (hx != 0) {
      ix = -1022;
      lx = (hx << 11);
    }
    /* each leading zero mantissa bit makes exponent smaller */
    for (; lx > 0; lx <<= 1) ix--;
    return ix;
  }
  if (hx < 0x7ff00000) /* normal x */
    return (hx >> 20) - 1023;
  if (FP_ILOGBNAN != (~0U >> 1)) {
    __GET_LO(lx, x);
    if (hx == 0x7ff00000 && lx == 0)  /* +-inf */
      return HUGE_VAL;  /* 2.11.1 fix, ~0U >> 1; */ /* = INT_MAX */
  }
  /* NAN. ilogb(NAN) = FP_ILOGBNAN */
  return AGN_NAN;  /* 2.11.1 fix, FP_ILOGBNAN; */
}


/* @(#)s_cbrt.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 *
 * Optimized by Bruce D. Evans.
 *
 * cbrt(x)
 * Return cube root of x
 */
static const uint32_t
B1 = 715094163, /* B1 = (1023-1023/3-0.03306235651)*2**20 */
B2 = 696219795; /* B2 = (1023-1023/3-54/3-0.03306235651)*2**20 */

/* |1/cbrt(x) - p(x)| < 2**-23.5 (~[-7.93e-8, 7.929e-8]). */
static const double
PP0 =  1.87595182427177009643,		/* 0x3ffe03e6, 0x0f61e692 */
PP1 = -1.88497979543377169875,		/* 0xbffe28e0, 0x92f02420 */
PP2 =  1.621429720105354466140,		/* 0x3ff9f160, 0x4a49d6c2 */
PP3 = -0.758397934778766047437,		/* 0xbfe844cb, 0xbee751d9 */
PP4 =  0.145996192886612446982;		/* 0x3fc2b000, 0xd4e4edd7 */

LUALIB_API double sun_cbrt (double x) {
  int32_t hx;
  union {
    double value;
    uint64_t bits;
  } u;
  double r, s, t = 0.0, w;
  uint32_t sign;
  uint32_t high, low;
  __EXTRACT_WORDS(hx, low, x);
  sign = hx & 0x80000000; 		/* sign= sign(x) */
  hx ^= sign;
  if (unlikely(hx >= 0x7ff00000)) return(x + x);  /* cbrt(NaN,INF) is itself, 2.5.15 optimisation */
  /*
   * Rough cbrt to 5 bits:
   *    cbrt(2**e*(1+m) ~= 2**(e/3)*(1+(e%3+m)/3)
   * where e is integral and >= 0, m is real and in [0, 1), and "/" and
   * "%" are integer division and modulus with rounding towards minus
   * infinity.  The RHS is always >= the LHS and has a maximum relative
   * error of about 1 in 16.  Adding a bias of -0.03306235651 to the
   * (e%3+m)/3 term reduces the error to about 1 in 32. With the IEEE
   * floating point representation, for finite positive normal values,
   * ordinary integer divison of the value in bits magically gives
   * almost exactly the RHS of the above provided we first subtract the
   * exponent bias (1023 for doubles) and later add it back.  We do the
   * subtraction virtually to keep e >= 0 so that ordinary integer
   * division rounds towards minus infinity; this is also efficient.
   */
  if (hx < 0x00100000) { 		/* zero or subnormal? */
    if( (hx | low) == 0)
      return(x);		/* cbrt(0) is itself */
    __SET_HI(t, 0x43500000); /* set t= 2**54 */
    t *= x;
    __GET_HI(high, t);
    __INSERT_WORDS(t, sign | ((high & 0x7fffffff)/3 + B2), 0);
  } else
    __INSERT_WORDS(t, sign | (hx/3 + B1), 0);
  /*
   * New cbrt to 23 bits:
   *    cbrt(x) = t*cbrt(x/t**3) ~= t*P(t**3/x)
   * where P(r) is a polynomial of degree 4 that approximates 1/cbrt(r)
   * to within 2**-23.5 when |r - 1| < 1/10.  The rough approximation
   * has produced t such than |t/cbrt(x) - 1| ~< 1/32, and cubing this
   * gives us bounds for r = t**3/x.
   *
   * Try to optimize for parallel evaluation as in k_tanf.c.
   */
  r = (t*t)*(t/x);
  t = t*((PP0 + r*(PP1 + r*PP2)) + ((r*r)*r)*(PP3 + r*PP4));
  /*
   * Round t away from zero to 23 bits (sloppily except for ensuring that
   * the result is larger in magnitude than cbrt(x) but not much more than
   * 2 23-bit ulps larger).  With rounding towards zero, the error bound
   * would be ~5/6 instead of ~4/6.  With a maximum error of 2 23-bit ulps
   * in the rounded t, the infinite-precision error in the Newton
   * approximation barely affects third digit in the final error
   * 0.667; the error in the rounded t can be up to about 3 23-bit ulps
   * before the final error is larger than 0.667 ulps.
   */
  u.value = t;
  u.bits = (u.bits + 0x80000000) & 0xffffffffc0000000ULL;
  t = u.value;
  /* one step Newton iteration to 53 bits with error < 0.667 ulps */
  s = t*t;				/* t*t is exact */
  r = x/s;				/* error <= 0.5 ulps; |r| < |t| */
  w = t + t;				/* t+t is exact */
  r = (r - t)/(w + r);			/* r-t is exact; w+r ~= 3*t */
  t = t + t*r;			/* error <= 0.5 + 0.5/3 + epsilon */
  return t;
}


/* Pcrocess int's ONLY. 20 % faster than the `iqr` function. d must not be 0 !
   Returns both quotient = n / d and remainder = math.modulus(n, d) for integers n and d. If n is non-negative and the computed
   remainder n % d is negative, then quotient = quotient + 1 and remainder = remainder - d. Likewise, if n is negative and the
   computed remainder is positive, then quotient = quotient - 1 and remainder = remainder + d. */
LUALIB_API div_t tools_div (int n, int d) {  /* 2.11.3 */
  div_t r;
  r.quot = n/d;
  r.rem =  n%d;
  if (n >= 0 && r.rem < 0) {
    r.quot++;
    r.rem -= d;
  } else if (n < 0 && r.rem > 0) {
    r.quot--;
    r.rem += d;
  }
  return (r);
}


/* Taken from: https://graphics.stanford.edu/~seander/bithacks.html
   four percent faster than graphics.sanford.edu's version without table look-up
   'Find the log base 2 of an N-bit integer in O(lg(N)) operations'. */
static const int MultiplyDeBruijnBitPosition[32] =  {
  0, 9, 1, 10, 13, 21, 2, 29, 11, 14, 16, 18, 22, 25, 3, 30,
  8, 12, 20, 28, 15, 17, 24, 7, 19, 27, 23, 6, 26, 5, 4, 31
};

LUALIB_API int tools_uintlog2 (uint32_t v) {  /* 2.14.0 */
  v |= v >> 1;  /* first round down to one less than a power of 2 */
  v |= v >> 2;
  v |= v >> 4;
  v |= v >> 8;
  v |= v >> 16;
  return MultiplyDeBruijnBitPosition[(uint32_t)(v * 0x07C4ACDDU) >> 27];
}


LUALIB_API float tools_lbfast (float x) {  /* 2.14.6, 18 percent faster than sun_log. */
  ieee_float_shape_type p, q;
  p.value = x;
  q.word = (p.word & 0x007FFFFF) | 0x3f000000;
  x = (float)p.word * 1.1920928955078125e-7f;
  return x - 124.22551499f - 1.498030302f * q.value  - 1.72587999f / (0.3520887068f + q.value);
  /* multiply result by 0.69314718f = (float)LN2 to get natural logarithm ln(x) */
}

/* Taken from: http://aggregate.org/MAGIC/#Absolute%20Value%20of%20a%20Float
   Population count of set bits (one bits) */
LUALIB_API INLINE uint32_t tools_onebits (register uint32_t x) {
#ifdef __GNUC__
  return __builtin_popcount(x);  /* 2.14.12 */
#else
  x -= ((x >> 1) & 0x55555555);
  x = (((x >> 2) & 0x33333333) + (x & 0x33333333));
  x = (((x >> 4) + x) & 0x0f0f0f0f);
  x += (x >> 8);
  x += (x >> 16);
  return(x & 0x0000003f);
#endif
}


/* flip n */
LUALIB_API INLINE uint32_t tools_flip (register uint32_t n) {
  int i, r = 0;
  for (i=0; i < 8 * sizeof(uint32_t); i++) {
    r += (n & 0x0001);
    r <<= 1;
    n >>= 1;
  }
  return r;
}

/* Remove trailing zeros from a uint32_t, 2.11.3
   Source: https://stackoverflow.com/questions/10343730/strip-leading-zeroes-in-binary-number-in-c-c */
LUALIB_API INLINE int tools_remtrailzeros (register uint32_t n) {
  int flipped = tools_flip(n);
  /* shift to remove trailing zeroes */
  while (!(flipped & 0x0001))
    flipped >>= 1;
  return flipped;
}


/* Count number of leading zeros (clz32). 2.11.3
   Source: http://aggregate.org/MAGIC/#Leading%20Zero%20Count */
LUALIB_API INLINE uint32_t tools_clz32 (register uint32_t x, uint32_t *y) {  /* 2.16.8 fix */
  x |= (x >> 1);
  x |= (x >> 2);
  x |= (x >> 4);
  x |= (x >> 8);
  x |= (x >> 16);
  *y = x;
#ifdef __GNUC__
  return (8 * sizeof(uint32_t) - __builtin_popcount(x));
#else
  return (8 * sizeof(uint32_t) - tools_onebits(x));
#endif
}


/* Count number of leading zeros (clz32). 2.14.10
   Source:  musl-1.1.22/src/internal/atomic.h */
LUALIB_API INLINE uint32_t tools_ctz32 (register uint32_t x, uint32_t *y) {  /* 2.16.8 fix */
  uint32_t clz = tools_clz32(x & -x, y);
  /* mask y ! */
  return 8*sizeof(uint32_t) - (x != 0)*(1 + clz);
}


/* See: https://stackoverflow.com/questions/2589096/find-most-significant-bit-left-most-that-is-set-in-a-bit-array, #1, by Jeff */
LUALIB_API INLINE uint32_t tools_msb (register uint32_t x) {
  int r = 0;
  if (x < 1) return 0;
  while (x >>= 1) r++;
  return r + 1;
}


LUALIB_API INLINE uint32_t tools_lsb (register uint32_t x) {  /* 2.12.3, least-significant bit */
  if (x < 1) return 0;
  x &= ~x + 1;  /* clears all but the lowest bit of x */
  return tools_uintlog2(x) + 1;
}


/* Modulo operator for unsigned ints, slower than C's `%' on x86 CPUs, 2.12.0 RC 4
   Source: https://stackoverflow.com/questions/24219482/alternative-of-modulo-operator-in-embedded-c */
LUALIB_API unsigned int tools_modulo (unsigned int num, unsigned int div) {
  if (div == 0) return -1;  /* avoid stack overflows, i.e. segfaults */
  if (num - div >= div)
    num = tools_modulo(num, div + div);
  while (num >= div)
    num -= div;
  return num;
}


#define _PHIM1  (PHI - 1.0)
#define _W      (4294967296.0)
#define _SI     (_W * _PHIM1)
/* See: https://de.wikipedia.org/wiki/Multiplikative_Methode */
LUALIB_API INLINE uint32_t tools_fibmod (uint32_t k, uint32_t m, int exact) {  /* 2.14.8 */
  if (exact) {
     return floor( ( m * fmod(_SI*k, _W) )/_W);
  } else {  /* "double p = k*_PHIM1; floor(m*(p - floor(p)))" is slower */
     double p = (double)k*_PHIM1;
     return m*(p - (uint32_t)(p));  /* 63 % faster but result may vary across platforms */
  }
  /* result is in the range 0 .. m - 1 if m is odd, and in 1 .. m - 1 otherwise, unless m = 0 or k = 0 where 0 is returned */
}


/* @(#)s_asinh.c 5.1 93/09/24 */
/*
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 *
 * asinh(x)
 * Method :
 *	Based on
 *		asinh(x) = sign(x) * log [ |x| + sqrt(x*x+1) ]
 *	we have
 *	asinh(x) := x  if  1+x*x=1,
 *		 := sign(x)*(log(x)+ln2)) for large |x|, else
 *		 := sign(x)*log(2|x|+1/(|x|+sqrt(x*x+1))) if|x|>2, else
 *		 := sign(x)*log1p(|x| + x^2/(1 + sqrt(1+x^2)))
 */

LUALIB_API double sun_asinh (double x) {  /* 2.14.13, 12 % faster than GCC's asinh */
  double t, w;
  int32_t hx, ix;
  __GET_HI(hx, x);
  ix = hx & 0x7fffffff;
  if (unlikely(ix >= 0x7ff00000)) return x + x;  /* x is inf or NaN, 2.5.15 optimisation */
  if (ix < 0x3e300000) {  /* |x| < 2**-28 */
    if(huge + x > one) return x;  /* return x inexact except 0 */
  }
  if (ix > 0x41b00000) {         /* |x| > 2**28 */
    w = sun_log(fabs(x)) + LN2;
  } else if (ix > 0x40000000) {  /* 2**28 > |x| > 2.0 */
    t = fabs(x);
    w = sun_log(2.0*t + one/(sqrt(x*x + one) + t));
  } else {  /* 2**-28 < |x| < 2.0 */
    t = x*x;
    w = log1p(fabs(x) + t/(one + sqrt(one + t)));  /* DO NOT use kernel_log1p, it returns wrong results in this context ! */
  }
  if (hx > 0) return w; else return -w;
}


LUALIB_API int tools_gcd (double x, double y) {
  int p, q, t;
  if (unlikely(tools_isfloat(x) || tools_isfloat(y))) return 1;  /* 2.5.15 optimisation */
  if (y == 0) return (x < 0) ? -x : x;
  p = x; q = y;
  while (q != 0) {
    t = p;
    p = q;
    q = t % q;
  }
  return (p < 0) ? -p : p;
}


LUALIB_API int tools_anykey (void) {
  int i;
#if defined(_WIN32)
  i = kbhit();
  if (i) { getch(); }  /* 0.31.4, `clear the buffer` */
#elif defined(__OS2__) || defined(__unix__) || defined(__APPLE__)
  i = kbhit();
#else
  i = -1;
#endif
  return i;
}


/* suited for negative values as well, 2.12.6, taken from: http://www.strudel.org.uk/itoa/ */
LUALIB_API char *tools_itoa (int64_t val, int base) {
  int i, isneg;
  static char buf[33] = { 0 };  /* NULL terminator, changed 2.12.7 for possible minus */
  isneg = 0;
  i = 31;  /* changed 2.12.7 for possible minus */
  if (val == 0) buf[i--] = '0';  /* 2.12.7 extension */
  else if (val < 0) { val = -val; isneg = 1; }  /* 2.12.7 extension */
  if (base < 1 || val > INT_MAX) return NULL;   /* 2.12.7 extension */
  for (; val && i ; --i, val /= base)
    buf[i] = "0123456789abcdef"[val % base];
  if (isneg) buf[i--] = '-';  /* 2.12.7 extension */
  return &buf[i + 1];
}


LUALIB_API double tools_koadd (volatile double s, volatile double x, volatile double *q) {  /* 2.13.0 */
  volatile lua_Number sold, u, v, w, t;
  v = x - *q;
  sold = s;
  s += v;
  if (fabs(x) < fabs(*q)) {
    t = x;
    x = -(*q);
    *q = t;
  }
  u = (v - x) + *q;
  if (fabs(sold) < fabs(v)) {
    t = sold;
    sold = v;
    v = t;
  }
  w = (s - sold) - v;
  *q = u + w;
  return s;  /* s now contains sumdata */
}


/* Evaluate polynomial
 *
 * SYNOPSIS:
 *
 * int N;
 * double x, y, coef[N+1], polevl[];
 *
 * y = polevl( x, coef, N );
 *
 * DESCRIPTION:
 *
 * Evaluates polynomial of degree N:
 *
 *                     2          N
 * y  =  C  + C x + C x  +...+ C x
 *        0    1     2          N
 *
 * Coefficients are stored in reverse order:
 *
 * coef[0] = C  , ..., coef[N] = C  .
 *            N                   0
 *
 * SPEED:
 *
 * In the interest of speed, there are no checks for out
 * of bounds arithmetic.  This routine is used by most of
 * the functio  ns in the library.  Depending on available
 * equipment features, the user may wish to rewrite the
 * program in microcode or assembly language.
 *
 */

/*
Cephes Math Library Release 2.1:  December, 1988
Copyright 1984, 1987, 1988 by Stephen L. Moshier
Direct inquiries to 30 Frost Street, Cambridge, MA 02140

Kahan-Ozawa extension by alex walz, 2.13.0 */

LUALIB_API double tools_polyeval (double x, double coeffs[], int n) {
  volatile double r, q, *p;
  p = coeffs;
  r = *p++;
  q = 0;
  do
    r = tools_koadd(r*x, *p++, &q);
  while (--n);
  return r + q;
}


/* 2.13.0, based on Sun's rint, s_rint.c, due to call overhead, only slightly faster than ISFLOAT macro. */
LUALIB_API INLINE int tools_isfloat (double x) {
  int32_t hx, ux;
  uint32_t i, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;  /* unbiased exponent: math.exponent(x) - 1, except 0 -> -1023 */
  if (unlikely(ux > 51))  /* very large number not correctly representible any more ?  2.5.15 optimisation */
    return 0;   /* covers +/-inf, nan, as well. */
  if (ux < 20) {
    if (ux < 0) {  /* |x| < 1 */
      return ((hx & 0x7fffffff) | lx) != 0;
    } else {
      i = (0x000fffff) >> ux;
      return ((hx & i) | lx) != 0;
	}
  } else {
    i = ((uint32_t)(0xffffffff)) >> (ux - 20);
    return (lx & i) != 0;
  }
}


/* 2.13.0, based on Sun's rint, s_rint.c, due to call overhead, only slightly faster than ISINT macro. */
LUALIB_API INLINE int tools_isint (double x) {
  int32_t hx, ux;
  uint32_t i, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;  /* unbiased exponent: math.exponent(x) - 1, except 0 -> -1023 */
  if (unlikely(ux > 51)) /* value not correctly representable, not +/-inf, nan ?  2.5.15 optimisation */
    /* (hx & 0x7fffffff) < 0x7ff00000) -> real number, but not +/-inf, nan */
    return ((hx & 0x7fffffff) < 0x7ff00000);  /* with +/-inf, nan, return 0 */
  if (ux < 20) {
    if (ux < 0) {  /* |x| < 1 */
      return ((hx & 0x7fffffff) | lx) == 0;
    } else {
      i = (0x000fffff) >> ux;
      return ((hx & i) | lx) == 0;
    }
  } else {
    i = ((uint32_t)(0xffffffff)) >> (ux - 20);
    return (lx & i) == 0;
  }
}


/* Checks for an integer that is also correctly representable. 2.16.6 */
LUALIB_API INLINE int tools_isregint (double x) {
  int32_t hx, ux;
  uint32_t i, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;  /* unbiased exponent: math.exponent(x) - 1, except 0 -> -1023 */
  if (unlikely(ux > 51)) {  /* value not correctly representable, not +/-inf, nan ?  2.5.15 optimisation */
    /* (hx & 0x7fffffff) < 0x7ff00000) -> real number, but not +/-inf, nan */
    if ((hx & 0x7fffffff) < 0x7ff00000) return 0;  /* with +/-inf, nan, return 0 */
    /* check for irregular numbers */
    hx &= 0x7fffffff;  /* absolute value */
    if (hx >= 0x43400000) return 0;
    return !(hx == 0x43400000 && lx == 0);
  }
  if (ux < 20) {
    if (ux < 0) {  /* |x| < 1 */
      return ((hx & 0x7fffffff) | lx) == 0;
    } else {
      i = (0x000fffff) >> ux;
      return ((hx & i) | lx) == 0;
    }
  } else {
    i = ((uint32_t)(0xffffffff)) >> (ux - 20);
    return (lx & i) == 0;
  }
}


LUALIB_API INLINE int tools_isregnonnegint (double x) {
  int32_t hx, ux;
  uint32_t i, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;  /* unbiased exponent: math.exponent(x) - 1, except 0 -> -1023 */
  if (unlikely(ux > 51)) {  /* value not correctly representable, not +/-inf, nan ?  2.5.15 optimisation */
    /* (hx & 0x7fffffff) < 0x7ff00000) -> real number, but not +/-inf, nan */
    if ((hx & 0x7fffffff) < 0x7ff00000) return 0;  /* with +/-inf, nan, return 0 */
    /* check for irregular numbers */
    hx &= 0x7fffffff;  /* absolute value */
    if (hx >= 0x43400000) return 0;
    return !(hx == 0x43400000 && lx == 0);
  }
  if (ux < 20) {
    if (ux < 0) {  /* |x| < 1 */
      return (((hx & 0x7fffffff) | lx) == 0) && !(hx >> 31);
    } else {
      i = (0x000fffff) >> ux;
      return (((hx & i) | lx) == 0) && !(hx >> 31);
    }
  } else {
    i = ((uint32_t)(0xffffffff)) >> (ux - 20);
    return ((lx & i) == 0) && !(hx >> 31);
  }
}


/* 2.13.0, based on Sun's rint, s_rint.c, due to call overhead, only slightly faster than ISINT macro. */
LUALIB_API INLINE int tools_isnegint (double x) {
  int32_t hx, ux;
  uint32_t i, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;  /* unbiased exponent: math.exponent(x) - 1, except 0 -> -1023 */
  if (unlikely(ux > 51))  /* very large number not correctly representible any more ?  2.5.15 optimisation */
    /* (hx & 0x7fffffff) < 0x7ff00000) -> real number, but not +/-inf, nan
       (hx >> 31) -> signbit (1 = minus, 0 = zero or +) */
    return ((hx & 0x7fffffff) < 0x7ff00000) && (hx >> 31);
  if (ux < 20) {
    if (ux < 0) {  /* |x| < 1 */
      return 0;
    } else {
      i = (0x000fffff) >> ux;
      return (((hx & i) | lx) == 0) && (hx >> 31);
	}
  } else {
    i = ((uint32_t)(0xffffffff)) >> (ux - 20);
    return ((lx & i) == 0) && (hx >> 31);
  }
}


/* Checks whether x is positive; idea taken from: newlib-3.1.0/newlib/libm/mathfp/s_ispos.c */
LUALIB_API int tools_ispos (double x) {  /* 2.16.1 not exposed */
  uint32_t hx, lx;
  EXTRACT_WORDS(hx, lx, x);
  if (((hx & 0x7fffffff) | ((lx | -lx) >> 31)) > 0x7ff00000) return 0;  /* x is NaN */
  return !(hx & 0x80000000);  /* for x = +/-0, finite or +/-inf */
}


LUALIB_API INLINE int tools_isposint (double x) {  /* 2.13.0 */
  int32_t hx, ux;
  uint32_t i, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;  /* unbiased exponent: math.exponent(x) - 1, except 0 -> -1023 */
  if (unlikely(ux > 51))  /* very large number not correctly representible any more ?  2.5.15 optimisation */
    /* (hx & 0x7fffffff) < 0x7ff00000) -> real number, but not +/-inf, nan
       (hx >> 31) -> signbit (1 = minus, 0 = zero or +) */
    return ((hx & 0x7fffffff) < 0x7ff00000) && !(hx >> 31);  /* 2.16.6 fix */
  if (ux < 20) {
    if (ux < 0) {  /* |x| < 1 */
      return 0;
    } else {
      i = (0x000fffff) >> ux;
      return (((hx & i) | lx) == 0) && !(hx >> 31);
	}
  } else {
    i = ((uint32_t)(0xffffffff)) >> (ux - 20);
    return ((lx & i) == 0) && !(hx >> 31);
  }
}


LUALIB_API INLINE int tools_isnonnegint (double x) {  /* 2.14.13 */
  int32_t hx, ux;
  uint32_t i, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;  /* unbiased exponent: math.exponent(x) - 1, except 0 -> -1023 */
  if (unlikely(ux > 51))  /* very large number not correctly representible any more ?  2.5.15 optimisation */
    /* (hx & 0x7fffffff) < 0x7ff00000) -> real number, but not +/-inf, nan
       (hx >> 31) -> signbit (1 = minus, 0 = zero or +) */
    return ((hx & 0x7fffffff) < 0x7ff00000) && (hx >> 31);
  if (ux < 20) {
    if (((hx & 0x7fffffff) | lx) == 0)  /* x == +-0 */
      return 1;
    if (ux < 0) {  /* |x| < 1 */
      return 0;
    } else {
      i = (0x000fffff) >> ux;
      return (((hx & i) | lx) == 0) && (hx >> 31 == 0);
	}
  } else {
    i = ((uint32_t)(0xffffffff)) >> (ux - 20);
    return ((lx & i) == 0) && (hx >> 31 == 0);
  }
}


LUALIB_API INLINE int tools_isuint (double x, int bits) {  /* 2.16.11, bits: 16 = 2-byte uint, 32 = 4-byte uint, etc. */
  int32_t hx, ux;
  uint32_t i, lx;
  __EXTRACT_WORDS(hx, lx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;  /* unbiased exponent: math.exponent(x) - 1, except 0 -> -1023 */
  if (!(hx & 0x80000000) && ux > bits - 1) return 0;
  if (unlikely(ux > 51))  /* very large number not correctly representible any more ?  2.5.15 optimisation */
    /* (hx & 0x7fffffff) < 0x7ff00000) -> real number, but not +/-inf, nan
       (hx >> 31) -> signbit (1 = minus, 0 = zero or +) */
    return ((hx & 0x7fffffff) < 0x7ff00000) && (hx >> 31);
  if (ux < 20) {
    if (((hx & 0x7fffffff) | lx) == 0)  /* x == +-0 */
      return 1;
    if (ux < 0) {  /* |x| < 1 */
      return 0;
    } else {
      i = (0x000fffff) >> ux;
      return (((hx & i) | lx) == 0) && (hx >> 31 == 0);
	}
  } else {
    i = ((uint32_t)(0xffffffff)) >> (ux - 20);
    return ((lx & i) == 0) && (hx >> 31 == 0);
  }
}


/* LUAI_UINT32 m_w = AGN_RANDOM_MW; */ /* must not be zero, nor 0x464fffff; for definition, see agnconf.h */
/* LUAI_UINT32 m_z = AGN_RANDOM_MZ; */ /* must not be zero, nor 0x9068ffff; for definition, see agnconf.h */

#define int2double	( 1.0/(UINT32_MAX + 1.0) )  /* 2.10.1 fix */
#define bias		( int2double/2.0 )          /* 2.10.1 fix */

/* returns a random number in the open interval (0, 1) */
LUALIB_API INLINE double tools_random (int mode) {
  if (mode) {  /* suggested by Slobodan, 2.5.1: create really random numbers */
    Time64_T tgmt = time(NULL);
    m_w = (tgmt + m_w) % 113 + 1;
    m_z = (tgmt + m_z) % 113 + 1;
  }
  m_z = 36969 * (m_z & 65535) + (m_z >> 16);
  m_w = 18000 * (m_w & 65535) + (m_w >> 16);
  /* 2.10.1 out-of-range fix, DO NOT CHANGE or divide by UINT32_MAX + x; no performance decrease */
  return ((m_z << 16) + m_w)*int2double + bias;
}


/* returns an integer in the range [l, u], 2.14.3 */
LUALIB_API INLINE int tools_randomrange (int l, int u, int mode) {
  double r = tools_random(mode);
  r *= 10000;
  r = (r - trunc(r)) * 10000;
  r = tools_reducerange(r, 0, 1);
  return floor(r*(u - l + 1)) + l;
}


LUALIB_API INLINE ptrdiff_t tools_posrelat (ptrdiff_t pos, size_t len) {  /* 2.14.4 */
  /* relative string position: negative means back from end */
  /* return (pos>=0) ? pos : (ptrdiff_t)len+pos+1; */
  /* Lua 5.1.3 patch 9 */
  if (pos < 0) pos += (ptrdiff_t)len + 1;
  return (pos >= 0) ? pos : 0;
}


/* based on Sun's __ieee754_pow
 *
 * @(#)e_pow.c 5.1 93/09/24
 *
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */
LUALIB_API INLINE int tools_isevenorodd (double x) {
  int32_t j, k, r, hx, ix;
  uint32_t lx;
  __EXTRACT_WORDS(hx, lx, x);
  ix = hx & 0x7fffffff;
  if ((ix | lx) == 0) return 2;    /* x = 0, even, must be checked explicitly */
  if (unlikely(ix >= 0x7ff00000)) return 0;  /* +/-inf or nan, 2.5.15 optimisation */
  /* determine if y is an odd int when x < 0
   * yisint = 0	... y is not an integer
   * yisint = 1	... y is an odd int
   * yisint = 2	... y is an even int
   */
  if (ix >= 0x43400000) return 2;  /* even integer y */
  r = 0;
  if (ix >= 0x3ff00000) {
    k = (ix >> 20) - 0x3ff;	 /* exponent */
    if (k > 20) {
      j = lx >> (52 - k);
      if ((j << (52 - k)) == lx) r = 2 - (j & 1);
    } else if (lx == 0) {
      j = ix >> (20 - k);
      if ((j << (20 - k)) == ix) r = 2 - (j & 1);
	}
  }
  return r;
}


/* Allocates a string buffer by internally determining its most efficient size, aligned along the "long" boundary. The
   return is a char* pointer to the beginning of the string. The function zeros only the last few bytes and assumes that
   the `leading` rest will be filled by real characters later on. Just pass l as the number of characters, excluding the
   terminating \0, and do not multiply it by sizeof(char). The function automatically adds a terminating \0.
   Works like lapi.c/agn_stralloc, 2.16.5. See also: tools_optstrlen. */
LUALIB_API INLINE char *tools_stralloc (size_t len) {
  char *buf;
  size_t rem, l;
  l = len;
  /* determine optimal size, including terminating \0 */
  rem = (++l) % AGN_BLOCKSIZE;  /* length including \0 */
  l += (rem != 0)*(AGN_BLOCKSIZE - rem);
  /* now commit memory */
  buf = (char *)malloc(l*sizeof(char));
  if (buf == NULL) return NULL;
  memset(buf + len, 0, l - len);  /* only zero the bytes that will not be filled by characters later on */
  return buf;
}


#if ULONG_MAX - 20L > 4294967275L
#define AGN_BLOCKSIZEGT32 1
#else
#define AGN_BLOCKSIZEGT32 0
#endif

#define charinchunk(p, mask)   (tools_strisnull((p) ^ (mask)))
#define createmask() { \
  mask = c0 << 8 | c0; \
  mask = mask << 16 | mask; \
}

/* similar to strpbrk, but only checks if any of the characters in p has been found in s; otherwise returns 0. Based on
   former tools_strpbrk which returns a pointer or NULL.
   1 % faster if there is a hit, 10 % faster if not. 2.16.12 */
LUALIB_API int tools_hasstrchr (const void *src, const void *p) {  /* p represents a SET of single chars */
#if AGN_BLOCKSIZEGT32 == 1
  int i;
#endif
  int32_t isunaligned;
  size_t len, l;
  unsigned char c0;
  unsigned long *s0, mask;
  const unsigned char *s, *c;
  s = (const unsigned char *)src;
  len = strlen((const char*)s);
  isunaligned = tools_strisunaligned(s);
  c0 = 0;
  l = len;
  for (c=p; *c; c++) {  /* iterate set of `needle` chars */
    c0 = uchar(*c);
    while (isunaligned && tools_strisunaligned(s)) {
      if (!l--) goto strpbrklabel;  /* proceed with next char in needle p, 2.16.12 patch */
      if (*s++ == c0) return 1;
    }
    if (tools_largeenough(l)) {  /* 4-/8-byte chunks instead of single chars can be compared */
      s0 = (unsigned long *)s;
      createmask();
#if AGN_BLOCKSIZEGT32 == 1
      for (i=32; i < AGN_BLOCKSIZE*8; i <<= 1)
        mask = (mask << i) | mask;
#endif
      while (tools_largeenough(l)) {  /* 4/8 byte chunks can be compared */
        if (charinchunk(*s0, mask)) return 1;  /* c has been found */
        l -= AGN_BLOCKSIZE;
        s0++;
      }
      s = (unsigned char *)s0;
    }
    while (l--) {  /* check the rest of the string, 2.16.12 patch */
      if (*s++ == c0) return 1;
    }
strpbrklabel:
    s = (const unsigned char *)src;  /* reset src pointer to its beginning */
    l = len;
  }  /* proceed with next char in needle p */
  return 0;
}


LUALIB_API INLINE void *tools_memchr (const void *src, int c, size_t l) {  /* 2.16.12 */
#if AGN_BLOCKSIZEGT32 == 1
  int i;
#endif
  const unsigned char *s;
  unsigned char c0;
  unsigned long *p, mask;
  s = (const unsigned char *)src;
  c0 = c;
  while (tools_strisunaligned(s)) {
    if (!l--) return NULL;
    if (*s == c0) return (void *)s;
    s++;
  }
  /* process 4 or 8 bytes at a time if string is long enough and aligned */
  if (tools_largeenough(l)) {
    p = (unsigned long *)s;
    createmask();
#if AGN_BLOCKSIZEGT32 == 1
    for (i=32; i < AGN_BLOCKSIZE*8; i <<= 1)
      mask = (mask << i) | mask;
#endif
    while (tools_largeenough(l)) {
      if (charinchunk(*p, mask)) break;  /* character found in chunk */
      l -= AGN_BLOCKSIZE;
      p++;
    }
    s = (unsigned char *)p;
  }
  /* move forward to character and return pointer to it */
  while (l--) {
    if (*s == c0) return (void *)s;
    s++;
  }
  return NULL;
}


/* l = number of chars to be checked in haystack */
LUALIB_API INLINE int tools_memcmp (const void *src, const void *p, size_t l) {  /* 2.16.12 */
  unsigned char *s, *t;
  s = (unsigned char *)src;
  t = (unsigned char *)p;
  if (tools_largeenough(l) && tools_stringsarealigned(s, t)) {
    /* process 4 or 8 bytes at a time if string src (s) is long enough and both strings are aligned */
    unsigned long *p, *q;
    p = (unsigned long *)s;
    q = (unsigned long *)t;
    while (tools_largeenough(l)) {
      if (*p != *q) break;  /* mismatch */
      p++; q++;
      l -= AGN_BLOCKSIZE;
    }
    s = (unsigned char*)p;
    t = (unsigned char*)q;
  }
  /* no further 4/8-byte chunk present, compare byte-by-byte */
  while (l--) {
    if (*s != *t) return *s - *t;
    s++; t++;
  }
  return 0;
}


/* 2.16.12, 10 % faster than lmemfind in lstrlib.c */
LUALIB_API INLINE const char *tools_lmemfind (const char *s1, size_t l1, const char *s2, size_t l2) {
  if (l2 == 0) return s1;  /* empty strings are everywhere */
  else if (l2 > l1) return NULL;  /* avoids a negative `l1' */
  else {
    const char *init;  /* to search for a `*s2' inside `s1' */
    l2--;  /* 1st char will be checked by `memchr' */
    l1 = l1 - l2;  /* `s2' cannot be found after that */
    while (l1 > 0 && (init = (const char *)tools_memchr(s1, *s2, l1)) != NULL) {
      init++;   /* 1st char is already checked */
      if (tools_memcmp(init, s2 + 1, l2) == 0)  /* compare the rest of the characters */
        return init - 1;
      else {  /* correct `l1' and `s1' to try again */
        l1 -= init - s1;
        s1 = init;
      }
    }
    return NULL;  /* not found */
  }
}


/* ************************** DO NOT USE OR REMOVE THE FOLLOWING FUNCTIONS ************************** */



